//
// Copyright (C) 2007 Institut fuer Telematik, Universitaet Karlsruhe (TH)
//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public License
// as published by the Free Software Foundation; either version 2
// of the License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
//

/**
 * @file DHTDataStorage.cc
 * @author Ingmar Baumgart
 */

#include <omnetpp.h>

#include "DHTDataStorage.h"

Define_Module(DHTDataStorage);

using namespace std;

std::ostream& operator<<(std::ostream& Stream, const DHTData entry)
{
    return Stream << "Value: " << entry.value
                  << " Endtime: " << entry.ttlMessage->arrivalTime()
                  << " Modifiable: " << entry.is_modifiable
                  << " SourceNode: " << entry.sourceNode;
}


void DHTDataStorage::initialize(int stage)
{
    if (stage != MIN_STAGE_APP)
        return;

    WATCH_MAP(dataMap);
}

void DHTDataStorage::handleMessage(cMessage* msg)
{
    error("This module doesn't handle messages!");
}

void DHTDataStorage::clear()
{
    map<OverlayKey, DHTData>::iterator iter;

    for( iter = dataMap.begin(); iter != dataMap.end(); iter++ ) {
        cancelAndDelete(iter->second.ttlMessage);
    }

    dataMap.clear();
}


uint DHTDataStorage::getSize()
{
    return dataMap.size();
}

bool DHTDataStorage::isEmpty()
{
    if (dataMap.size() == 0)
        return true;
    else
        return false;
}

const BinaryValue& DHTDataStorage::getData(const OverlayKey& key) {

    std::map<OverlayKey, DHTData>::iterator it =  dataMap.find(key);

    if (it == dataMap.end())
        return BinaryValue::UNSPECIFIED_VALUE;
    else
        return it->second.value;
}

const NodeHandle& DHTDataStorage::getSourceNode(const OverlayKey& key) {

    std::map<OverlayKey, DHTData>::iterator it =  dataMap.find(key);

    if (it == dataMap.end())
        return NodeHandle::UNSPECIFIED_NODE;
    else
        return it->second.sourceNode;
}

cMessage* DHTDataStorage::getTtlMessage(const OverlayKey& key){
    std::map<OverlayKey, DHTData>::iterator it =  dataMap.find(key);

    if (it == dataMap.end())
        return NULL;
    else
        return it->second.ttlMessage;
}

const bool DHTDataStorage::isModifiable(const OverlayKey& key) {

    std::map<OverlayKey, DHTData>::iterator it =  dataMap.find(key);

    if (it == dataMap.end())
        return true;
    else
        return it->second.is_modifiable;
}


const BinaryValue& DHTDataStorage::getDataAtPos(uint pos)
{
    if (pos >= dataMap.size()) {
        error("Index out of bound (DHTDataStorage, getDataAtPos())");
    }

    std::map<OverlayKey, DHTData>::iterator it = dataMap.begin();

    for (uint i= 0; i < pos; i++) {
        it++;
        if (i == (pos-1))
            return it->second.value;
    }

    return it->second.value;
}

const std::map<OverlayKey, DHTData>::iterator DHTDataStorage::begin()
{
    return dataMap.begin();
}

void DHTDataStorage::addData(OverlayKey key, BinaryValue value, cMessage* ttlMessage, bool is_modifiable, NodeHandle sourceNode, bool responsible)
{
    DHTData entry;
    entry.value = value;
    entry.ttlMessage = ttlMessage;
    entry.sourceNode = sourceNode;
    entry.is_modifiable = is_modifiable;
    entry.responsible = responsible;
    // replace with new value
    dataMap.erase(key);
    dataMap.insert(make_pair(key, entry));
}

void DHTDataStorage::removeData(OverlayKey key)
{
    dataMap.erase(key);
}

DhtDataVector* DHTDataStorage::get_range(const OverlayKey lkey, const OverlayKey rkey)
{
    DhtDataVector* vect = new DhtDataVector();
    map<OverlayKey, DHTData>::iterator iter;
    if (lkey > rkey) { //the zero key is between this to keys
        //we fetch everything after lkey
        for( iter = dataMap.lower_bound(lkey); iter != dataMap.end(); iter++ ) {
            vect->insert(vect->end(), std::make_pair(iter->first, iter->second));
        }
        //and everything before rkey
        for( iter = dataMap.upper_bound(rkey); iter != (dataMap.begin()--); iter-- ) {
            vect->insert(vect->end(), std::make_pair(iter->first, iter->second));
        }
    }
    else {
        for (iter = dataMap.lower_bound(lkey); iter != (dataMap.upper_bound(rkey) ++); iter++) {
            vect->insert(vect->end(), std::make_pair(iter->first, iter->second));
        }
    }

    return vect;
}

DhtDumpVector* DHTDataStorage::dumpDht()
{
    DhtDumpVector* vect = new DhtDumpVector();
    DhtDumpEntry entry;

    map<OverlayKey, DHTData>::iterator iter;

    for (iter = dataMap.begin(); iter != dataMap.end(); iter++) {
        entry.key = iter->first;
        entry.value = iter->second.value;
        entry.ttl = iter->second.ttlMessage->arrivalTime() - simTime();
        entry.owner = iter->second.sourceNode;
        entry.is_modifiable = iter->second.is_modifiable;
        entry.responsible = iter->second.responsible;
        vect->push_back(entry);
    }

    return vect;
}


void DHTDataStorage::updateDisplayString()
{
    // FIXME: doesn't work without tcl/tk
    //    	if (ev.isGUI()) {
    if (1) {
        char buf[80];

        if (dataMap.size() == 1) {
            sprintf(buf, "1 data item");
        } else {
            sprintf(buf, "%zi data items", dataMap.size());
        }

        displayString().setTagArg("t", 0, buf);
        displayString().setTagArg("t", 2, "blue");
    }

}

void DHTDataStorage::updateTooltip()
{
    if (ev.isGUI()) {
        std::stringstream str;
        for (uint i = 0; i < dataMap.size(); i++)	{
            str << getDataAtPos(i);

            if ( i != dataMap.size() - 1 )
                str << endl;
        }


        char buf[1024];
        sprintf(buf, "%s", str.str().c_str());
        displayString().setTagArg("tt", 0, buf);
    }
}

void DHTDataStorage::display()
{
    cout << "Content of DHTDataStorage:" << endl;
    for (std::map<OverlayKey, DHTData>::iterator it = dataMap.begin();
    it != dataMap.end(); it++) {
        cout << "Key: " << it->first << " Value: " << it->second.value << "End-time: " << it->second.ttlMessage->arrivalTime() << endl;
    }
}
