//
// Copyright (C) 2007 Institut fuer Telematik, Universitaet Karlsruhe (TH)
//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public License
// as published by the Free Software Foundation; either version 2
// of the License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
//

/**
 * @file DHTDataStorage.h
 * @author Ingmar Baumgart
 */

#ifndef __DHTDATASTORAGE_H_
#define __DHTDATASTORAGE_H_

#include <set>
#include <vector>
#include <sstream>

#include <omnetpp.h>

#include <NodeHandle.h>
#include <InitStages.h>
#include <BinaryValue.h>

/**
 * DHT data storage module
 *
 * This modul contains the data storage of the DHT implementation.
 *
 * @author Ingmar Baumgart
 * @see DHT
 */


struct DHTData
{
    BinaryValue value;
    cMessage* ttlMessage;
    bool is_modifiable;
    NodeHandle sourceNode;
    bool responsible; //is this node responsible for this key ?
    friend std::ostream& operator<<(std::ostream& Stream, const DHTData entry);
};

struct DhtDumpEntry
{
    OverlayKey key;
    BinaryValue value;
    int ttl;
    NodeHandle owner;
    bool is_modifiable;
    bool responsible;
};

typedef std::vector<DhtDumpEntry> DhtDumpVector;

typedef std::vector<std::pair<OverlayKey, DHTData> > DhtDataVector;

class DHTDataStorage : public cSimpleModule
{
  public:
    virtual int numInitStages() const
    {
        return MAX_STAGE_APP + 1;
    }
    virtual void initialize(int stage);
    virtual void handleMessage(cMessage* msg);

    /**
     * Returns number of stored data items in the map
     *
     * @return number of stored data items
     */
    virtual uint getSize();

    /**
     * Clears all stored data items
     */
    virtual void clear();

    /**
     * Checks if the data storage map is empty
     *
     * @return returns false if there are stored data items, true otherwise.
     */
    virtual bool isEmpty();

    /**
     * Returns the value of a stored data item with a given key
     *
     * @param key The key of the data item
     * @return The value of the data item with the given key
     */
    virtual const BinaryValue& getData(const OverlayKey& key);

    /**
     * Returns the source node of a stored data item with a given key
     *
     * @param key The key of the data item
     * @return The source node of the data item with the given key
     */
    virtual const NodeHandle& getSourceNode(const OverlayKey& key);

    /**
     * Returns the ttlMessage of a stored data item with a given key
     *
     * @param key The key of the data item
     * @return The ttlMessage of the data item with the given key
     */
    virtual cMessage* getTtlMessage(const OverlayKey& key);

    /**
     * Returns a boolean telling if this data if modifiable
     *
     * @param key The key of the data item
     * @return The value of the is_modifiable value
     */
    virtual const bool isModifiable(const OverlayKey& key);


    /**
     * Returns the value of the data item stored at position pos
     *
     * @param pos position in data storage map
     * @return The value of the data item at position pos
     */
    virtual const BinaryValue& getDataAtPos(uint pos = 0);

    /**
     * Returns an iterator to the beginning of the map
     *
     * @return An iterator
     */
    virtual const std::map<OverlayKey, DHTData>::iterator begin();

    /**
     * Store a new data item in the map
     *
     * @param key The key of the data item to be stored
     * @param value The value of the data item to be stored
     * @param ttlMessage The self-message sent for the ttl expiration
     * @param is_modifiable Flag that tell if the data can be change by anyone, or just by the sourceNode
     * @param sourceNode Node which asked to store the value
     * @param responsible
     */
    virtual void addData(OverlayKey key, BinaryValue value, cMessage* ttlMessage, bool is_modifiable=true, NodeHandle sourceNode=NodeHandle::UNSPECIFIED_NODE, bool responsible=true);

    /**
     * Removes a certain data item from the map
     *
     * @param key The key of the data item to be removed
     */
    virtual void removeData(OverlayKey key);
    void display ();

    /**
     * Return a vector containing every keys/values between the two bounds
     * @param lkey The beginning of the range
     * @param rkey The end of the range
     */
    DhtDataVector* get_range(const OverlayKey lkey, const OverlayKey rkey);
    
    /**
     * Dump all local data records into a vector
     * 
     * @return the vector containing all local data records
     */
    DhtDumpVector* dumpDht();

  protected:
    std::map<OverlayKey, DHTData> dataMap; /**< internal representation of the data storage */

    /**
     * Displays the current number of successors in the list
     */
    void updateDisplayString();

    /**
     * Displays the first 4 successor nodes as tooltip.
     */
    void updateTooltip();
};

#endif
