//
// Copyright (C) 2006 Institut fuer Telematik, Universitaet Karlsruhe (TH)
//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public License
// as published by the Free Software Foundation; either version 2
// of the License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
//

/**
 * @file KBRTestApp.cc
 * @author Bernhard Heep
 */

#include <IPAddressResolver.h>
#include <CommonMessages_m.h>
#include <GlobalStatistics.h>
#include <UnderlayConfigurator.h>
#include <BootstrapOracle.h>

#include "KBRTestApp.h"
#include "KBRTestMessage_m.h"

Define_Module(KBRTestApp);

KBRTestApp::KBRTestApp()
{
    testTimer = NULL;
}

KBRTestApp::~KBRTestApp()
{
    cancelAndDelete(testTimer);
}

void KBRTestApp::initializeApp(int stage)
{
    if (stage != MIN_STAGE_APP) {
        return;
    }

    lookupNodeIds = par("lookupNodeIds");
    mean = par("testMsgInterval");
    deviation = mean / 10;
    activeNetwInitPhase = par("activeNetwInitPhase");
    msgHandleBufSize = par("msgHandleBufSize");

    numDelivered = 0;
    bytesDelivered = 0;
    WATCH(numDelivered);
    WATCH(bytesDelivered);

    numDropped = 0;
    bytesDropped = 0;
    WATCH(numDropped);
    WATCH(bytesDropped);

    sequenceNumber = 0;

    nodeIsLeavingSoon = false;

    // initialize circular buffer
    if (msgHandleBufSize > 0) {
        mhBuf.resize(msgHandleBufSize);
        mhBufBegin = mhBuf.begin();
        mhBufEnd = mhBuf.end();
        mhBufNext = mhBufBegin;
    }

#if 0
    bindToPort(1025);
    thisNode.port = 1025;
#endif

    // start periodic timer
    testTimer = new cMessage("KBRTestApp Timer");
    scheduleAt(simulation.simTime() + truncnormal(mean, deviation), testTimer);
}

void KBRTestApp::handleTimerEvent(cMessage* msg)
{
    if (msg == testTimer) {
        // schedule next timer event
        scheduleAt(simulation.simTime() + truncnormal(mean, deviation), msg);

        // do nothing if the network is still in the initialization phase
        if ((!activeNetwInitPhase && underlayConfigurator->isInInitPhase())
                || underlayConfigurator->isSimulationEndingSoon()
                || nodeIsLeavingSoon) {

            return;
        }

        OverlayKey destKey;
        if (lookupNodeIds) {
            destKey = bootstrapOracle->getBootstrapNode().key;
            // do nothing if there are currently no nodes in the network
            if (destKey.isUnspecified())
                return;
        } else {
            // generate random destination key
            destKey = OverlayKey::random();
        }

        // create a 100 byte test message
        KBRTestMessage* testMsg = new KBRTestMessage("KBRTestMessage");
        testMsg->setId(id());
        testMsg->setSeqNum(sequenceNumber++);
        testMsg->setByteLength(100);
        testMsg->setMeasurementPhase(globalStatistics->isMeasuring());

        RECORD_STATS(globalStatistics->sentKBRTestAppMessages++);

        callRoute(destKey, testMsg);

#if 0
        thisNode.port = 1025;
        NodeHandle handle = bootstrapOracle->getBootstrapNode();
        handle.port = 1025;
        pingNode(handle, -1, -1, NULL, "TestPING", NULL, -1, UDP_TRANSPORT);

        LookupCall* call = new LookupCall();
        call->setLength(0);
        call->setKey(destKey);
        call->setNumSiblings(overlay->getMaxNumSiblings());
        sendInternalRpcCall(OVERLAY_COMP, call);
#endif

    }
}
void KBRTestApp::pingResponse(PingResponse* response, cPolymorphic* context,
                              int rpcId, simtime_t rtt)
{
    //std::cout << rtt << std::endl;
}

void KBRTestApp::handleRpcResponse(BaseResponseMessage* msg,
                                   cPolymorphic* context, int rpcId,
                                   simtime_t rtt)
{
    RPC_SWITCH_START(msg)
    RPC_ON_RESPONSE(Lookup) {
        handleLookupResponse(_LookupResponse);
        EV << "[KBRTestApp::handleRpcResponse() @ " << overlay->getThisNode().ip
           << " (" << overlay->getThisNode().key.toString(16) << ")]\n"
           << "    Lookup RPC Response received: id=" << rpcId << "\n"
           << "    msg=" << *_LookupResponse << " rtt=" << rtt
           << endl;
        break;
    }
    RPC_SWITCH_END( )
}

void KBRTestApp::handleLookupResponse(LookupResponse* msg)
{
    EV << "[KBRTestApp::handleLookupResponse() @ " << overlay->getThisNode().ip
       << " (" << overlay->getThisNode().key.toString(16) << ")]\n"
       << "    Lookup response for key " << msg->getKey()<< " : ";
    if (msg->getIsValid()) {
        for (uint i=0; i<msg->getSiblingsArraySize(); i++) {
            EV << endl << "    " << msg->getSiblings(i);
        }
    } else {
        EV << "failed";
    }

    EV << endl;
}

void KBRTestApp::handleNodeLeaveNotification()
{
    nodeIsLeavingSoon = true;
}


void KBRTestApp::deliver(OverlayKey& key, cMessage* msg)
{
    KBRTestMessage* testMsg = check_and_cast<KBRTestMessage*>(msg);
    OverlayCtrlInfo* overlayCtrlInfo =
        check_and_cast<OverlayCtrlInfo*>(msg->removeControlInfo());

    if (overlay->getThisNode().key.isUnspecified())
        error("key");

    // check for duplicate
    if ((msgHandleBufSize > 0 )
            && checkSeen(overlayCtrlInfo->getSrcNode().key, testMsg->getSeqNum())) {
        EV << "KBRTestApp: duplicate dropped." << std::endl;
        delete overlayCtrlInfo;
        delete testMsg;
        return;
    }

    // Return statistical data to the sender.
    if (cModule* mod = simulation.module(testMsg->getId())) {
        if (KBRTestApp* sender = dynamic_cast<KBRTestApp*>(mod)) {
            if ((!lookupNodeIds) || (overlay->getThisNode().key == key)) {
                if (testMsg->getMeasurementPhase() == true) {
                        sender->evaluateData((simTime() - testMsg->creationTime()),
                                             overlayCtrlInfo->getHopCount(),
                                             testMsg->byteLength());
                }
            } else if(lookupNodeIds) {
                RECORD_STATS(numDropped++;
                             bytesDropped += testMsg->byteLength());
                EV << "[KBRTestApp::deliver() @ " << overlay->getThisNode().ip
                << " (" << overlay->getThisNode().key.toString(16) << ")]\n"
                << "    Error: Lookup of NodeIDs and KBRTestMessage"
                << " received with different destKey!"
                << endl;
            }
        }
    }

    EV << "[KBRTestApp::deliver() @ " << overlay->getThisNode().ip
       << " (" << overlay->getThisNode().key.toString(16) << ")]\n"
       << "    Received \"" << testMsg->name() << "(seqNr: "
       << testMsg->getSeqNum() << ")\n"
       << "    with destination key: " << key.toString(16)
       << endl;

    delete overlayCtrlInfo;
    delete testMsg;
}

void KBRTestApp::forward(OverlayKey* key, cMessage** msg,
                         NodeHandle* nextHopNode)
{
    KBRTestMessage* tempMsg = check_and_cast<KBRTestMessage*>(*msg);

    tempMsg->setVisitedNodesArraySize(tempMsg->getVisitedNodesArraySize() + 1);
    tempMsg->setVisitedNodes(tempMsg->getVisitedNodesArraySize() - 1,
                             overlay->getThisNode().ip);
}

bool KBRTestApp::checkSeen(const OverlayKey& key, int seqNum)
{
    MsgHandle hdl(key, seqNum);

    for (MsgHandleBuf::iterator it = mhBufBegin; it != mhBufEnd; ++it) {
        if (it->key.isUnspecified()) {
            continue;
        }

        if (*it == hdl) {
            return true;
        }
    }

    *(mhBufNext++) = hdl;
    if (mhBufNext == mhBufEnd) {
        mhBufNext = mhBufBegin;
    }
    return false;
}

void KBRTestApp::evaluateData(simtime_t latency, int hopCount, long int bytes)
{
    // count the number and size of successfully delivered messages
    RECORD_STATS(numDelivered++; bytesDelivered += bytes;
                 globalStatistics->deliveredKBRTestAppMessages++);

    if (numOverlaySent < numDelivered) {
        error("KBRTestApp::evaluateData(): numSent < numDelivered!");
    }

    RECORD_STATS(globalStatistics->recordOutVector("KBRTestApp: KBR Hop "
                                                   "Count", hopCount));
    RECORD_STATS(globalStatistics->recordOutVector("KBRTestApp: KBR Latency",
                                                   latency));
}

void KBRTestApp::finishApp()
{
    simtime_t time = globalStatistics->calcMeasuredLifetime(creationTime);

    if (time != 0) {
        globalStatistics->addStdDev("KBRTestApp: Delivered Messages/s",
                                    numDelivered / time);
        globalStatistics->addStdDev("KBRTestApp: Delivered Bytes/s",
                                    bytesDelivered / time);
        globalStatistics->addStdDev("KBRTestApp: Dropped Messages/s",
                                    numDropped / time);
        globalStatistics->addStdDev("KBRTestApp: Dropped Bytes/s",
                                    bytesDropped / time);
        if (numOverlaySent != 0) {
            globalStatistics->addStdDev("KBRTestApp: KBR Delivery Ratio",
                                        (float)numDelivered /
                                        (float)numOverlaySent);
        } else {
            globalStatistics->addStdDev("KBRTestApp: KBR Delivery Ratio", 0);
        }
    }
}

