//
// Copyright (C) 2006 Institut fuer Telematik, Universitaet Karlsruhe (TH)
//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public License
// as published by the Free Software Foundation; either version 2
// of the License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
//

/**
 * @file BaseOverlay.cc
 * @author Bernhard Heep (initial)
 * @author Sebastian Mies (rpcs)
 */


#include <assert.h>

#include <RpcMacros.h>

#include <UDPAppBase.h>
#include <UDPSocket.h>
#include <IPAddressResolver.h>
#include <NotificationBoard.h>

#include <BootstrapOracleAccess.h>
#include <UnderlayConfiguratorAccess.h>
#include <GlobalStatisticsAccess.h>
#include <GlobalParametersAccess.h>

#include <LookupListener.h>
#include <RecursiveLookup.h>
#include <IterativeLookup.h>

#include <BootstrapList.h>

#include "BaseOverlay.h"

using namespace std;

//------------------------------------------------------------------------
//--- Initialization & finishing -----------------------------------------
//------------------------------------------------------------------------

BaseOverlay::BaseOverlay()
{
    bootstrapOracle = NULL;
    underlayConfigurator = NULL;
    notificationBoard = NULL;
    globalParameters = NULL;
    bootstrapList = NULL;
}

BaseOverlay::~BaseOverlay()
{
    finishLookups();
    finishRpcs();
}

int BaseOverlay::numInitStages() const
{
    return NUM_STAGES_ALL;
}

void BaseOverlay::initialize(int stage)
{
    if (stage == MIN_STAGE_OVERLAY) {
        OverlayKey::setKeyLength(par("keyLength"));

        // find friend modules
        bootstrapOracle = BootstrapOracleAccess().get();
        underlayConfigurator = UnderlayConfiguratorAccess().get();
        notificationBoard = NotificationBoardAccess().get();
        globalParameters = GlobalParametersAccess().get();
        bootstrapList = check_and_cast<BootstrapList*>(parentModule()->
                parentModule()->submodule("bootstrapList", 0));

        // fetch some parameters
        debugOutput = par("debugOutput");
        collectPerHopDelay = par("collectPerHopDelay");
        localPort = par("localPort");
        hopCountMax = par("hopCountMax");
        drawOverlayTopology = par("drawOverlayTopology");

        // we assume most overlays don't provide KBR services
        kbr = false;

        // set routing type
        std::string temp(par("routingType"));
        if (temp == "iterative")
            defaultRoutingType = ITERATIVE_ROUTING;
        else if (temp == "exhaustive-iterative")
            defaultRoutingType = EXHAUSTIVE_ITERATIVE_ROUTING;
        else if (temp == "semi-recursive")
            defaultRoutingType = SEMI_RECURSIVE_ROUTING;
        else if (temp == "full-recursive")
            defaultRoutingType = FULL_RECURSIVE_ROUTING;
        else if (temp == "source-routing-recursive")
            defaultRoutingType = RECURSIVE_SOURCE_ROUTING;
        else throw new cRuntimeError((std::string("Wrong routing type: ")
                                      + temp).c_str());

        collectPerHopDelay = par("collectPerHopDelay");
        useCommonAPIforward = par("useCommonAPIforward");
        routeMsgAcks = par("routeMsgAcks");
        recNumRedundantNodes = par("recNumRedundantNodes");

        // set base lookup parameters
        iterativeLookupConfig.redundantNodes = par("lookupRedundantNodes");
        iterativeLookupConfig.parallelPaths = par("lookupParallelPaths");
        iterativeLookupConfig.parallelRpcs = par("lookupParallelRpcs");
        iterativeLookupConfig.secure = par("lookupSecure");
        iterativeLookupConfig.merge = par("lookupMerge");
        iterativeLookupConfig.failedNodeRpcs = par("lookupFailedNodeRpcs");
        iterativeLookupConfig.strictParallelRpcs =
            par("lookupStrictParallelRpcs");
        iterativeLookupConfig.useAllParallelResponses =
            par("lookupUseAllParallelResponses");
        iterativeLookupConfig.newRpcOnEveryTimeout =
            par("lookupNewRpcOnEveryTimeout");
        iterativeLookupConfig.newRpcOnEveryResponse =
            par("lookupNewRpcOnEveryResponse");
        iterativeLookupConfig.finishOnFirstUnchanged =
            par("lookupFinishOnFirstUnchanged");

        // statistics
        numAppDataSent = 0;
        bytesAppDataSent = 0;
        numAppLookupSent = 0;
        bytesAppLookupSent = 0;
        numMaintenanceSent = 0;
        bytesMaintenanceSent = 0;
        numAppDataReceived = 0;
        bytesAppDataReceived = 0;
        numAppLookupReceived = 0;
        bytesAppLookupReceived = 0;
        numMaintenanceReceived = 0;
        bytesMaintenanceReceived = 0;
        numAppDataForwarded = 0;
        bytesAppDataForwarded = 0;
        numAppLookupForwarded = 0;
        bytesAppLookupForwarded = 0;
        numMaintenanceForwarded = 0;
        bytesMaintenanceForwarded = 0;

        numDropped = 0;
        bytesDropped = 0;
        numFindNodeSent = 0;
        bytesFindNodeSent = 0;
        numFindNodeResponseSent = 0;
        bytesFindNodeResponseSent = 0;
        numFailedNodeSent = 0;
        bytesFailedNodeSent = 0;
        numFailedNodeResponseSent = 0;
        bytesFailedNodeResponseSent = 0;

        numInternalSent = 0;
        bytesInternalSent = 0;
        numInternalReceived = 0;
        bytesInternalReceived = 0;

        WATCH(numAppDataSent);
        WATCH(bytesAppDataSent);
        WATCH(numAppLookupSent);
        WATCH(bytesAppLookupSent);
        WATCH(numMaintenanceSent);
        WATCH(bytesMaintenanceSent);
        WATCH(numAppDataReceived);
        WATCH(bytesAppDataReceived);
        WATCH(numAppLookupReceived);
        WATCH(bytesAppLookupReceived);
        WATCH(numMaintenanceReceived);
        WATCH(bytesMaintenanceReceived);
        WATCH(numAppDataForwarded);
        WATCH(bytesAppDataForwarded);
        WATCH(numAppLookupForwarded);
        WATCH(bytesAppLookupForwarded);
        WATCH(numMaintenanceForwarded);
        WATCH(bytesMaintenanceForwarded);

        WATCH(numDropped);
        WATCH(bytesDropped);
        WATCH(numFindNodeSent);
        WATCH(bytesFindNodeSent);
        WATCH(numFindNodeResponseSent);
        WATCH(bytesFindNodeResponseSent);
        WATCH(numFailedNodeSent);
        WATCH(bytesFailedNodeSent);
        WATCH(numFailedNodeResponseSent);
        WATCH(bytesFailedNodeResponseSent);

        if (isInSimpleMultiOverlayHost()) {
            WATCH(numInternalSent);
            WATCH(bytesInternalSent);
            WATCH(numInternalReceived);
            WATCH(bytesInternalReceived);
        }

        // set up local nodehandle
        thisNode.ip = IPAddressResolver().
                      addressOf(parentModule()->parentModule()).get4();
        thisNode.key = OverlayKey::UNSPECIFIED_KEY;

        // set up UDP
        bindToPort(localPort);

        // subscribe to the notification board
        notificationBoard->subscribe(this, NF_OVERLAY_TRANSPORTADDRESS_CHANGED);
        notificationBoard->subscribe(this, NF_OVERLAY_NODE_LEAVE);
        notificationBoard->subscribe(this, NF_OVERLAY_NODE_GRACEFUL_LEAVE);

        // init visualization with terminal ptr
        if (drawOverlayTopology)
            initVis(parentModule()->parentModule());

        // init rpcs
        initRpcs();
        initLookups();

        // statistics
        globalStatistics->nodesInitialized++;
        creationTime = simTime();
    }

    if (stage >= MIN_STAGE_OVERLAY && stage <= MAX_STAGE_OVERLAY)
        initializeOverlay(stage);

    if (stage == MAX_STAGE_TIER_1) {
        // bootstrapList registered its gate to the overlay 0
        // if this is not overlay 0, we may not have the gate, so retrieve it
	    // this assumes that the overlay is in a container module!
        if (!compModuleList.count(BOOTSTRAPLIST_COMP)) {
            BaseOverlay *firstOverlay = dynamic_cast<BaseOverlay*>
                    (parentModule()->parentModule()
                    ->submodule("overlay", 0)->gate("from_app")
                    ->toGate()->ownerModule());
            if (!firstOverlay) {
                opp_error("BaseOverlay.cc: Couldn't obtain bootstrap gate");
            }
            registerComp(BOOTSTRAPLIST_COMP,
                         firstOverlay->getCompModule(BOOTSTRAPLIST_COMP));
        }
    }

    // join after all components are registered
    if (stage == MAX_STAGE_TIER_3) {
        if ((bool)par("joinOnApplicationRequest") == false) {
            join();
        }
    }
}

void BaseOverlay::initializeOverlay(int stage)
{
}

void BaseOverlay::finish()
{
    finishOverlay();

    globalStatistics->nodesFinished++;

    simtime_t time = globalStatistics->calcMeasuredLifetime(creationTime);
    if (time == 0) return;

    if (collectPerHopDelay) {
        std::ostringstream singleHopName;
        HopDelayRecord* hdrl = NULL;
        HopDelayRecord* hdr = NULL;
        for (size_t i = 0; i < singleHopDelays.size();) {
            hdrl = singleHopDelays[i++];
            hdr = hdrl;
            for (size_t j = 1; j <= i; ++j) {
                if (hdr->count == 0) continue;
                singleHopName.str("");
                singleHopName << "BaseOverlay: Average Delay in Hop "
                              << j << " of " << i;
                globalStatistics->addStdDev(singleHopName.str(),
                                            hdr->val / hdr->count);
                ++hdr;
            }
            delete[] hdrl;
        }
        singleHopDelays.clear();
    }

    globalStatistics->addStdDev("BaseOverlay: Sent App Data Messages/s",
                                numAppDataSent / time);
    globalStatistics->addStdDev("BaseOverlay: Sent App Data Bytes/s",
                                bytesAppDataSent / time);
    if (isInSimpleMultiOverlayHost()) {
        globalStatistics->addStdDev("BaseOverlay: Internal Sent Messages/s",
                                    numInternalReceived / time);
        globalStatistics->addStdDev("BaseOverlay: Internal Sent Bytes/s",
                                    bytesInternalReceived / time);
    }
    globalStatistics->addStdDev("BaseOverlay: Sent App Lookup Messages/s",
                                numAppLookupSent / time);
    globalStatistics->addStdDev("BaseOverlay: Sent App Lookup Bytes/s",
                                bytesAppLookupSent / time);
    globalStatistics->addStdDev("BaseOverlay: Sent Maintenance Messages/s",
                                numMaintenanceSent / time);
    globalStatistics->addStdDev("BaseOverlay: Sent Maintenance Bytes/s",
                                bytesMaintenanceSent / time);

    globalStatistics->addStdDev("BaseOverlay: Sent Total Messages/s",
                                (numAppDataSent + numAppLookupSent +
                                 numMaintenanceSent) / time);
    globalStatistics->addStdDev("BaseOverlay: Sent Total Bytes/s",
                                (bytesAppDataSent + bytesAppLookupSent +
                                 bytesMaintenanceSent) / time);
    globalStatistics->addStdDev("BaseOverlay: Sent FindNode Messages/s",
                                numFindNodeSent / time);
    globalStatistics->addStdDev("BaseOverlay: Sent FindNode Bytes/s",
                                bytesFindNodeSent / time);

    globalStatistics->addStdDev("BaseOverlay: Sent FindNodeResponse Messages/s",
                                numFindNodeResponseSent / time);
    globalStatistics->addStdDev("BaseOverlay: Sent FindNodeResponse Bytes/s",
                                bytesFindNodeResponseSent / time);
    globalStatistics->addStdDev("BaseOverlay: Sent FailedNode Messages/s",
                                numFailedNodeSent / time);
    globalStatistics->addStdDev("BaseOverlay: Sent FailedNode Bytes/s",
                                bytesFailedNodeSent / time);
    globalStatistics->addStdDev("BaseOverlay: Sent FailedNodeResponse Messages/s",
                                numFailedNodeResponseSent / time);
    globalStatistics->addStdDev("BaseOverlay: Sent FailedNodeResponse Bytes/s", bytesFailedNodeResponseSent / time);

    globalStatistics->addStdDev("BaseOverlay: Received App Data Messages/s",
                                numAppDataReceived / time);
    globalStatistics->addStdDev("BaseOverlay: Received App Data Bytes/s",
                                bytesAppDataReceived / time);
    if (isInSimpleMultiOverlayHost()) {
        globalStatistics->addStdDev("BaseOverlay: Internal Received Messages/s",
                                    numInternalReceived / time);
        globalStatistics->addStdDev("BaseOverlay: Internal Received Bytes/s",
                                    bytesInternalReceived / time);
    }
    globalStatistics->addStdDev("BaseOverlay: Received App Lookup Messages/s",
                                numAppLookupReceived / time);
    globalStatistics->addStdDev("BaseOverlay: Received App Lookup Bytes/s",
                                bytesAppLookupReceived / time);
    globalStatistics->addStdDev("BaseOverlay: Received Maintenance Messages/s",
                                numMaintenanceReceived / time);
    globalStatistics->addStdDev("BaseOverlay: Received Maintenance Bytes/s",
                                bytesMaintenanceReceived / time);

    globalStatistics->addStdDev("BaseOverlay: Received Total Messages/s",
                                (numAppDataReceived + numAppLookupReceived +
                                 numMaintenanceReceived)/time);
    globalStatistics->addStdDev("BaseOverlay: Received Total Bytes/s",
                                (bytesAppDataReceived + bytesAppLookupReceived +
                                 bytesMaintenanceReceived)/time);
    globalStatistics->addStdDev("BaseOverlay: Forwarded App Data Messages/s",
                                numAppDataForwarded / time);
    globalStatistics->addStdDev("BaseOverlay: Forwarded App Data Bytes/s",
                                bytesAppDataForwarded / time);
    globalStatistics->addStdDev("BaseOverlay: Forwarded App Lookup Messages/s",
                                numAppLookupForwarded / time);
    globalStatistics->addStdDev("BaseOverlay: Forwarded Signaling Bytes/s",
                                bytesAppLookupForwarded / time);
    globalStatistics->addStdDev("BaseOverlay: Forwarded Maintenance Messages/s",
                                numMaintenanceForwarded / time);
    globalStatistics->addStdDev("BaseOverlay: Forwarded Maintenance Bytes/s",
                                bytesMaintenanceForwarded / time);
    globalStatistics->addStdDev("BaseOverlay: Forwarded Total Messages/s",
                                (numAppDataForwarded + numAppLookupForwarded +
                                 numMaintenanceForwarded) / time);
    globalStatistics->addStdDev("BaseOverlay: Forwarded Total Bytes/s",
                                (bytesAppDataForwarded + bytesAppLookupForwarded +
                                 bytesMaintenanceForwarded) / time);

    globalStatistics->addStdDev("BaseOverlay: Dropped Messages/s",
                                numDropped / time);
    globalStatistics->addStdDev("BaseOverlay: Dropped Bytes/s",
                                bytesDropped / time);

    globalStatistics->addStdDev("BaseOverlay: Measured Session Time",
                                simTime() - creationTime);

    globalStatistics->addStdDev("BaseRpc: Sent Ping Messages/s",
                                numPingSent / time);
    globalStatistics->addStdDev("BaseRpc: Sent Ping Bytes/s",
                                bytesPingSent / time);
    globalStatistics->addStdDev("BaseRpc: Sent Ping Response Messages/s",
                                numPingResponseSent / time);
    globalStatistics->addStdDev("BaseRpc: Sent Ping Response Bytes/s",
                                bytesPingResponseSent / time);

    globalStatistics->doFinish();

}

void BaseOverlay::finishOverlay()
{
}

//------------------------------------------------------------------------
//--- General Overlay Parameters (getter and setters) --------------------
//------------------------------------------------------------------------
bool BaseOverlay::isMalicious()
{
    return bootstrapOracle->isMalicious(getThisNode());
}

CompType BaseOverlay::getThisCompType()
{
    return OVERLAY_COMP;
}

//------------------------------------------------------------------------
//--- UDP functions copied from the INET framework .----------------------
//------------------------------------------------------------------------
void BaseOverlay::bindToPort(int port)
{
    EV << "[BaseOverlay::bindToPort() @ " << thisNode.ip
       << " (" << thisNode.key.toString(16) << ")]\n"
       << "    Binding to UDP port " << port
       << endl;

    thisNode.port = port;

    // TODO UDPAppBase should be ported to use UDPSocket sometime, but for now
    // we just manage the UDP socket by hand...
    cMessage *msg = new cMessage("UDP_C_BIND", UDP_C_BIND);
    UDPControlInfo *ctrl = new UDPControlInfo();
    ctrl->setSrcPort(port);
    ctrl->setSockId(UDPSocket::generateSocketId());
    msg->setControlInfo(ctrl);
    send(msg, "to_udp");
}


//------------------------------------------------------------------------
//--- Overlay Common API: Key-based Routing ------------------------------
//------------------------------------------------------------------------

void BaseOverlay::callDeliver(BaseOverlayMessage* msg,
                              const OverlayKey& destKey)
{
    KBRdeliver* deliverMsg = new KBRdeliver();

    OverlayCtrlInfo* overlayCtrlInfo =
        check_and_cast<OverlayCtrlInfo*>(msg->removeControlInfo());

    BaseAppDataMessage* appDataMsg = dynamic_cast<BaseAppDataMessage*>(msg);

    // TODO GIA
    if (appDataMsg != NULL) {
        overlayCtrlInfo->setSrcComp(appDataMsg->getSrcComp());
        overlayCtrlInfo->setDestComp(appDataMsg->getDestComp());
    }

    deliverMsg->setControlInfo(overlayCtrlInfo);
    deliverMsg->setDestKey(destKey);
    deliverMsg->encapsulate(msg->decapsulate());
    deliverMsg->setType(KBR_DELIVER);

    cGate* destGate = getCompRpcGate(static_cast<CompType>(
            overlayCtrlInfo->getDestComp()));

    if (destGate == NULL) {
        opp_error("BaseOverlay::callDeliver(): Unknown destComp!");
    }

    sendDirect(deliverMsg, 0, destGate);

    delete msg;
}

void BaseOverlay::callForward(const OverlayKey& key, BaseRouteMessage* msg,
                              const NodeHandle& nextHopNode)
{
    KBRforward* forwardMsg = new KBRforward();

    forwardMsg->setDestKey(msg->getDestKey());
    forwardMsg->setNextHopNode(nextHopNode);
    forwardMsg->encapsulate(msg->encapsulatedMsg()->decapsulate());

    OverlayCtrlInfo* overlayCtrlInfo =
        new OverlayCtrlInfo();
    overlayCtrlInfo->setTransportType(ROUTE_TRANSPORT);
    overlayCtrlInfo->setRoutingType(msg->getRoutingType());
    overlayCtrlInfo->setHopCount(msg->getHopCount());
    overlayCtrlInfo->setSrcNode(msg->getSrcNode());
    overlayCtrlInfo->setSrcComp(check_and_cast<BaseAppDataMessage*>
        (msg->encapsulatedMsg())->getSrcComp());
    overlayCtrlInfo->setDestComp(check_and_cast<BaseAppDataMessage*>
        (msg->encapsulatedMsg())->getDestComp());

    if (msg->controlInfo() != NULL) {
        OverlayCtrlInfo* ctrlInfo =
            check_and_cast<OverlayCtrlInfo*>(msg->removeControlInfo());

        overlayCtrlInfo->setLastHop(ctrlInfo->getLastHop());

        delete ctrlInfo;
    }

    forwardMsg->setControlInfo(overlayCtrlInfo);

    forwardMsg->setType(KBR_FORWARD);

    send(forwardMsg, "to_app");

    delete msg;
}

NodeVector* BaseOverlay::local_lookup(const OverlayKey& key,
                                      int num, bool safe)
{
    Enter_Method("local_lookup()");

    if (safe == true) {
        error("BaseOverlay::local_lookup(): safe flag is not implemented!");
    }

    NodeVector* nodeVector = findNode(key, min(num, getMaxNumRedundantNodes()),
                                      min(num,getMaxNumSiblings()));

    if (((int)nodeVector->size()) > num)
        nodeVector->resize(num);

    return nodeVector;
}

void BaseOverlay::join(const OverlayKey& nodeID)
{
    Enter_Method("join()");

    // set nodeID and IP
    thisNode.ip =
        IPAddressResolver().addressOf(parentModule()->parentModule()).get4();
    if (nodeID.isUnspecified())  {
        setOwnNodeID();
    } else {
        thisNode.key = nodeID;
    }

    joinOverlay();
}

void BaseOverlay::joinForeignPartition(const NodeHandle& node)
{
    throw new cRuntimeError("BaseOverlay::joinForeignPartition(): "
                            "This overlay doesn't support merging!");
}

void BaseOverlay::setOwnNodeID()
{
    thisNode.key = OverlayKey::random();
}

NodeVector* BaseOverlay::neighborSet(int num)
{
    Enter_Method("neighborSet()");

    return local_lookup(thisNode.key, num, false);
}

void BaseOverlay::callUpdate(const NodeHandle& node, bool joined)
{
    if ((!node.isUnspecified()) && (node != thisNode)) {
        if (joined) {
            EV << "[BaseOverlay::callUpdate() @ " << thisNode.ip
               << " (" << thisNode.key.toString(16) << ")]\n"
               << "    (" << node << ", " << joined << ") joined"
               << endl;
        } else {
            EV << "[BaseOverlay::callUpdate() @ " << thisNode.ip
               << " (" << thisNode.key.toString(16) << ")]\n"
               << "    (" << node << ", " << joined << ") left"
               << endl;
        }
    }

    KBRupdate* updateMsg = new KBRupdate("UPDATE");

    updateMsg->setNode(node);
    updateMsg->setJoined(joined);

    updateMsg->setType(KBR_UPDATE);

    send(updateMsg, "to_app");
}

bool BaseOverlay::isSiblingFor(const NodeHandle& node, const OverlayKey& key,
                               int numSiblings, bool* err)
{
    Enter_Method("isSiblingFor()");

    opp_error( "isSiblingFor: Not implemented!" );

    return false;
}

int BaseOverlay::getMaxNumSiblings()
{
    Enter_Method("getMaxNumSiblings()");

    opp_error( "getMaxNumSiblings: Not implemented!" );

    return false;
}

int BaseOverlay::getMaxNumRedundantNodes()
{
    Enter_Method("getMaxNumRedundantNodes()");

    opp_error( "getMaxNumRedundantNodes: Not implemented!" );

    return false;
}


//------------------------------------------------------------------------
//--- Message Handlers ---------------------------------------------------
//------------------------------------------------------------------------

//private
void BaseOverlay::handleMessage(cMessage* msg)
{
    // process self-messages
    if (msg->isSelfMessage()) {
        // process rpc self-messages
        BaseRpcMessage* rpcMessage = dynamic_cast<BaseRpcMessage*>(msg);
        if (rpcMessage!=NULL) {
            internalHandleRpcMessage(rpcMessage);
            return;
        }
        // process all other self-messages
        handleTimerEvent(msg);
    }

    // process messages from UDP
    else if (msg->arrivedOn("from_udp")) {
        UDPControlInfo* udpControlInfo =
            check_and_cast<UDPControlInfo*>(msg->removeControlInfo());
        OverlayCtrlInfo* overlayCtrlInfo = new OverlayCtrlInfo;
        overlayCtrlInfo
            ->setLastHop(TransportAddress(udpControlInfo->srcAddr(),
                                          udpControlInfo->srcPort()));
        overlayCtrlInfo->setTransportType(UDP_TRANSPORT);

        msg->setControlInfo(overlayCtrlInfo);
        delete udpControlInfo;

        // debug message
        if (debugOutput) {
            EV << "[BaseOverlay:handleMessage() @ " << thisNode.ip
            << " (" << thisNode.key.toString(16) << ")]\n"
            << "    Received " << *msg << " from "
            << overlayCtrlInfo->getLastHop().ip << endl;
        }

        BaseOverlayMessage* baseOverlayMsg =
            dynamic_cast<BaseOverlayMessage*>(msg);

        if (baseOverlayMsg == NULL) {
            RECORD_STATS(numDropped++; bytesDropped += msg->byteLength());
            delete msg;
            return;
        }

        // records stats if message is not a UDP "self message"
        if (overlayCtrlInfo->getLastHop() != thisNode) {
            // is this from anywhere else?
            if (baseOverlayMsg->getStatType() == APP_DATA_STAT)
                RECORD_STATS(numAppDataReceived++; bytesAppDataReceived +=
                             baseOverlayMsg->byteLength());
            else if (baseOverlayMsg->getStatType() == APP_LOOKUP_STAT)
                RECORD_STATS(numAppLookupReceived++;bytesAppLookupReceived +=
                             baseOverlayMsg->byteLength());
            else // MAINTENANCE_STAT
                RECORD_STATS(numMaintenanceReceived++;
                             bytesMaintenanceReceived +=
                                 baseOverlayMsg->byteLength());
        }
        if (overlayCtrlInfo->getLastHop().ip == thisNode.ip) {
            // is this from the same node?
            RECORD_STATS(numInternalReceived++; bytesInternalReceived +=
                             baseOverlayMsg->byteLength());
        }

        handleBaseOverlayMessage(baseOverlayMsg);
    }

    // process CommonAPIMessages from App
    else if (dynamic_cast<CommonAPIMessage*>(msg) != NULL) {
        if (dynamic_cast<KBRroute*>(msg) != NULL) {
            KBRroute* apiMsg = static_cast<KBRroute*>(msg);

            std::vector<TransportAddress> sourceRoute;
            for (uint i = 0; i < apiMsg->getSourceRouteArraySize(); ++i)
                sourceRoute.push_back(apiMsg->getSourceRoute(i));

            route(apiMsg->getDestKey(), static_cast<CompType>(apiMsg->getDestComp()),
                  static_cast<CompType>(apiMsg->getSrcComp()), apiMsg->decapsulate(),
                          sourceRoute);
        } else if (dynamic_cast<KBRforward*>(msg) != NULL) {
            KBRforward* apiMsg = static_cast<KBRforward*>(msg);
            OverlayCtrlInfo* overlayCtrlInfo =
                check_and_cast<OverlayCtrlInfo*>
                (msg->removeControlInfo());

            BaseAppDataMessage* dataMsg =
                new BaseAppDataMessage();
            dataMsg->setType(APPDATA);
            dataMsg->setLength(BASEAPPDATA_L(dataMsg));
            dataMsg->setName(msg->encapsulatedMsg()->name());
            dataMsg->encapsulate(msg->decapsulate());
            dataMsg->setSrcComp(overlayCtrlInfo->getSrcComp());
            dataMsg->setDestComp(overlayCtrlInfo->getDestComp());
            dataMsg->setStatType(APP_DATA_STAT);

            BaseRouteMessage* routeMsg = new BaseRouteMessage(dataMsg->name());
            routeMsg->setType(OVERLAYROUTE);
            routeMsg->setLength(BASEROUTE_L(routeMsg));
            routeMsg->encapsulate(dataMsg);

            routeMsg->setStatType(APP_DATA_STAT);
            routeMsg->setRoutingType(overlayCtrlInfo->getRoutingType());
            routeMsg->setDestKey(apiMsg->getDestKey());
            routeMsg->setSrcNode(overlayCtrlInfo->getSrcNode());
            routeMsg->setHopCount(overlayCtrlInfo->getHopCount());
            routeMsg->setControlInfo(overlayCtrlInfo);

            // message marked with this-pointer as already forwarded to tier1
            routeMsg->setContextPointer(this);

            std::vector<TransportAddress> sourceRoute;
            sourceRoute.push_back(apiMsg->getNextHopNode());
            sendToKey(apiMsg->getDestKey(), routeMsg, 1, sourceRoute);
        }

        delete msg;
    } else if (dynamic_cast<BaseRpcMessage*>(msg) != NULL) {
        BaseRpcMessage* rpcMessage =
            static_cast<BaseRpcMessage*>(msg);
        internalHandleRpcMessage(rpcMessage);
    }

    // process other messages from App
    else if (msg->arrivedOn("from_app")) {
        handleAppMessage(msg);

    } else {
        opp_error("BaseOverlay::handleMessage(): Received msg with "
                  "unknown type!");
        delete msg;
    }
}

void BaseOverlay::handleBaseOverlayMessage(BaseOverlayMessage* msg,
                                           const OverlayKey& destKey)
{
    switch (msg->getType()) {
    case OVERLAYSIGNALING:
        handleUDPMessage(msg);
        return;

    case RPC: {
        // process rpc-messages
        BaseRpcMessage* rpcMessage =
            check_and_cast<BaseRpcMessage*>(msg);
        internalHandleRpcMessage(rpcMessage);
        return;
    }

    case APPDATA: {
        //TODO use route messages? here: set transport type to ROUTE for "naked"
        // app messages
        OverlayCtrlInfo* overlayCtrlInfo = check_and_cast<OverlayCtrlInfo*>(msg->controlInfo());
        overlayCtrlInfo->setTransportType(ROUTE_TRANSPORT);

        // TODO: Should be removed if we used route messages here (definition of srcNode?)
        //if (overlayCtrlInfo->getSrcNode().isUnspecified()) {
        //    overlayCtrlInfo->setSrcNode(NodeHandle(overlayCtrlInfo->getLastHop()));
        //}

        BaseAppDataMessage* baseAppDataMsg =
            check_and_cast<BaseAppDataMessage*>(msg);
        callDeliver(baseAppDataMsg, destKey);
        return;
    }

    case OVERLAYROUTE: {
        BaseRouteMessage* baseRouteMsg =
            check_and_cast<BaseRouteMessage*>(msg);

        // if this node is malicious drop the message
        if (isMalicious()) {
            EV << "[BaseOverlay::handleBaseOverlayMessage() @ " << thisNode.ip
            << " (" << thisNode.key.toString(16) << ")]\n"
            << "    BaseRouteMessage gets dropped because this node is malicious"
            << endl;
            RECORD_STATS(numDropped++;
                         bytesDropped += baseRouteMsg->byteLength());
            delete msg;
            return;
        }

        // collect delay-value of completed hop
        if (collectPerHopDelay) {
            baseRouteMsg->setHopDelayArraySize(baseRouteMsg->
                                               getHopDelayArraySize() + 1);
            baseRouteMsg->setHopDelay(baseRouteMsg->getHopDelayArraySize() - 1,
                                      simTime() - baseRouteMsg->getHopStamp());
        }

        OverlayCtrlInfo* overlayCtrlInfo
            = check_and_cast<OverlayCtrlInfo*>(baseRouteMsg
                                               ->removeControlInfo());
        // set transport type
        overlayCtrlInfo->setTransportType(ROUTE_TRANSPORT);

        // source routing: save visited nodes, copy next hops
        std::vector<TransportAddress> sourceRoute;
        if ((baseRouteMsg->getNextHopsArraySize() > 0) ||
             (baseRouteMsg->getRoutingType() == RECURSIVE_SOURCE_ROUTING)) {
            // save this node as visited
            baseRouteMsg
                ->setVisitedHopsArraySize(baseRouteMsg
                                          ->getVisitedHopsArraySize() + 1);
            baseRouteMsg->setVisitedHops(baseRouteMsg
                                             ->getVisitedHopsArraySize() - 1,
                                         thisNode);

            // remove nodes from next hops and copy them to sourceRoute
            if (baseRouteMsg->getNextHopsArraySize() > 0) {
                sourceRoute.resize(baseRouteMsg->getNextHopsArraySize()- 1);
                for (uint i = 1; i < baseRouteMsg->getNextHopsArraySize();
                     ++i) {
                    sourceRoute[i - 1] = baseRouteMsg->getNextHops(i);
                }
                baseRouteMsg->setNextHopsArraySize(0);
            }
        }

        // decapsulate msg if node is sibling for destKey
        // or message is at its destination node
        bool err;
        if ((sourceRoute.size() == 0) &&
            (baseRouteMsg->getDestKey().isUnspecified() ||
             isSiblingFor(thisNode, baseRouteMsg->getDestKey(), 1, &err)
             /*&& !err*/)) {
            overlayCtrlInfo->setHopCount(baseRouteMsg->getHopCount());
            overlayCtrlInfo->setSrcNode(baseRouteMsg->getSrcNode());
            overlayCtrlInfo->setRoutingType(baseRouteMsg->getRoutingType());

            // copy visited nodes to control info
            overlayCtrlInfo
                ->setVisitedHopsArraySize(baseRouteMsg
                                              ->getVisitedHopsArraySize());
            for (uint i = 0; i < baseRouteMsg->getVisitedHopsArraySize(); ++i)
                overlayCtrlInfo
                    ->setVisitedHops(i, baseRouteMsg->getVisitedHops(i));

            BaseOverlayMessage* tmpMsg
                = check_and_cast<BaseOverlayMessage*>(baseRouteMsg
                                                      ->decapsulate());
            tmpMsg->setControlInfo(overlayCtrlInfo);

            // delay between hops
            if (collectPerHopDelay) {
                RECORD_STATS(
                    size_t i;
                    for (i = singleHopDelays.size();
                             i < baseRouteMsg->getHopDelayArraySize();) {
                        singleHopDelays.push_back(new HopDelayRecord[++i]);
                    }

                    i = baseRouteMsg->getHopDelayArraySize() - 1;
                    HopDelayRecord* hdr = singleHopDelays[i];

                    for (size_t j = 0; j <= i; ++j) {
                        hdr[j].count++;
                        hdr[j].val += baseRouteMsg->getHopDelay(j);
                    }
                );
            }

            // handle encapsulated message at destination node
            if (((baseRouteMsg->getRoutingType() == ITERATIVE_ROUTING)
                    || (baseRouteMsg->getRoutingType() == EXHAUSTIVE_ITERATIVE_ROUTING))
                    || forwardMessageRecursive(thisNode, baseRouteMsg)) {
                handleBaseOverlayMessage(tmpMsg, baseRouteMsg->getDestKey());
                delete baseRouteMsg;
            }
            return;
        } else {
            // forward msg if this node is not responsible for the key
            baseRouteMsg->setControlInfo(overlayCtrlInfo);
            sendToKey(baseRouteMsg->getDestKey(), baseRouteMsg, 1, sourceRoute);
            return;
        }
        break;
    }

    default:
        EV << "[BaseOverlay::handleBaseOverlayMessage() @ " << thisNode.ip
        << " (" << thisNode.key.toString(16) << ")]\n"
        << "    Received unknown message from UDP of type " << msg->name()
        << endl;
        break;
    }
}

void BaseOverlay::receiveChangeNotification(int category, cPolymorphic* details)
{
    Enter_Method_Silent();
    if (category == NF_OVERLAY_TRANSPORTADDRESS_CHANGED) {
        handleTransportAddressChangedNotification();
    } else if (category == NF_OVERLAY_NODE_LEAVE) {
        handleNodeLeaveNotification();
    } else if (category == NF_OVERLAY_NODE_GRACEFUL_LEAVE) {
        handleNodeGracefulLeaveNotification();
    }
}

void BaseOverlay::handleTransportAddressChangedNotification()
{
    // get new ip address
    thisNode.ip = IPAddressResolver().addressOf(
                      parentModule()->parentModule()).get4();

    joinOverlay();
}

void BaseOverlay::handleNodeLeaveNotification()
{
    // ...
}

void BaseOverlay::handleNodeGracefulLeaveNotification()
{
    // ...
}

//virtual protected
void BaseOverlay::handleTimerEvent(cMessage* msg)
{
    // timer...
}

//virtual protected
void BaseOverlay::handleAppMessage(cMessage* msg)
{
    delete msg;
}

void BaseOverlay::handleUDPMessage(BaseOverlayMessage* msg)
{
    delete msg;
}

//virtual protected
void BaseOverlay::recordOverlaySentStats(BaseOverlayMessage* msg)
{
    // collect statistics ...
}

void BaseOverlay::setOverlayReady(bool ready)
{
    //TODO new setOverlayState(State state) function
    //if (ready && state == READY) return;

    displayString().setTagArg("i", 1, ready ? "" : "red");
    bootstrapOracle->setOverlayReadyIcon(getThisNode(), ready);

    if (ready) {
        bootstrapList->registerBootstrapNode(thisNode);
    } else {
        bootstrapList->removeBootstrapNode(thisNode);
    }

    if (globalParameters->getPrintStateToStdOut()) {
        std::cout << "OVERLAY STATE: " << (ready ? "READY (" : "OFFLINE (")
                  << thisNode << ")" << std::endl;
    }

    CompReadyMessage* msg = new CompReadyMessage;
    msg->setReady(ready);
    msg->setComp(OVERLAY_COMP);
    // TODO: fix MMOG protocols to handle new CompReadyMessage
    //sendMessageToAllComp(msg, OVERLAY_COMP);
    sendDirect((cMessage*)msg->dup(), 0, getCompRpcGate(TIER1_COMP));
    sendDirect(msg, 0, getCompRpcGate(BOOTSTRAPLIST_COMP));
}



//------------------------------------------------------------------------
//--- Messages -----------------------------------------------------------
//------------------------------------------------------------------------

void BaseOverlay::sendRouteMessage(const TransportAddress& dest,
                                   BaseRouteMessage* msg,
                                   bool ack)
{
    if (msg && (dest != thisNode)) {
        msg->setHopCount(msg->getHopCount() + 1);
    }
    if (!ack)
        sendMessageToUDP(dest, msg);
    else {
        NextHopCall* nextHopCall = new NextHopCall(msg->name());
        nextHopCall->setLength(NEXTHOPCALL_L(nextHopCall));
        nextHopCall->encapsulate(msg);
        nextHopCall->setStatType(msg->getStatType());
        sendUdpRpcCall(dest, nextHopCall);
    }
}
void BaseOverlay::sendMessageToUDP(const TransportAddress& dest,
                                   cMessage* msg)
{
    // if there's still a control info attached to the message, remove it
    cPolymorphic* ctrlInfo = msg->removeControlInfo();
    if (ctrlInfo != NULL)
        delete ctrlInfo;

    // debug message
    if (debugOutput) {
        EV << "[BaseOverlay::sendMessageToUDP() @ " << thisNode.ip
        << " (" << thisNode.key.toString(16) << ")]\n"
        << "    Sending " << *msg << " to " << dest.ip
        << endl;
    }

    msg->setKind(UDP_C_DATA);
    UDPControlInfo* udpControlInfo = new UDPControlInfo();
    udpControlInfo->setSrcAddr(thisNode.ip);
    udpControlInfo->setSrcPort(thisNode.port);
    udpControlInfo->setDestAddr(dest.ip);
    udpControlInfo->setDestPort(dest.port);
    msg->setControlInfo(udpControlInfo);

    send(msg, "to_udp");

    if (dest != thisNode) {
        BaseOverlayMessage* baseOverlayMsg
            = check_and_cast<BaseOverlayMessage*>(msg);
        // record statistics, if message is not local
        if (baseOverlayMsg->getStatType() == APP_DATA_STAT) {
            RECORD_STATS(numAppDataSent++;
                         bytesAppDataSent += msg->byteLength());
        } else if (baseOverlayMsg->getStatType() == APP_LOOKUP_STAT){
            RECORD_STATS(numAppLookupSent++; bytesAppLookupSent +=
                         msg->byteLength());
        } else { // MAINTENANCE_STAT
            RECORD_STATS(numMaintenanceSent++; bytesMaintenanceSent +=
                         msg->byteLength());
        }
        recordOverlaySentStats(baseOverlayMsg);
    } else {
        if (dest.ip == thisNode.ip) { // to same node, but different port
            RECORD_STATS(numInternalSent++; bytesInternalSent += msg->byteLength());
        }
    }
}

//------------------------------------------------------------------------
//--- Basic Routing ------------------------------------------------------
//------------------------------------------------------------------------

static int pendingLookups = 0;

void BaseOverlay::initLookups()
{
    lookups = LookupSet();
}

void BaseOverlay::finishLookups()
{
    while (lookups.size() > 0) {
        (*lookups.begin())->abortLookup();
    }
    lookups.clear();
}

class SendToKeyListener : public LookupListener
{
private:
    BaseOverlay* overlay;
    BaseOverlayMessage* msg;
    GlobalStatistics* globalStatistics;
public:
    SendToKeyListener( BaseOverlay* overlay, BaseOverlayMessage* msg ) {
        this->overlay = overlay;
        this->msg = msg;
        globalStatistics = overlay->globalStatistics;
        pendingLookups++;
    }

    ~SendToKeyListener() {
        pendingLookups--;
        overlay = NULL;
        if (msg != NULL) {
            delete msg;
            msg = NULL;
        }
    }

    virtual void lookupFinished(AbstractLookup *lookup) {
        if (dynamic_cast<BaseRouteMessage*>(msg)) {
            BaseRouteMessage* routeMsg = static_cast<BaseRouteMessage*>(msg);
            if (lookup->isValid()) {
                if (lookup->getResult().size()==0) {
                    EV << "[SendToKeyListener::lookupFinished()]\n"
                          "    [ERROR] SendToKeyListener: Valid result, "
                          "but empty array." << endl;
                } else {
                    routeMsg->setHopCount(routeMsg->getHopCount()
                                          + lookup->getAccumulatedHops());

                    for (uint i=0; i<lookup->getResult().size(); i++) {
                        overlay->sendRouteMessage(lookup->getResult()[i],
                                                  static_cast<BaseRouteMessage*>
                                                   (routeMsg->dup()),
                                                  overlay->routeMsgAcks);
                    }
                }
            } else {
                EV << "[SendToKeyListener::lookupFinished()]\n"
                   << "    Lookup failed - dropping message"
                   << endl;
                RECORD_STATS(overlay->numDropped++;
                             overlay->bytesDropped += routeMsg->byteLength());
            }
        } else if (dynamic_cast<LookupCall*>(msg)) {
            LookupCall* call = static_cast<LookupCall*>(msg);
            LookupResponse* response = new LookupResponse();

            if (lookup->isValid()) {

                response->setIsValid(true);
                response->setKey(call->getKey());
                response->setSiblingsArraySize(lookup->getResult().size());
                for (uint i=0; i<lookup->getResult().size(); i++) {
                    response->setSiblings(i, lookup->getResult()[i]);
                }
                if (lookup->getResult().size() == 0) {
                    EV << "[SendToKeyListener::lookupFinished() @ "
                       << overlay->thisNode.ip
                       << " (" << overlay->thisNode.key.toString(16) << ")]\n"
                       << "    LookupCall "
                       << call->getNonce()
                       << " failed! (size=0)" << endl;
                }
            } else {
                response->setIsValid(false);
                EV << "[SendToKeyListener::lookupFinished() @ "
                   << overlay->thisNode.ip
                   << " (" << overlay->thisNode.key.toString(16) << ")]\n"
                   << "    LookupCall "
                   << call->getNonce()
                   << " failed!" << endl;
            }
            overlay->sendRpcResponse(call, response);
            msg = NULL;
        } else {
            throw new cRuntimeError("SendToKeyListener::lookupFinished(): "
                                    "Unkown message type!");
        }
        delete this;
    }
};

void BaseOverlay::route(const OverlayKey& key, CompType destComp,
                        CompType srcComp, cMessage* msg,
                        const std::vector<TransportAddress>& sourceRoute,
                        RoutingType routingType)
{
    if (key.isUnspecified() &&
        (!sourceRoute.size() || sourceRoute[0].isUnspecified()))
        error("route(): Key and hint unspecified!");

    // create base route message
    BaseAppDataMessage* baseAppDataMsg =
        new BaseAppDataMessage("BaseAppDataMessage");
    baseAppDataMsg->setType(APPDATA);
    baseAppDataMsg->setDestComp(destComp);
    baseAppDataMsg->setSrcComp(srcComp);
    baseAppDataMsg->setLength(BASEAPPDATA_L(baseAppDataMsg));
    baseAppDataMsg->setName(msg->name());

    baseAppDataMsg->setStatType(APP_DATA_STAT);
    baseAppDataMsg->encapsulate(msg);

    // debug output
    if (debugOutput) {
        EV << "[BaseOverlay::route() @ " << thisNode.ip
        << " (" << thisNode.key.toString(16) << ")]\n"
        << "    Received message from application"
        << endl;
    }

    if (key.isUnspecified()) {
        sendMessageToUDP(sourceRoute[0], baseAppDataMsg);
    } else {
        sendToKey(key, baseAppDataMsg, 1, sourceRoute, routingType);
    }
}

bool BaseOverlay::forwardMessageRecursive(const TransportAddress& dest,
                                          BaseRouteMessage* msg)
{
    return true;
}

void BaseOverlay::sendToKey(const OverlayKey& key, BaseOverlayMessage* msg,
                            int numSiblings,
                            const std::vector<TransportAddress>& sourceRoute,
                            RoutingType routingType)
{
    BaseRouteMessage* routeMsg = NULL;

    if (routingType == DEFAULT_ROUTING) routingType = defaultRoutingType;

    if (debugOutput) {
        EV << "[BaseOverlay::sendToKey() @ " << thisNode.ip
        << " (" << thisNode.key.toString(16) << ")]\n"
        << "    Sending " << msg <<  " to " << key
        << endl;
    }

    if (key.isUnspecified() &&
        !(sourceRoute.size() && !sourceRoute[0].isUnspecified()))
        error("BaseOverlay::sendToKey(): unspecified destination key!");

    if (msg->getType() != OVERLAYROUTE) {
        assert(!msg->controlInfo());
        routeMsg = new BaseRouteMessage("BaseRouteMessage");
        routeMsg->setType(OVERLAYROUTE);
        routeMsg->setRoutingType(routingType);
        routeMsg->setDestKey(key);
        routeMsg->setSrcNode(thisNode);
        routeMsg->setStatType(msg->getStatType());
        // copy the name of the inner message
        routeMsg->setName(msg->name());
        routeMsg->setLength(BASEROUTE_L(routeMsg));
        routeMsg->encapsulate(msg);

        OverlayCtrlInfo* routeCtrlInfo = new OverlayCtrlInfo;
        routeCtrlInfo->setLastHop(thisNode);
        routeCtrlInfo->setTransportType(ROUTE_TRANSPORT);
        routeCtrlInfo->setRoutingType(routingType);
        routeMsg->setControlInfo(routeCtrlInfo);

        //message marked as not already forwarded to tier1
        routeMsg->setContextPointer(NULL);
    } else {
        routeMsg = check_and_cast<BaseRouteMessage*>(msg);
        routingType = static_cast<RoutingType>(routeMsg->getRoutingType());
    }

    /* // only used as workaround - uncomment if assert fails!
    if (!routeMsg->controlInfo()) {
        // workaround
        OverlayCtrlInfo* routeCtrlInfo = new OverlayCtrlInfo;
        routeCtrlInfo->setLastHopAddr(thisNode.ip);
        routeCtrlInfo->setLastHopPort(thisNode.port);
        routeCtrlInfo->setTransportType(ROUTE_TRANSPORT);
        routeMsg->setControlInfo(routeCtrlInfo);
    }
    */

    // set timestamp for next hop
    if (collectPerHopDelay) {
        routeMsg->setHopStamp(simTime());
    }

    if (sourceRoute.size() && !sourceRoute[0].isUnspecified()) {
        // send msg to nextHop if specified (used for e.g. join rpcs)
        OverlayCtrlInfo* ctrlInfo = check_and_cast<OverlayCtrlInfo*>
            (routeMsg->controlInfo());
        ctrlInfo->setTransportType(UDP_TRANSPORT);
        assert(routeMsg->getNextHopsArraySize() == 0);
        routeMsg->setNextHopsArraySize(sourceRoute.size());
        for (uint i = 0; i < sourceRoute.size(); ++i)
            routeMsg->setNextHops(i, sourceRoute[i]);
        sendRouteMessage(sourceRoute[0], routeMsg, routeMsgAcks);
        return;
    }

    if ((routingType == ITERATIVE_ROUTING)
        || (routingType == EXHAUSTIVE_ITERATIVE_ROUTING)) {

        // create lookup and sent to key
        AbstractLookup* lookup = createLookup(routingType, routeMsg, NULL, true);
        lookup->lookup(routeMsg->getDestKey(), numSiblings, hopCountMax,
                       0, new SendToKeyListener(this, routeMsg));
    } else  {
        // recursive routing
        NodeVector* nextHops = findNode(routeMsg->getDestKey(),
                                        recNumRedundantNodes,
                                        numSiblings, routeMsg);

        if (nextHops->size() == 0) {
            EV << "[BaseOverlay::sendToKey() @ " << thisNode.ip
               << " (" << thisNode.key.toString(16) << ")]\n"
               << "    FindNode() returned NULL - dropping message"
               << endl;
            //std::cout << simTime() << " " << thisNode.ip << " " << state
            //          << " FindNode() returned NULL - dropping message "
            //          << routeMsg->name() << " from "
            //          << routeMsg->getSrcNode() << std::endl;
            // statistics
            RECORD_STATS(numDropped++; bytesDropped += routeMsg->byteLength());
            delete routeMsg;
        } else {
            // delete message if the hop count maximum is exceeded
            if (routeMsg->getHopCount() >= hopCountMax) {

                EV << "[BaseOverlay::sendToKey() @ " << thisNode.ip
                   << " (" << thisNode.key.toString(16) << ")]\n"
                   << "    Discards " << routeMsg->name() << " from "
                   << routeMsg->getSrcNode().ip << "\n"
                   << "    The hop count maximum has been exceeded ("
                   << routeMsg->getHopCount() << ">="
                   << hopCountMax << ")"
                   << endl;
                // statistics
                RECORD_STATS(numDropped++;
                             bytesDropped += routeMsg->byteLength());
                delete routeMsg;
                delete nextHops;
                return;
            }

            OverlayCtrlInfo* overlayCtrlInfo =
                dynamic_cast<OverlayCtrlInfo*>(routeMsg->controlInfo());
            assert(overlayCtrlInfo);

            // check and choose nextHop candidate
            NodeHandle* nextHop = NULL;
            bool err, isSibling;
            isSibling = isSiblingFor(thisNode, routeMsg->getDestKey(),
                                     numSiblings, &err);

            for (uint index = 0; nextHop == NULL && nextHops->size() > index;
                 ++index) {
                nextHop = &((*nextHops)[index]);
                // loop detection
                if (((overlayCtrlInfo->getLastHop() == *nextHop) &&
                     (*nextHop != thisNode)) ||
                     // do not forward msg to source node
                    ((*nextHop == routeMsg->getSrcNode()) &&
                     (thisNode != routeMsg->getSrcNode())) ||
                     // nextHop is thisNode, but isSiblingFor() is false
                    ((*nextHop == thisNode) && (!isSibling))) {
                    nextHop = NULL;
                }
            }

            if (nextHop == NULL) {
                if (!checkFindNode(routeMsg)) {
                    EV << "[BaseOverlay::sendToKey() @ " << thisNode.ip
                       << " (" << thisNode.key.toString(16) << ")]\n"
                       << "    Discards " << routeMsg->name() << " from "
                       << routeMsg->getSrcNode().ip << "\n"
                       << "    No useful nextHop found!"
                       << endl;
                    //std::cout << thisNode.ip << " packet dropped" << std::endl;
                    RECORD_STATS(numDropped++;
                                 bytesDropped += routeMsg->byteLength());
                }
                delete routeMsg;
                delete nextHops;
                return;
            }

            assert(!nextHop->isUnspecified());

            // callForward to app
            if (useCommonAPIforward &&
                dynamic_cast<BaseAppDataMessage*>(
                        routeMsg->encapsulatedMsg()) &&
                routeMsg->contextPointer() == NULL) {
                callForward(routeMsg->getDestKey(), routeMsg, *nextHop);
                delete nextHops;
                return;
            }
            //message marked as not already forwarded
            routeMsg->setContextPointer(NULL);

            // is this node responsible?
            if (*nextHop == thisNode) {
                if (isSibling && !err) {
                    //EV << "[BaseOverlay::sendToKey() @ " << thisNode.ip
                    //   << " (" << thisNode.key.toString(16) << ")]\n"
                    //   << "    Forwards msg for key " << routeMsg->getDestKey() "\n"
                    //   << "    to node " << (*nextHops)[0]
                    //   << endl;
                    delete nextHops;
                    assert(routeMsg->controlInfo());
                    handleBaseOverlayMessage(routeMsg, key);
                    return;
                } else {
                    error("isSiblingsFor() is true with an error: "
                          "Erroneous method isSiblingFor()!");
                }
            } // else forward msg if this node is not responsible for the key

            // records statistics, if we forward this message
            if (overlayCtrlInfo->getLastHop().ip != thisNode.ip) {
                if (routeMsg->getStatType() == APP_DATA_STAT) {
                    RECORD_STATS(numAppDataForwarded++; bytesAppDataForwarded +=
                                 routeMsg->byteLength());
                } else if (routeMsg->getStatType() == APP_LOOKUP_STAT){
                    RECORD_STATS(numAppLookupForwarded++;
                                 bytesAppLookupForwarded +=
                                     routeMsg->byteLength());
                } else {
                    RECORD_STATS(numMaintenanceForwarded++;
                                 bytesMaintenanceForwarded +=
                                     routeMsg->byteLength());
                }
            }
            overlayCtrlInfo->setHopCount(routeMsg->getHopCount());
            if (forwardMessageRecursive(*nextHop, routeMsg)) {
                delete routeMsg->removeControlInfo();
                sendRouteMessage(*nextHop, routeMsg, routeMsgAcks);
            }
        }
        delete nextHops;
    }
}

bool BaseOverlay::checkFindNode(BaseRouteMessage* routeMsg)
{
    if (dynamic_cast<FindNodeCall*>(routeMsg->encapsulatedMsg())) {
        FindNodeCall* findNodeCall =
            static_cast<FindNodeCall*>(routeMsg->decapsulate());
        findNodeCall
            ->setControlInfo(check_and_cast<OverlayCtrlInfo*>
            (routeMsg->removeControlInfo()));
        findNodeRpc(findNodeCall);
        //std::cout << "+" << std::endl;
        return true;
    }
    //std::cout << "-" << std::endl;
    return false;
}

//protected: create a lookup class
AbstractLookup* BaseOverlay::createLookup(RoutingType routingType,
                                          const BaseOverlayMessage* msg,
                                          const cObject* findNodeExt,
                                          bool appLookup)
{
    AbstractLookup* newLookup;

    if (routingType == DEFAULT_ROUTING) {
        routingType = defaultRoutingType;
    }

    switch (routingType) {
        case ITERATIVE_ROUTING:
        case EXHAUSTIVE_ITERATIVE_ROUTING:
            newLookup = new IterativeLookup(this, routingType,
                                            iterativeLookupConfig, findNodeExt,
                                            appLookup);
            break;
        case RECURSIVE_SOURCE_ROUTING:
        case SEMI_RECURSIVE_ROUTING:
        case FULL_RECURSIVE_ROUTING:
            newLookup = new RecursiveLookup(this, routingType, appLookup);
            break;
        default:
            throw new cRuntimeError("BaseOverlay::createLookup():"
                                    " Unknown routingType!");
            break;
    }

    lookups.insert(newLookup);
    return newLookup;
}

void BaseOverlay::removeLookup(AbstractLookup* lookup)
{
    lookups.erase(lookup);
}

//virtual public
OverlayKey BaseOverlay::distance( const OverlayKey& x,
                                  const OverlayKey& y ) const
{
    opp_error("BaseOverlay::distance(): Not implemented!");
    return OverlayKey::UNSPECIFIED_KEY;
}

//protected: find closest nodes
NodeVector* BaseOverlay::findNode(const OverlayKey& key,
                                  int numRedundantNodes,
                                  int numSiblings,
                                  BaseOverlayMessage* msg)
{
    opp_error( "findNode: Not implemented!" );
    return NULL;
}

//protected: join the overlay with a given nodeID
void BaseOverlay::joinOverlay()
{
//  std::cout << "BaseOverlay::joinOverlay(): Not implemented!" << endl;
    return;
}

bool BaseOverlay::handleFailedNode(const TransportAddress& failed)
{
    return true;
}

//------------------------------------------------------------------------
//--- RPCs ---------------------------------------------------------------
//------------------------------------------------------------------------

//private
bool BaseOverlay::internalHandleRpc( BaseCallMessage* msg )
{
    // call rpc stubs
    RPC_SWITCH_START( msg );
    RPC_DELEGATE( FindNode, findNodeRpc );
    RPC_DELEGATE( FailedNode, failedNodeRpc );
    RPC_DELEGATE( Lookup, lookupRpc );
    RPC_DELEGATE( NextHop, nextHopRpc );
    RPC_SWITCH_END( );

    // if RPC was handled return true, else tell the parent class to handle it
    return RPC_HANDLED || BaseRpc::internalHandleRpc(msg);
}

bool BaseOverlay::internalHandleRpcResponse(BaseResponseMessage* msg,
                                            cPolymorphic* context,
                                            int rpcId, simtime_t rtt)
{
    // if RPC was handled return true, else tell the parent class to handle it
    //RPC_SWITCH_START( msg );
    //RPC_ON_RESPONSE( NextHop ) {
        //if (rtt > 1.0) std::cout << rtt  << " to "
        //                         << msg->getSrcNode() << std::endl;
    //}
    //RPC_SWITCH_END( );

    return /*RPC_HANDLED ||*/ BaseRpc::internalHandleRpcResponse(msg, context,
                                                             rpcId, rtt);
}

void BaseOverlay::internalHandleRpcTimeout(BaseCallMessage* msg,
                                           const TransportAddress& dest,
                                           cPolymorphic* context, int rpcId,
                                           const OverlayKey& destKey)
{
    RPC_SWITCH_START( msg )
        RPC_ON_CALL( NextHop )
        {
            //std::cout << "nextHopRpc timeout" << std::endl;
            BaseRouteMessage* tempMsg
                = check_and_cast<BaseRouteMessage*>(msg->decapsulate());

            if (!tempMsg->controlInfo()) {
                OverlayCtrlInfo* overlayCtrlInfo = new OverlayCtrlInfo;
                overlayCtrlInfo->setLastHop(thisNode);
                overlayCtrlInfo->setHopCount(tempMsg->getHopCount());
                overlayCtrlInfo->setSrcNode(tempMsg->getSrcNode());
                overlayCtrlInfo->setRoutingType(tempMsg->getRoutingType());//test
                overlayCtrlInfo->setTransportType(UDP_TRANSPORT);
                tempMsg->setControlInfo(overlayCtrlInfo);
            }
            // remove node from local routing tables
            // + send message again
            if (handleFailedNode(dest))
                handleBaseOverlayMessage(tempMsg, destKey);
            else {
                delete tempMsg;
                //std::cout << thisNode.ip << "   ***" << std::endl;
                RECORD_STATS(numDropped++;
                             bytesDropped += tempMsg->byteLength());
                joinOverlay();
            }
            break;
        }
    RPC_SWITCH_END( )

    BaseRpc::internalHandleRpcTimeout(msg, dest, context, rpcId, destKey);
}

void BaseOverlay::internalSendRouteRpc(BaseRpcMessage* message,
                                       const OverlayKey& destKey,
                                       const std::vector<TransportAddress>&
                                       sourceRoute,
                                       RoutingType routingType) {
    FindNodeCall* findNodeCall;
    uint numSiblings = 1;
    if ((findNodeCall = dynamic_cast<FindNodeCall*>(message)))
        numSiblings = findNodeCall->getNumSiblings();

    sendToKey(destKey, message, numSiblings, sourceRoute, routingType);
}

void BaseOverlay::internalSendRpcResponse(BaseCallMessage* call,
                                          BaseResponseMessage* response)
{
    OverlayCtrlInfo* overlayCtrlInfo =
        check_and_cast<OverlayCtrlInfo*>(call->controlInfo());

    TransportType transportType = ROUTE_TRANSPORT;
    const TransportAddress* destNode;
    if (overlayCtrlInfo->getSrcNode().isUnspecified())
        destNode = &(overlayCtrlInfo->getLastHop());
    else
        destNode = &(overlayCtrlInfo->getSrcNode());
    const OverlayKey* destKey = &OverlayKey::UNSPECIFIED_KEY;

    RoutingType routingType
        = static_cast<RoutingType>(overlayCtrlInfo->getRoutingType());

    assert(overlayCtrlInfo->getTransportType() != INTERNAL_TRANSPORT);

    if ((overlayCtrlInfo->getTransportType() == UDP_TRANSPORT) ||
        (routingType == SEMI_RECURSIVE_ROUTING) ||
        (routingType == ITERATIVE_ROUTING) ||
        (routingType == EXHAUSTIVE_ITERATIVE_ROUTING)) {
        // received by UDP or direct response (IR, EIR or SRR routing)
        transportType = UDP_TRANSPORT;
        overlayCtrlInfo->setVisitedHopsArraySize(0); //???
    } else if ((static_cast<RoutingType> (overlayCtrlInfo->getRoutingType())
            == FULL_RECURSIVE_ROUTING)) {
        // full recursive routing
        destKey = &(overlayCtrlInfo->getSrcNode().key);
        destNode = &NodeHandle::UNSPECIFIED_NODE;
    }
    // else: source routing -> route back over visited hops

    sendRpcResponse(transportType,
                    static_cast<CompType>(overlayCtrlInfo->getSrcComp()),
                    *destNode, *destKey, call, response);
}

//protected: statistic helpers for IterativeLookup
void BaseOverlay::countFindNodeCall( const FindNodeCall* call )
{
    RECORD_STATS(numFindNodeSent++;
                 bytesFindNodeSent += call->byteLength());
}

void BaseOverlay::countFailedNodeCall( const FailedNodeCall* call )
{
    RECORD_STATS(numFailedNodeSent++;
                 bytesFailedNodeSent += call->byteLength());
}

//private: rpc stub
void BaseOverlay::findNodeRpc( FindNodeCall* call )
{
    // if this node is malicious don't answer a findNodeCall
    if (isMalicious()) {
        EV << "[BaseOverlay::findNodeRpc() @ " << thisNode.ip
        << " (" << thisNode.key.toString(16) << ")]\n"
        << "    Node ignores findNodeCall because this node is malicious"
        << endl;
        delete call;
        return;
    }

    FindNodeResponse* findNodeResponse =
        new FindNodeResponse("FindNodeResponse");

    NodeVector* nextHops = findNode(call->getLookupKey(),
                                    call->getNumRedundantNodes(),
                                    call->getNumSiblings(), call);

    findNodeResponse->setClosestNodesArraySize(nextHops->size());

    for (uint i=0; i < nextHops->size(); i++) {
        findNodeResponse->setClosestNodes(i, (*nextHops)[i]);
    }

    bool err;
    if (isSiblingFor(thisNode, call->getLookupKey(), call->getNumSiblings(),
                     &err)) {
        findNodeResponse->setSiblings(true);
    }

    findNodeResponse->setLength(FINDNODERESPONSE_L(findNodeResponse));

    if (call->hasObject("findNodeExt")) {
        cMessage* findNodeExt = (cMessage*)call->removeObject("findNodeExt");
        findNodeResponse->addObject(findNodeExt);
        findNodeResponse->addLength(findNodeExt->length());
    }

    RECORD_STATS(numFindNodeResponseSent++; bytesFindNodeResponseSent +=
                     findNodeResponse->byteLength());

    delete nextHops;

    sendRpcResponse(call, findNodeResponse);
}

void BaseOverlay::failedNodeRpc( FailedNodeCall* call )
{
    FailedNodeResponse* failedNodeResponse =
        new FailedNodeResponse("FailedNodeResponse");
    failedNodeResponse->setTryAgain(handleFailedNode(call->getFailedNode()));
    failedNodeResponse->setLength(FAILEDNODERESPONSE_L(failedNodeResponse));

    if (call->hasObject("findNodeExt")) {
        cMessage* findNodeExt = check_and_cast<cMessage*>(
                                    call->removeObject("findNodeExt"));
        failedNodeResponse->addObject(findNodeExt);
        failedNodeResponse->addLength(findNodeExt->length());
    }

    RECORD_STATS(numFailedNodeResponseSent++; bytesFailedNodeResponseSent +=
                     failedNodeResponse->byteLength());

    sendRpcResponse(call, failedNodeResponse);
}

void BaseOverlay::lookupRpc(LookupCall* call)
{
    // create lookup and sent to key
    AbstractLookup* lookup = createLookup(DEFAULT_ROUTING, call, NULL, true);
    lookup->lookup(call->getKey(), call->getNumSiblings(), hopCountMax,
                   1, new SendToKeyListener( this, call ));

}

void BaseOverlay::nextHopRpc(NextHopCall* call)
{
    //test
    if (state != READY) {
        delete call;
        return;
    }

    BaseRouteMessage* routeMsg
        = check_and_cast<BaseRouteMessage*>(call->decapsulate());

    OverlayCtrlInfo* overlayCtrlInfo =
        check_and_cast<OverlayCtrlInfo*>(call->controlInfo()->dup());
    overlayCtrlInfo->setHopCount(routeMsg->getHopCount());
    overlayCtrlInfo->setSrcNode(routeMsg->getSrcNode());
    overlayCtrlInfo->setRoutingType(routeMsg->getRoutingType());

    routeMsg->setControlInfo(overlayCtrlInfo);
    assert(routeMsg->controlInfo());

    std::string temp("ACK: [");
    (temp += routeMsg->name()) += "]";

    NextHopResponse* response
        = new NextHopResponse(temp.c_str());
    response->setLength(NEXTHOPRESPONSE_L(response));
    sendRpcResponse(call, response);

    handleBaseOverlayMessage(routeMsg, routeMsg->getDestKey());
}

void BaseOverlay::registerComp(CompType compType, cModule *module)
{
    cGate *gate = NULL;

    if (module != NULL)
        gate = module->gate("direct_in");

    compModuleList[compType] = make_pair<cModule*, cGate*>(module, gate);
}

cModule* BaseOverlay::getCompModule(CompType compType)
{
    CompModuleList::iterator it = compModuleList.find(compType);

    if (it != compModuleList.end())
        return it->second.first;
    else
        return NULL;
}

cGate* BaseOverlay::getCompRpcGate(CompType compType)
{
    CompModuleList::iterator it = compModuleList.find(compType);

    if (it != compModuleList.end())
        return it->second.second;
    else
        return NULL;
}

void BaseOverlay::sendMessageToAllComp(cMessage* msg, CompType srcComp)
{
    for (CompModuleList::iterator it = compModuleList.begin();
         it != compModuleList.end(); it++) {

        // don't send message to the origination component
        if (it->first != srcComp)
            sendDirect((cMessage*)msg->dup(), 0, it->second.second);
    }

    delete msg;
}

bool BaseOverlay::isInSimpleMultiOverlayHost()
{
    return isVector() || parentModule()->isVector();
}

