//
// Copyright (C) 2006 Institut fuer Telematik, Universitaet Karlsruhe (TH)
//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public License
// as published by the Free Software Foundation; either version 2
// of the License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
//

/**
 * @file BaseRpc.cc
 * @author Bernhard Heep
 * @author Sebastian Mies
 * @author Gregoire Menuel
 */

#include <CommonMessages_m.h>
#include <BootstrapOracleAccess.h>
#include <UnderlayConfiguratorAccess.h>
#include <GlobalStatisticsAccess.h>
#include <GlobalParametersAccess.h>
#include <NeighborCache.h>
#include <OverlayAccess.h>
#include <assert.h>

#include "BaseRpc.h"
#include "RpcMacros.h"


//------------------------------------------------------------------------
//--- Initialization & finishing -----------------------------------------
//------------------------------------------------------------------------


//------------------------------------------------------------------------
//--- RPC Handling -------------------------------------------------------
//------------------------------------------------------------------------

BaseRpc::BaseRpc()
{
    defaultRpcListener = NULL;
}

//private
void BaseRpc::initRpcs()
{
    // set friend modules
    globalStatistics = GlobalStatisticsAccess().get();
    globalParameters = GlobalParametersAccess().get();

    rpcsPending = 0;
    rpcStates.clear();

    defaultRpcListener = new RpcListener();

    //set ping cache
    numPingSent = 0;
    bytesPingSent = 0;
    numPingResponseSent = 0;
    bytesPingResponseSent = 0;

    WATCH(numPingSent);
    WATCH(bytesPingSent);
    WATCH(numPingResponseSent);
    WATCH(bytesPingResponseSent);

    // set overlay pointer
    overlay = OverlayAccess().get(this);

    // register component
    thisCompType = getThisCompType();
    overlay->registerComp(thisCompType, this);

    cModule *mod = parentModule();
    neighborCache = NULL;
    while (neighborCache == NULL) {
        neighborCache = (NeighborCache*)mod->submodule("neighborCache");
        mod = mod->parentModule();
        if (!mod)
            opp_error ("BaseRpc::initRpc: Module type contains no ping cache!");
    }
}

//private
void BaseRpc::finishRpcs()
{
    // stop all rpcs
    for (RpcStates::iterator i = rpcStates.begin();
        i != rpcStates.end(); i++) {
        cancelAndDelete(i->second.callMsg);
        cancelAndDelete(i->second.timeoutMsg);
        delete i->second.dest;
        i->second.dest = NULL;
        delete i->second.context;
        i->second.context = NULL;
    }
    rpcStates.clear();

    // delete default rpc listener
    if (defaultRpcListener != NULL) {
        delete defaultRpcListener;
        defaultRpcListener = NULL;
    }
}

uint32_t BaseRpc::sendRpcCall(TransportType transportType,
                              CompType destComp,
                              const TransportAddress& dest, //TODO vector
                              //const std::vector<TransportAddress>& sourceRoute,
                              const OverlayKey& destKey,
                              BaseCallMessage* msg,
                              cPolymorphic* context,
                              RoutingType routingType,
                              simtime_t timeout,
                              int retries,
                              int rpcId,
                              RpcListener* rpcListener)
{
    // create nonce, timeout and set default parameters
    uint nonce;
    do {
        nonce = intuniform(1, 2147483647);
    } while (rpcStates.count(nonce) > 0);

    if (timeout == -1) {
        switch (transportType) {
        case INTERNAL_TRANSPORT:
            timeout = 0;
            break;
        case UDP_TRANSPORT:
            timeout = globalParameters->getRpcUdpTimeout();
            break;
        case ROUTE_TRANSPORT:
            timeout = (destKey.isUnspecified() ?
                       globalParameters->getRpcUdpTimeout() :
                       globalParameters->getRpcKeyTimeout());
            break;
        default:
            throw new cException("BaseRpc::sendRpcMessage(): "
                                 "Unknown RpcTransportType!");
        }
    }

    if (rpcListener == NULL)
        rpcListener = defaultRpcListener;

    // create state
    RpcState state;
    state.id = rpcId;
    state.timeSent = simulation.simTime();
    state.dest = dest.dup();
    state.destKey = destKey;
    state.srcComp = getThisCompType();
    state.destComp = destComp;
    state.listener = rpcListener;
    state.timeoutMsg = new RpcTimeoutMessage();
    state.timeoutMsg->setNonce(nonce);
    state.retries = retries;
    state.transportType = transportType;
    state.routingType = routingType;
    state.context = context;

    if (rpcStates.count(nonce) > 0)
        error("RPC nonce collision");

    // set message parameters
    msg->setNonce(nonce);
    if (transportType == ROUTE_TRANSPORT)
        msg->setSrcNode(overlay->getThisNode());
    else
        msg->setSrcNode(thisNode);
    msg->setType(RPC);

    // save copy of call message in RpcState
    state.callMsg = dynamic_cast<BaseCallMessage*>(msg->dup());
    assert(!msg->encapsulatedMsg() || !msg->encapsulatedMsg()->controlInfo());

    // register state
    rpcStates[nonce] = state;

    // schedule timeout message
    if (timeout!=0)
        scheduleAt( simulation.simTime() + timeout, state.timeoutMsg);

    //TODO
    std::vector<TransportAddress> sourceRoute;
    sourceRoute.push_back(dest);
    sendRpcMessageWithTransport(transportType, destComp, routingType,
                                sourceRoute, destKey, msg);

    return nonce;
}


//public
void BaseRpc::cancelRpcMessage(uint32_t nonce)
{
    if (rpcStates.count(nonce)==0)
        return;
    RpcState state = rpcStates[nonce];
    rpcStates.erase(nonce);
    cancelAndDelete(state.callMsg);
    cancelAndDelete(state.timeoutMsg);
    delete state.dest;
    state.dest = NULL;
    delete state.context;
    state.context = NULL;
}

//protected
void BaseRpc::internalHandleRpcMessage(BaseRpcMessage* msg)
{
    // check if this is a rpc call message
    BaseCallMessage* rpCall = dynamic_cast<BaseCallMessage*>(msg);
    if (rpCall != NULL) {
        bool rpcHandled = true;
        if (!handleRpc(rpCall)) rpcHandled = internalHandleRpc(rpCall);
        if (!rpcHandled) {
            EV << "[BaseRpc::internalHandleRpcMessage() @ " << thisNode.ip
               << " (" << thisNode.key.toString(16) << ")]\n"
               << "    Error: RPC '" << msg->fullName()<< "' was not handled"
               << endl;
            delete msg;
        }
        return;
    }

    // get nonce
    int nonce = msg->getNonce();

    // nonce known? no -> delete message and return
    if (rpcStates.count(nonce)==0) {
        EV << "[BaseRpc::internalHandleRpcMessage() @ " << thisNode.ip
           << " " << thisNode.key.toString(16) << ")]\n"
           << "    RPC: Nonce Unknown"
           << endl;
        delete msg;
        return;
    }

    // get state and remove from map
    RpcState state = rpcStates[nonce];
    rpcStates.erase(nonce);

    // is timeout message?
    if (msg->isSelfMessage() &&
        (dynamic_cast<RpcTimeoutMessage*>(msg) != NULL)) {
        // yes-> inform listener

        // retry?
        state.retries--;
        simtime_t timeout = simulation.simTime() - state.timeSent;
        if (state.retries>=0) {
            //TODO
            std::vector<TransportAddress> sourceRoute;
            sourceRoute.push_back(*state.dest);
            sendRpcMessageWithTransport(state.transportType, state.destComp,
                                        state.routingType,
                                        sourceRoute,
                                        state.destKey,
                                        dynamic_cast<BaseCallMessage*>
                                        (state.callMsg->dup()));
            if (timeout!=0)
                scheduleAt(simulation.simTime() + timeout, msg);

            state.timeSent = simulation.simTime();
            rpcStates[nonce] = state;
            return;
        }
        // inform listener
        if (state.listener != NULL)
            state.listener->handleRpcTimeout(state.callMsg, *state.dest,
                                             state.context, state.id,
                                             state.destKey);

        // inform overlay
        internalHandleRpcTimeout(state.callMsg, *state.dest, state.context,
                                 state.id, state.destKey);
        handleRpcTimeout(state.callMsg, *state.dest, state.context,
                         state.id, state.destKey);

    } else { // no-> handle rpc response

        // get parameters
        simtime_t rtt = simulation.simTime() - state.timeSent;
        BaseResponseMessage* response
            = dynamic_cast<BaseResponseMessage*>(msg);

        //if (state.transportType == UDP_TRANSPORT)
        //    globalStatistics->recordOutVector("BaseRpc: UDP Round Trip Time",
        //                                      rtt);

        // inform listener
        if (state.listener != NULL)
            state.listener->handleRpcResponse(response, state.context,
                                              state.id, rtt);

        // inform overlay
        internalHandleRpcResponse(response, state.context, state.id, rtt);
        handleRpcResponse(response, state.context, state.id, rtt);

        // delete response
        delete response->removeControlInfo();
        delete response;
    }

    // delete messages
    delete state.callMsg;
    cancelAndDelete(state.timeoutMsg);
    delete state.dest;

    // clean up pointers
    state.dest = NULL;
    state.context = NULL;
    state.callMsg = NULL;
    state.timeoutMsg = NULL;
}

//private
bool BaseRpc::internalHandleRpc( BaseCallMessage* msg )
{
    RPC_SWITCH_START( msg );
    RPC_DELEGATE( Ping, pingRpc );
    RPC_SWITCH_END( );

    return RPC_HANDLED;
}

bool BaseRpc::internalHandleRpcResponse(BaseResponseMessage* msg,
                                        cPolymorphic* context,
                                        int rpcId, simtime_t rtt)
{
    // call rpc stubs
    RPC_SWITCH_START( msg );
    RPC_ON_RESPONSE( Ping ) {
        pingRpcResponse(_PingResponse, context, rpcId, rtt);
    }
    RPC_SWITCH_END( );

    return RPC_HANDLED; //TODO always false
}

void BaseRpc::internalHandleRpcTimeout(BaseCallMessage* msg,
                                       const TransportAddress& dest,
                                       cPolymorphic* context,
                                       int rpcId, const OverlayKey& destKey)
{
    RPC_SWITCH_START( msg ) {
        RPC_ON_CALL( Ping ) {
            pingRpcTimeout(_PingCall, dest, context, rpcId);
        }
    }
    RPC_SWITCH_END( )
}

//virtual protected
bool BaseRpc::handleRpc( BaseCallMessage* msg )
{
    return false;
}

void BaseRpc::sendRpcResponse(TransportType transportType,
                              CompType compType,
                              const TransportAddress &dest,
                              const OverlayKey &destKey,
                              BaseCallMessage* call,
                              BaseResponseMessage* response)
{
    if (call == NULL || response == NULL) {
        throw new cRuntimeError("call or response = NULL!");
    }

    assert(transportType == INTERNAL_TRANSPORT ||
           !dest.isUnspecified() ||
           !destKey.isUnspecified());

    if (transportType == ROUTE_TRANSPORT)
        response->setSrcNode(overlay->getThisNode());
    else
        response->setSrcNode(thisNode);
    response->setType(RPC);
    response->setNonce(call->getNonce());
    response->setStatType(call->getStatType());

    RoutingType routingType = NO_OVERLAY_ROUTING;
    OverlayCtrlInfo* overlayCtrlInfo = NULL;
    if (dynamic_cast<OverlayCtrlInfo*>(call->controlInfo()))
        overlayCtrlInfo =
            static_cast<OverlayCtrlInfo*>(call->removeControlInfo());
    else
        delete call->removeControlInfo();

    // source routing
    std::vector<TransportAddress> sourceRoute;
    if (overlayCtrlInfo && transportType == ROUTE_TRANSPORT) {
        routingType =
            static_cast<RoutingType>(overlayCtrlInfo->getRoutingType());
        for (uint i = overlayCtrlInfo->getVisitedHopsArraySize(); i > 0; --i) {
            if (overlayCtrlInfo->getVisitedHops(i - 1) !=
                    overlay->getThisNode()) {
                sourceRoute.push_back(overlayCtrlInfo->getVisitedHops(i - 1));
            }
        }
    }
    sourceRoute.push_back(dest);

    sendRpcMessageWithTransport(transportType, compType,
                                routingType, sourceRoute,
                                destKey, response);
    delete overlayCtrlInfo;
    delete call;
}

//protected
void BaseRpc::sendRpcResponse(BaseCallMessage* call,
                              BaseResponseMessage* response)
{
    const TransportAddress* destNode = &(call->getSrcNode());
    const OverlayKey* destKey = &(call->getSrcNode().key);

    OverlayCtrlInfo* overlayCtrlInfo =
        dynamic_cast<OverlayCtrlInfo*>(call->controlInfo());

    // "magic" transportType selection: internal
    if (overlayCtrlInfo &&
        overlayCtrlInfo->getTransportType() == INTERNAL_TRANSPORT) {
        sendRpcResponse(INTERNAL_TRANSPORT,
                        static_cast<CompType>(overlayCtrlInfo->getSrcComp()),
                        *destNode, *destKey, call, response);
    } else {
        internalSendRpcResponse(call, response);
    }
}

void BaseRpc::sendRpcMessageWithTransport(TransportType transportType,
                                          CompType destComp,
                                          RoutingType routingType,
                                          const std::vector<TransportAddress>& sourceRoute,
                                          const OverlayKey& destKey,
                                          BaseRpcMessage* message)
{
    switch (transportType) {
    case UDP_TRANSPORT: {
        sendMessageToUDP(sourceRoute[0], message);
        break;
    }
    case ROUTE_TRANSPORT: {
        internalSendRouteRpc(message, destKey,
                             sourceRoute, routingType);
        break;
    }
    case INTERNAL_TRANSPORT: {
        cGate *destCompGate = overlay->getCompRpcGate(destComp);
        if (destCompGate == NULL) {
            opp_error("BaseRpc::sendRpcMessageWithTransport():"
                      " INTERNAL_RPC to unknown RpcCompType!");
        }
        OverlayCtrlInfo *overlayCtrlInfo = new OverlayCtrlInfo();
        overlayCtrlInfo->setSrcComp(getThisCompType());
        overlayCtrlInfo->setDestComp(destComp);
        overlayCtrlInfo->setTransportType(INTERNAL_TRANSPORT);
        message->setControlInfo(overlayCtrlInfo);
        sendDirect(message, 0, destCompGate);
        break;
    }
    default:
        opp_error("BaseRpc::sendRpcMessageWithTransport: "
                  "invalid transportType!");
        break;
    }
}

// ping RPC stuff
void BaseRpc::pingResponse(PingResponse* response, cPolymorphic* context,
                           int rpcId, simtime_t rtt)
{
}

void BaseRpc::pingTimeout(PingCall* call, const TransportAddress& dest,
                          cPolymorphic* context, int rpcId)
{
}

void BaseRpc::pingRpc( PingCall* call )
{
    std::string pongName(call->name());
    if (pongName == "PING")
        pongName = "PONG";
    else {
        pongName = "PONG: [ ";
        pongName += call->name();
        pongName += " ]";
    }

    PingResponse* response = new PingResponse(pongName.c_str());
    response->setLength( PINGRESPONSE_L(response) );
    RECORD_STATS(numPingResponseSent++; bytesPingResponseSent +=
        response->byteLength());

    sendRpcResponse(call, response );
}

void BaseRpc::pingRpcResponse(PingResponse* response,
                              cPolymorphic* context, int rpcId, simtime_t rtt)
{
    NodeHandle& srcNode = response->getSrcNode();
    neighborCache->insertAddressRtt(srcNode, rtt);
    uint numContexts = neighborCache->numInhibitedPings(srcNode);
    WaitingContexts* waitingContexts =
        neighborCache->getWaitingContexts(srcNode);

    pingResponse(response, context, rpcId, rtt);

    if (numContexts > 0) {
        for (uint i = 0; i < waitingContexts->pingListeners.size(); ++i) {
            PingResponse *pr = new PingResponse();
            pr->setNonce(0);
            if (neighborCache->getNodeHandle(srcNode).isUnspecified())
                pr->setSrcNode(srcNode);
            else
                pr->setSrcNode(neighborCache->getNodeHandle(srcNode));

            // inform listener
            if (waitingContexts->pingListeners[i] != NULL)
                waitingContexts->pingListeners[i]
                    ->handleRpcResponse(pr, waitingContexts->pingContexts[i],
                                        waitingContexts->pingIds[i], rtt);

            handleRpcResponse(pr, waitingContexts->pingContexts[i],
                              waitingContexts->pingIds[i], rtt);
            pingResponse(pr, waitingContexts->pingContexts[i],
                         waitingContexts->pingIds[i], rtt);

            delete pr;
        }
        //neighborCache->clearNodeContexts(srcNode);
    }
    delete waitingContexts;
}

void BaseRpc::pingRpcTimeout(PingCall* pingCall,
                             const TransportAddress& dest,
                             cPolymorphic* context,
                             int rpcId)
{
    neighborCache->addressTimeout(dest);
    uint numContexts = neighborCache->numInhibitedPings(dest);
    WaitingContexts* waitingContexts =
        neighborCache->getWaitingContexts(dest);

    pingTimeout(pingCall, dest, context, rpcId);

    if (numContexts > 0) {
        for (uint i = 0; i < waitingContexts->pingListeners.size(); ++i) {
            PingCall* copy = static_cast<PingCall*>(pingCall->dup());
            copy->setNonce(0);

            // inform listener
            if (waitingContexts->pingListeners[i] != NULL)
                waitingContexts->pingListeners[i]
                    ->handleRpcTimeout(copy, dest,
                                       waitingContexts->pingContexts[i],
                                       waitingContexts->pingIds[i],
                                       OverlayKey::UNSPECIFIED_KEY);

            handleRpcTimeout(copy, dest, waitingContexts->pingContexts[i],
                             waitingContexts->pingIds[i],
                             OverlayKey::UNSPECIFIED_KEY);
            pingTimeout(copy, dest, waitingContexts->pingContexts[i],
                        waitingContexts->pingIds[i]);

            delete copy;
        }
        //neighborCache->addressTimeout(dest);
    }
    delete waitingContexts;
    //neighborCache->addressTimeout(dest); //TODO problem with new pings to context nodes
    // there must not be any contexts an rtt must be TIMEOUT before calling pingTimeout()
}
void BaseRpc::pingNode(const TransportAddress& dest, simtime_t timeout,
                       int retries, cPolymorphic* context,
                       const char* caption, RpcListener* rpcListener,
                       int rpcId, TransportType transportType,
                       bool overrideCache)
{
    simtime_t rtt;
    if (overrideCache || neighborCache->getAddressRtt(dest) ==
        NEIGHBORSTATE_UNKNOWN || neighborCache->getAddressRtt(dest) ==
        NEIGHBORSTATE_TIMEOUT) {
        // if unknown or overriding, send ping call as normal
        PingCall* call = new PingCall(caption);
        call->setLength(PINGCALL_L(call));
        RECORD_STATS(numPingSent++; bytesPingSent += call->byteLength());

        if (transportType == UDP_TRANSPORT ||
            (transportType != ROUTE_TRANSPORT &&
             getThisCompType() == OVERLAY_COMP)) {
            sendUdpRpcCall(dest, call, context, timeout, retries, rpcId,
                           rpcListener);
        } else {
            sendRouteRpcCall(getThisCompType(), dest, call, context,
                             DEFAULT_ROUTING, timeout, retries, rpcId,
                             rpcListener);
        }

        //neighborCache->setAddressPinged(dest);
        //neighborCache->cleanupCache();
        neighborCache->insertPingedNode(dest, context, rpcListener, rpcId);
    } else if ((rtt = neighborCache->getAddressRtt(dest)) !=
               NEIGHBORSTATE_WAITING) {
        // if it's time-out or a known value, reply with fake ping response
        PingResponse *pr = new PingResponse();
        pr->setNonce(0);
        if (neighborCache->getNodeHandle(dest).isUnspecified())
            pr->setSrcNode(dest);
        else
            pr->setSrcNode(neighborCache->getNodeHandle(dest));

        // inform listener
        if (rpcListener != NULL)
            rpcListener->handleRpcResponse(pr, context, rpcId, rtt);

        pingResponse(pr, context, rpcId, rtt);
        handleRpcResponse(pr, context, rpcId, rtt);

        //delete context;
        delete pr;
    } else {
        assert(neighborCache->getAddressRtt(dest) == NEIGHBORSTATE_WAITING);
        neighborCache->insertPingedNode(dest, context, rpcListener, rpcId);
    }
}
