
#ifndef __COMPARATOR_H
#define __COMPARATOR_H

#include <OverlayKey.h>

/**
 * Default Comparator
 */
template<class T>
class Comparator
{
public:
    /**
     * virtual destructor
     */
    virtual ~Comparator()
    {}

    /**
     * compares two variables of type T and indicates
     * which one is smaller or if they are equal
     *
     * @param lhs first variable to compare
     * @param rhs second variable to compare
     * @return -1 if rhs is smaller, 0 if lhs and rhs are
     *         equal and 1 if rhs is greater
     */
    virtual int compare( const T& lhs, const T& rhs ) const
    {
        return lhs.compareTo(rhs);
    }
};

/**
 * OverlayKey comparator
 */
class KeyComparator : public Comparator<OverlayKey>
    {}
;

/**
 * OverlayKey standard metric
 */
class KeyStdMetric
{
public:
    /**
     * calculates the distance from x to y with the standard metric
     *
     * @param x origination key
     * @param y destination key
     * @return |y-x|
     */
    inline OverlayKey distance(const OverlayKey& x,
                               const OverlayKey& y) const
    {
        return x > y ? (x-y) : (y-x);
    }
};

/**
 * OverlayKey XOR Metric
 */
class KeyXorMetric
{
public:
    /**
     * calculates the distance from x to y with the XOR metric
     *
     * @param x origination key
     * @param y destination key
     * @return x XOR y
     */
    inline OverlayKey distance(const OverlayKey& x,
                               const OverlayKey& y) const
    {
        return x^y;
    }
};


/**
 * OverlayKey Ring Metric
 */
class KeyRingMetric
{
public:
    /**
     * calculates the distance from x to y on a bidirectional ring
     *
     * @param x origination key
     * @param y destination key
     * @return |y-x| on a bidirectional ring
     */
    inline OverlayKey distance(const OverlayKey& x,
                               const OverlayKey& y) const
    {
        OverlayKey dist1(x - y);
        OverlayKey dist2(y - x);

	if (dist1 > dist2)
	    return dist2;
	else
	    return dist1;
    }
};


/**
 * OverlayKey Unidirectional Ring Metric
 */
class KeyUniRingMetric
{
public:
    /**
     * calculates the distance from x to y on a unidirectional ring
     *
     * @param x origination key
     * @param y destination key
     * @return distance from x to y on a unidirectional ring
     */
    inline OverlayKey distance(const OverlayKey& x,
                               const OverlayKey& y) const
    {
        return y-x;
    }
};


/**
 * OverlayKey prefix metric
 */
class KeyPrefixMetric
{
public:
    KeyPrefixMetric()
    {
        bitsPerDigit = 1;
    }

    /**
     * calculates the distance from x to y with the prefix metric
     *
     * @param x origination key
     * @param y destination key
     * @return |differing suffix|
     */
    inline OverlayKey distance(const OverlayKey& x,
                               const OverlayKey& y) const
    {
        return OverlayKey::getLength() / bitsPerDigit
               - x.sharedPrefixLength(y, bitsPerDigit);
    }

    inline void setBitsPerDigit(uint bitsPerDigit)
    {
        this->bitsPerDigit = bitsPerDigit;
    }

private:
    uint bitsPerDigit;
};

/**
 * OverlayKey distance comparator
 */
template<class Metric = KeyStdMetric>
class KeyDistanceComparator : public Comparator<OverlayKey>
{
private:
    Metric m; /**< indicates which metric to use for the comparison */
    OverlayKey key; /**< the relative key to which distances are compared */
public:

    /**
     * constructor
     */
    KeyDistanceComparator( const OverlayKey& relativeKey )
    {
        this->key = relativeKey;
    }

    /**
     * indicates which of the two given keys is closer to
     * the relative key
     *
     * @param lhs first key
     * @param rhs second key
     * @return -1 if lhs is closer, 0 if lhs and rhs are equal and 1
     *         if rhs closer to the relative key
     */
    int compare( const OverlayKey& lhs, const OverlayKey& rhs ) const
    {
        return m.distance(lhs, key).compareTo(m.distance(rhs, key));
    }
};

template<>
class KeyDistanceComparator<KeyPrefixMetric> : public Comparator<OverlayKey>
{
private:
    KeyPrefixMetric m; /**< indicates which metric to use for the comparison */
    OverlayKey key; /**< the relative key to which distances are compared */
public:

    /**
     * constructor
     */
    KeyDistanceComparator(const OverlayKey& relativeKey, uint bitsPerDigit = 1)
    {
        key = relativeKey;
        m.setBitsPerDigit(bitsPerDigit);
    }
    /**
     * indicates which of the two given keys has a longer distance to
     * the relative key
     *
     * @param lhs first key
     * @param rhs second key
     * @return -1 if lhs is closer, 0 if lhs and rhs are equal and 1
     *         if rhs closer to the relative key
     */
    int compare( const OverlayKey& lhs, const OverlayKey& rhs ) const
    {
        return m.distance(lhs, key).compareTo(m.distance(rhs, key));
    }
};

#endif

