//
// Copyright (C) 2006 Institut fuer Telematik, Universitaet Karlsruhe (TH)
//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public License
// as published by the Free Software Foundation; either version 2
// of the License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
//

/**
 * @file NeighborCache.cc
 * @author Antonio Zea
 */

#include <cassert>

#include <TransportAddress.h>
#include <NodeHandle.h>

#include "NeighborCache.h"


std::ostream& operator<<(std::ostream& os,
                         const NeighborCache::NeighborCacheEntry &p)
{
    os << p.rtt << " (inserted: " << p.insertTime /*<< ", #contexts: "
       << p.pingContexts.size() */<< ", key: "
       << (p.nodeRef.isUnspecified() ? "<unspec>" : p.nodeRef.key.toString(16))
       << ")";

    return os;
}


Define_Module(NeighborCache);

void NeighborCache::initialize()
{
    neighborCache.clear();
    WATCH_MAP(neighborCache);

    enableNeighborCache = par("enableNeighborCache");
    rttExpirationTime = par("rttExpirationTime");
    maxSize = par("maxSize");
}

NeighborCache::~NeighborCache()
{
    for (std::map<TransportAddress, NeighborCacheEntry>::iterator it =
        neighborCache.begin(); it != neighborCache.end(); ++it) {
        if (it->second.waitingContexts) {
            for (uint i = 0;
                 i < it->second.waitingContexts->pingContexts.size();
                 ++i) {
                delete it->second.waitingContexts->pingContexts[i];
            }

            delete it->second.waitingContexts;
        }
    }
}

bool NeighborCache::insertPingedNode(const TransportAddress& handle,
                                     cPolymorphic* context,
                                     RpcListener* rpcListener,
                                     int rpcId)
{
    if (!enableNeighborCache) return false;
    if (neighborCache.count(handle) == 0) {// || !enableNeighborCache) {
        NeighborCacheEntry entry;

        entry.insertTime = simulation.simTime();
        entry.rtt = NEIGHBORSTATE_WAITING;
        entry.nodeRef = NodeHandle::UNSPECIFIED_NODE;

        neighborCache[handle] = entry;
        neighborCacheExpireMap.insert(std::make_pair(entry.insertTime, handle));

        cleanupCache();

        assert(neighborCache.size() == neighborCacheExpireMap.size());
        return false;
    } else {
        NeighborCacheEntry& entry = neighborCache[handle];

        // waiting?
        if (entry.rtt == NEIGHBORSTATE_WAITING) {
            if (entry.waitingContexts == NULL)
                entry.waitingContexts = new WaitingContexts();
            entry.waitingContexts->pingContexts.push_back(context);
            entry.waitingContexts->pingListeners.push_back(rpcListener);
            entry.waitingContexts->pingIds.push_back(rpcId);

            return true;
        } else {
            if (entry.waitingContexts != NULL &&
                entry.waitingContexts->pingContexts.size() != 0) {
                throw new cRuntimeError("not waiting for response,"
                                        " but additional contexts found!");
            }

            updateEntry(handle, entry.insertTime);

            entry.rtt = NEIGHBORSTATE_WAITING;
            entry.insertTime = simulation.simTime();
            entry.nodeRef = NodeHandle::UNSPECIFIED_NODE;

            return false;
        }
    }
}

//void NeighborCache::clearNodeContexts(const TransportAddress& handle)
//{
//    //std::cout << simTime() << " get " << handle.ip << std::endl;
//    if (neighborCache.count(handle) == 0) {
//        //return;
//        throw new cRuntimeError("NeighborCache.cc: clearNodeContexts was asked"
//                                "for a non-existant node context.");
//    }
//    NeighborCacheEntry& entry = neighborCache[handle];
//
//    for (uint i = 0; i < entry.pingContexts.size(); ++i) {
//        delete entry.pingContexts[i];
//    }
//    entry.pingContexts.clear();
//    entry.pingListeners.clear();
//    entry.pingIds.clear();
//}

void NeighborCache::addressTimeout(const TransportAddress& handle)
{
    if (!enableNeighborCache) return;
    if (neighborCache.count(handle) == 0) {
        NeighborCacheEntry entry;

        entry.insertTime = simulation.simTime();
        entry.rtt = NEIGHBORSTATE_TIMEOUT;
        entry.nodeRef = NodeHandle::UNSPECIFIED_NODE;

        neighborCache[handle] = entry;
        neighborCacheExpireMap.insert(std::make_pair(entry.insertTime, handle));

        cleanupCache();
    } else {
        NeighborCacheEntry& entry = neighborCache[handle];

        updateEntry(handle, entry.insertTime);

        entry.insertTime = simulation.simTime();
        entry.rtt = NEIGHBORSTATE_TIMEOUT;
    }
    assert(neighborCache.size() == neighborCacheExpireMap.size());
}


void NeighborCache::insertAddressRtt(const NodeHandle &add, simtime_t rtt)
{
    if (!enableNeighborCache) return;
    TransportAddress ta = add;
    if (neighborCache.count(ta) == 0) {
        NeighborCacheEntry entry;

        entry.insertTime = simulation.simTime();
        entry.rtt = rtt;
        entry.nodeRef = add;

        neighborCache[ta] = entry;
        neighborCacheExpireMap.insert(std::make_pair(entry.insertTime, add));

        cleanupCache();
    } else {
        updateEntry(add, neighborCache[ta].insertTime);

        neighborCache[ta].insertTime = simulation.simTime();
        neighborCache[ta].rtt = rtt;
        neighborCache[ta].nodeRef = add;
    }
    assert(neighborCache.size() == neighborCacheExpireMap.size());
}

simtime_t NeighborCache::getAddressRtt(const TransportAddress &add)
{
    // cache disabled or entry not there
    if (!enableNeighborCache ||
        add.isUnspecified() ||
        (neighborCache.count(add) == 0)) {
        //std::cout << "NeighobrCache: MISS!" << std::endl;
        return NEIGHBORSTATE_UNKNOWN;
    }

    NeighborCacheEntry entry = neighborCache[add];

    if (entry.rtt == NEIGHBORSTATE_WAITING ||
        entry.rtt == NEIGHBORSTATE_UNKNOWN)
        return entry.rtt;
    // entry expired
    if ((simulation.simTime() - entry.insertTime) >= rttExpirationTime) {
        entry.rtt = NEIGHBORSTATE_UNKNOWN;
        return NEIGHBORSTATE_UNKNOWN;
    }
    //std::cout << "NeighborCache: HIT!" << std::endl;
    return entry.rtt;
}

NodeHandle NeighborCache::getNodeHandle(const TransportAddress &add)
{
    if (neighborCache.count(add) == 0) {
        throw new cRuntimeError("NeighborCache.cc: getNodeHandle was asked for "
                                "a non-existent node reference.");
    }
    return neighborCache[add].nodeRef;
}

bool NeighborCache::cleanupCache()
{
    bool result = false;
    uint size = neighborCache.size();
    if (size > maxSize) {
        std::map<simtime_t, TransportAddress>::iterator it;
        for (uint i = 0; i < (size - (maxSize / 2)); ++i) {
            it = neighborCacheExpireMap.begin();
            if ((neighborCache[it->second].rtt == NEIGHBORSTATE_WAITING) ||
                (neighborCache[it->second].insertTime == simTime())) {
                break;
            }
            neighborCache.erase(it->second);
            neighborCacheExpireMap.erase(it);
            result = true;
        }
    }
    assert(neighborCache.size() == neighborCacheExpireMap.size());
    return result;
}

void NeighborCache::updateEntry(const TransportAddress& address,
                                simtime_t insertTime)
{
    std::multimap<simtime_t, TransportAddress>::iterator it =
        neighborCacheExpireMap.lower_bound(insertTime);
    while (it->second != address) ++it;
    neighborCacheExpireMap.erase(it);
    neighborCacheExpireMap.insert(std::make_pair(simTime(), address));
    assert(neighborCache.size() == neighborCacheExpireMap.size());
}
