//
// Copyright (C) 2006 Institut fuer Telematik, Universitaet Karlsruhe (TH)
//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public License
// as published by the Free Software Foundation; either version 2
// of the License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
//

/**
 * @file Bamboo.cc
 * @author Gerhard Petruschat
 */

#include <IPAddressResolver.h>
#include <IPvXAddress.h>
#include <InterfaceTable.h>
#include <IPv4InterfaceData.h>
#include <RpcMacros.h>
#include <InitStages.h>
#include <SimpleNetConfigurator.h>
#include <GlobalStatistics.h>
#include <LookupListener.h>
#include <AbstractLookup.h>

#include "Bamboo.h"

Define_Module(Bamboo);

Bamboo::~Bamboo()
{

    // destroy self timer messages
    cancelAndDelete(joinTimeout);
    cancelAndDelete(readyWait);
    cancelAndDelete(repairTaskTimeout);
    cancelAndDelete(leafsetMaintenanceTimeout);
    cancelAndDelete(globalTuningTimeout);

    purgeVectors();
}

class BambooLookupListener : public LookupListener
{
private:
    Bamboo* overlay;
public:
    BambooLookupListener(Bamboo* overlay)
    {
        this->overlay = overlay;
    }

    virtual void lookupFinished(AbstractLookup *lookup)
    {
        overlay->lookupFinished(lookup);
        delete this;
    }
};

void Bamboo::initializeOverlay(int stage)
{
    if ( stage != MIN_STAGE_OVERLAY )
        return;

    // Bamboo provides KBR services
    kbr = true;

    baseInit();

    repairTaskTimeoutAmount = par("repairTaskTimeoutAmount");
    leafsetMaintenanceTimeoutAmount = par("leafsetMaintenanceTimeoutAmount");
    globalTuningTimeoutAmount = par("globalTuningTimeoutAmount");

    joinTimeout = new cMessage("joinTimeout");
    readyWait = new cMessage("readyWait");
    //joinUpdateWait = new cMessage("joinUpdateWait");

    repairTaskTimeout = new cMessage("repairTaskTimeout");
    leafsetMaintenanceTimeout = new cMessage("leafsetMaintenanceTimeout");
    globalTuningTimeout = new cMessage("globalTuningTimeout");

    sendPullFlag = true;

}

void Bamboo::joinOverlay()
{
    changeState(INIT);

    if (bootstrapNode.isUnspecified()) {
        // no existing pastry network -> first node of a new one
        changeState(READY);
    } else {
        // join existing pastry network
        changeState(JOINING_2);
    }
}

void Bamboo::changeState(int toState)
{

    baseChangeState(toState);

    switch (toState) {
    case INIT:

        break;

    case DISCOVERY:

        break;

    case JOINING_2:

    {
        PastryLeafsetMessage* msg = new PastryLeafsetMessage("Leafset");
         msg->setPastryMsgType(PASTRY_MSG_LEAFSET_PULL);
         msg->setStatType(MAINTENANCE_STAT);
         msg->setSender(thisNode);
         msg->setSendStateTo(thisNode);
         leafSet->dumpToStateMessage(msg);
         msg->setLength(PASTRYLEAFSET_L(msg));
         RECORD_STATS(leafsetSent++; leafsetBytesSent += msg->byteLength());
         std::vector<TransportAddress> sourceRoute;
         sourceRoute.push_back(bootstrapNode);
         sendToKey(thisNode.key, msg, 0/*1*/, sourceRoute);
     }

        break;

    case READY:

        // schedule routing table maintenance task
        cancelEvent(repairTaskTimeout);
        scheduleAt(simTime() + repairTaskTimeoutAmount, repairTaskTimeout);
        //EV << "local tuning scheduled for " << repairTaskTimeoutAmount << endl;
        //leaf2ask = NULL;

        cancelEvent(leafsetMaintenanceTimeout);
        //scheduleAt(simTime() + leafsetMaintenanceTimeoutAmount, leafsetMaintenanceTimeout);
        scheduleAt(simTime() + strToSimtime("200ms"), leafsetMaintenanceTimeout);

        cancelEvent(globalTuningTimeout);
        scheduleAt(simTime() + globalTuningTimeoutAmount, globalTuningTimeout);

        break;
    }
}

void Bamboo::handleTimerEvent(cMessage* msg)
{

    if (msg->isName("joinTimeout")) {
        EV << "Pastry: join timeout expired, restarting..." << endl;
        joinOverlay();
    } else if (msg->isName("readyWait")) {
        if (partialJoinPath) {
            RECORD_STATS(joinPartial++);
            sort(stReceived.begin(), stReceived.end(), stateMsgIsSmaller);

            // start pinging the nodes found in the first state message:
	    stReceivedPos = stReceived.begin();
            stateCache = *stReceivedPos;
            EV << "Pastry: joining despite some missing STATE messages."
               << endl;
            pingNodes();
        } else {
            EV << "Pastry: timeout waiting for missing state messages in JOIN "
                "state, restarting..." << endl;
            joinOverlay();
        }
    }
    else if (msg->isName("repairTaskTimeout")) {
        EV << "Pastry: starting routing table maintenance" << endl;
        doRoutingTableMaintenance();
        scheduleAt(simTime()+repairTaskTimeoutAmount, repairTaskTimeout);
    }
    else if (msg->isName("leafsetMaintenanceTimeout")) {
        EV << "Pastry: starting leafset maintenance" << endl;
        doLeafsetMaintenance();
        scheduleAt(simTime()+leafsetMaintenanceTimeoutAmount, leafsetMaintenanceTimeout);
    }
    else if (msg->isName("globalTuningTimeout")) {
        EV << "Pastry: starting global tuning" << endl;
        doGlobalTuning();
        scheduleAt(simTime()+globalTuningTimeoutAmount, globalTuningTimeout);
    }
}

void Bamboo::handleUDPMessage(BaseOverlayMessage* msg)
{
    PastryMessage* pastryMsg = check_and_cast<PastryMessage*>(msg);
    uint type = pastryMsg->getPastryMsgType();

    if (debugOutput) {
        EV << "Pastry: incoming message of type ";
        switch(type) {
        case PASTRY_MSG_STD:
            EV << "PASTRY_MSG_STD";
            break;
        case PASTRY_MSG_JOIN:
            EV << "PASTRY_MSG_JOIN";
            break;
        case PASTRY_MSG_STATE:
            EV << "PASTRY_MSG_STATE";
            break;
        case PASTRY_MSG_LEAFSET:
            EV << "PASTRY_MSG_LEAFSET";
            break;
        case PASTRY_MSG_LEAFSET_PULL:
            EV << "PASTRY_MSG_LEAFSET_PULL";
            break;
        case PASTRY_MSG_ROWREQ:
            EV << "PASTRY_MSG_ROWREQ";
            break;
        case PASTRY_MSG_RROW:
            EV << "PASTRY_MSG_RROW";
            break;
        case PASTRY_MSG_REQ:
            EV << "PASTRY_MSG_REQ";
            break;
        default:
            EV << "UNKNOWN (" << type <<")";
            break;
        }
        EV << endl;
    }

    switch (type) {
    case PASTRY_MSG_STD:
        opp_error("Pastry received PastryMessage of unknown type!");
        break;
    case PASTRY_MSG_JOIN:

        break;

    case PASTRY_MSG_LEAFSET: {
        PastryLeafsetMessage* lmsg =
            check_and_cast<PastryLeafsetMessage*>(pastryMsg);
        RECORD_STATS(leafsetReceived++; leafsetBytesReceived +=
            lmsg->byteLength());

        if (state == JOINING_2) {
            cancelEvent(joinTimeout);
        }

        if ((state == JOINING_2) || (state == READY)) {
            handleLeafsetMessage(lmsg, true);
        } else {
            delete lmsg;
        }
    }
        break;

    case PASTRY_MSG_LEAFSET_PULL: {
        PastryLeafsetMessage* lmsg =
            check_and_cast<PastryLeafsetMessage*>(pastryMsg);
        RECORD_STATS(leafsetReceived++; leafsetBytesReceived +=
            lmsg->byteLength());

        if (state == READY) {

            sendLeafset(lmsg->getSendStateTo());
            handleLeafsetMessage(lmsg, true);

        } else {
            delete lmsg;
        }
    }
        break;

    case PASTRY_MSG_ROWREQ: {

        PastryRoutingRowRequestMessage* rtrmsg =
            check_and_cast<PastryRoutingRowRequestMessage*>(pastryMsg);
        RECORD_STATS(routingTableReqReceived++; routingTableReqBytesReceived +=
            rtrmsg->byteLength());
        if (state == READY)
            if (rtrmsg->getRow() == -1)
                sendRoutingRow(rtrmsg->getSendStateTo(), routingTable->getLastRow());
                else if (rtrmsg->getRow() > routingTable->getLastRow())
                    EV << "Pastry: received request for nonexistent routing"
                       << "table row, dropping message!" << endl;
                else sendRoutingRow(rtrmsg->getSendStateTo(), rtrmsg->getRow());
        else
            EV << "Pastry: received routing table request before reaching "
               << "READY state, dropping message!" << endl;
       delete rtrmsg;
    }
        break;

    case PASTRY_MSG_RROW: {
        PastryRoutingRowMessage* rtmsg =
            check_and_cast<PastryRoutingRowMessage*>(pastryMsg);
        RECORD_STATS(routingTableReceived++; routingTableBytesReceived +=
            rtmsg->byteLength());

        if (state == READY) {

            uint nodesPerRow = routingTable->nodesPerRow;
            PastryStateMessage* stateMsg;

            stateMsg = new PastryStateMessage("STATE");
            stateMsg->setTimestamp(rtmsg->getTimestamp());
            stateMsg->setPastryMsgType(PASTRY_MSG_STATE);
            stateMsg->setStatType(MAINTENANCE_STAT);
            stateMsg->setSender(rtmsg->getSender());
            stateMsg->setLeafSetArraySize(0);
            stateMsg->setNeighborhoodSetArraySize(0);
            stateMsg->setRoutingTableArraySize(nodesPerRow);

            for (uint i = 0; i < nodesPerRow; i++) {
                stateMsg->setRoutingTable(i, rtmsg->getRoutingTable(i));
            }

            handleStateMessage(stateMsg);
        }

        delete rtmsg;
    }
        break;

    case PASTRY_MSG_REQ: {
        PastryRequestMessage* lrmsg =
            check_and_cast<PastryRequestMessage*>(pastryMsg);
        handleRequestMessage(lrmsg);
    }
        break;

    case PASTRY_MSG_STATE: {
        PastryStateMessage* stateMsg =
            check_and_cast<PastryStateMessage*>(msg);
        RECORD_STATS(stateReceived++; stateBytesReceived +=
                     stateMsg->byteLength());
        handleStateMessage(stateMsg);
    }
        break;
    }
}

void Bamboo::doLeafsetMaintenance(void)
{

    const TransportAddress& ask = leafSet->getRandomNode();
    if (!ask.isUnspecified()) {
        sendLeafset(ask, true);
        EV << "Bamboo: leafset maintenance: pulling leafset from "
           << ask << endl;
    }

}

void Bamboo::doGlobalTuning(void)
{

    OverlayKey newKey = OverlayKey::random();
    //OverlayKey OverlayKey::randomSuffix  	(  	uint   	 pos  	 )   	 const

    createLookup()->lookup(newKey, 1, 0, 0,
                 new BambooLookupListener(this));

}

bool Bamboo::handleFailedNode(const TransportAddress& failed)
{
    if (state != READY) {
        return false;
    }

    if (failed.isUnspecified())
        opp_error("Bamboo::handleFailedNode(): failed is unspecified!");

    const TransportAddress& lsAsk = leafSet->failedNode(failed);
    /*const TransportAddress& rtAsk = */routingTable->failedNode(failed);
    neighborhoodSet->failedNode(failed);

    if (lsAsk.isUnspecified() && (! leafSet->isValid())) {
        EV << "Pastry: lost connection to the network, trying to re-join."
           << endl;
        joinOverlay();
        return false;
    }

    return true;
}

void Bamboo::checkProxCache(void)
{
    if (state == JOINING_2) {
        changeState(READY);
        return;
    }

    // state == READY
    simtime_t now = simTime();

    // no cached STATE message?
    if (!(stateCache.msg && stateCache.prox)) return;

    // some entries not yet determined?
    if (find(stateCache.prox->pr_rt.begin(), stateCache.prox->pr_rt.end(),
             PASTRY_PROX_PENDING) != stateCache.prox->pr_rt.end()
        ) return;
    if (find(stateCache.prox->pr_ls.begin(), stateCache.prox->pr_ls.end(),
             PASTRY_PROX_PENDING) != stateCache.prox->pr_ls.end()
        ) return;
    if (find(stateCache.prox->pr_ns.begin(), stateCache.prox->pr_ns.end(),
             PASTRY_PROX_PENDING) != stateCache.prox->pr_ns.end()
        ) return;

    // merge info in own state tables
    // except leafset (was already handled in handleStateMessage)
    if (neighborhoodSet->mergeState(stateCache.msg, *stateCache.prox))
    lastStateChange = now;
    EV << "Merging nodes into routing table" << endl;
    if (routingTable->mergeState(stateCache.msg, *stateCache.prox)) {
        lastStateChange = now;
        EV << "merged nodes into routing table" << endl;
    }

    updateTooltip();

    delete stateCache.msg;
    stateCache.msg = NULL;
    delete stateCache.prox;
	  stateCache.prox = NULL;

    // process next queued message:
    if (! stateCacheQueue.empty()) {
        stateCache = stateCacheQueue.front();
	      stateCacheQueue.pop();
        pingNodes();
    }

}

void Bamboo::handleStateMessage(PastryStateMessage* msg)
{
    if (debugOutput) {
        EV << "Pastry::handleStateMessage() new STATE message to process "
           << static_cast<void*>(msg) << " in state " <<
            ((state == READY)?"READY":((state == JOINING_2)?"JOIN":"INIT"))
           << endl;
    }

    if (state == INIT) {
        EV << "Pastry: can't handle state messages until at least reaching "
            "JOIN state." << endl;
        delete msg;
        return;
    }

    PastryStateMsgHandle handle(msg);

    if (state == JOINING_2) {
        determineAliveTable(msg);
        leafSet->mergeState(msg, aliveTable);
        // merged state into leafset right now
        lastStateChange = simTime();
        newLeafs();
        updateTooltip();

        // no state message is processed right now, start immediately:
       stateCache = handle;
       pingNodes();

        return;
    }

    // determine aliveTable to prevent leafSet from merging nodes that are
    // known to be dead:
    determineAliveTable(msg);
    if (leafSet->mergeState(msg, aliveTable)) {
        // merged state into leafset right now
        lastStateChange = simTime();
        newLeafs();
        updateTooltip();
    }
    // in READY state, only ping nodes to get proximity metric:
    if (!stateCache.msg) {
        // no state message is processed right now, start immediately:
        stateCache = handle;
        pingNodes();
    } else {
        // enqueue message for later processing:
        stateCacheQueue.push(handle);
        prePing(msg);
    }
}

void Bamboo::lookupFinished(AbstractLookup *lookup)
{
    EV << "[BambooListener::lookupFinished()]\n";
    if (lookup->isValid()) {
            EV  << "    Lookup successful" << endl;
    } else {
            EV << "    Lookup failed" << endl;
    }

}
