//
// Copyright (C) 2007 Institut fuer Telematik, Universitaet Karlsruhe (TH)
//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public License
// as published by the Free Software Foundation; either version 2
// of the License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
//

/**
 * @file Broose.h
 * @author Jochen Schenk
 */

#ifndef __BROOSE_H_
#define __BROOSE_H_

#include <omnetpp.h>
#include <BaseOverlay.h>
#include <RpcListener.h>
#include <OverlayKey.h>
#include "BrooseHandle.h"
#include "BrooseBucket.h"
#include "BrooseMessage_m.h"

#include <map>

class BrooseBucket;

/**
 * Broose overlay module
 *
 * Implementation of the Broose KBR overlay as described in
 * "Broose: A Practical Distributed Hashtable Based on the
 * De-Bruijn Topology" by Anh-Tuan Gai and Laurent Viennot
 *
 * @author Jochen Schenk
 * @see Bucket
 */


// protocol states
enum
{
    INIT,
    RSET,
    BSET,
    READY
};


class Broose : public BaseOverlay
{
  public:
    ~Broose();

    // see BaseOverlay.h
    virtual void initializeOverlay(int stage);

    // see BaseOverlay.h
    virtual void finishOverlay();

    // see BaseOverlay.h
    virtual bool isSiblingFor(const NodeHandle& node,
                              const OverlayKey& key,
                              int numSiblings,
                              bool* err);

    // see BaseOverlay.h
    virtual void joinOverlay();

    // see BaseOverlay.h
    virtual void handleUDPMessage(BaseOverlayMessage* msg);

    // see BaseOverlay.h
    virtual void recordOverlaySentStats(BaseOverlayMessage* msg);

    // see BaseOverlay.h
    virtual bool handleRpc(BaseCallMessage* msg);

    // see BaseOverlay.h
    virtual void handleTimerEvent(cMessage* msg);

    /**
     * updates information shown in tk-environment
     */
    void updateTooltip();

  protected:
    //parameter
    int chooseLookup; /**< decides which kind of lookup (right/left shifting) is used */
    int joinDelay; /**< time interval between two join tries */
    int protoState; /**< the state in which a node currently is */
    int receivedJoinResponse; /**< number of received join response messages */
    int receivedBBucketLookup; /**< number of received lookup responses for the B bucket */
    int numberBBucketLookup; /**< maximal number of lookup reponses for the B bucket */
    int receivedLBucketLookup; /**< number of received lookup responses for the L bucket */
    int numberLBucketLookup;  /**< maximal number of lookup reponses for the L bucket */
    int shiftingBits; /**< number of bits shifted in/out each step */
    int powShiftingBits; /**< 2^{variable shiftingBits} */
    uint bucketSize;  /**< maximal number of bucket entries */
    uint rBucketSize; /**< maximal number of entries in the r buckets */
    int parallelRequests;  /**< number ob parallel requests */
    int keyLength; /**< length of the node and data IDs */
    bool refresh; /**< is the node restarting the bootstrap protocol or is it a new node */
    int pingDelay; /**< time intervall between bucket refreshs */
    int refreshTime; /**< time intervall after which a ping is done */
    uint userDist; /**< how many hops are added to the estimated hop count */
    int numPings; /**< actual number of received ping messages */
    int maxPings; /**< total number of ping messages */
    int numberRetries; /**< number of retries in case of timeout */

    //statistics
    int bucketCount; /**< number of Bucket messages */
    int bucketBytesSent; /**< length of all Bucket messages */
    int broosePingCount; /**< number of Ping messages */
    int broosePingBytesSent; /**< length of all Ping messages */
    int numFailedPackets; /**< number of packets which couldn't be routed correctly */

    //module references
    BrooseBucket *lBucket, *bBucket;  /**< */
    BrooseBucket **rBucket;  /**< */

    // timer
    cMessage* join_timer;  /**< */
    cMessage* bucket_timer; /**< timer to reconstruct all buckets */

    //node handles
    NodeHandle bootstrapNode;  /**< node handle holding the bootstrap node */
    BrooseHandle thisBrooseNode; /**< this is a BrooseHandle of the current node like thisNode */

    //functions

    /**
     * handles a expired join timer
     *
     * @param msg the timer self-message
     */
    void handleJoinTimerExpired(cMessage* msg);

    /**
     * handles a expired bucket refresh timer
     *
     * @param msg the bucket refresh self-message
     */
    void handleBucketTimerExpired(cMessage* msg);

    /**
     * changes the node's state
     *
     * @param state the state to which a node is changing
     */
    void changeState(int state);

    // see BaseOverlay.h
    NodeVector* findNode(const OverlayKey& key,
			 int numRedundantNodes,
			 int numSiblings,
			 BaseOverlayMessage* msg);

    /** decides if a specific key is managed by this node
    *
    * @param key the key
    * @return bool value, if the key is managed by this node
    */
    bool keyBelongsToNode(const OverlayKey& key);

    // see BaseOverlay.h
    int getMaxNumSiblings();

    // see BaseOverlay.h
    int getMaxNumRedundantNodes();

    /**
     * debug function which output the content of the node's buckets
     */
    void displayBucketState();

    /**
     * outputs an Overlay key in the binary system
     *
     * @param key the key to output
     */
    void binaryOutput(const OverlayKey& key);

    // see BaseOverlay.h
    void handleRpcResponse(BaseResponseMessage* msg,
                           cPolymorphic* context,
                           int rpcId, simtime_t rtt);

    // see BaseOverlay.h
    void handleRpcTimeout(BaseCallMessage* msg, const TransportAddress& dest,
                          cPolymorphic* context, int rpcId,
                          const OverlayKey& destKey);

    /**
     * This method is called if an Find Node Call timeout has been reached.
     *
     * @param findNode The original FindNodeCall
     * @param dest the destination node
     * @param destKey the destination OverlayKey
     */
    void handleFindNodeTimeout(FindNodeCall* findNode,
			       const TransportAddress& dest,
			       const OverlayKey& destKey);

    /**
     * handles a received Bucket request
     *
     * @param msg the message to process
     */
    void handleBucketRequestRpc(BucketCall* msg);

    /**
     * handles a received Bucket response
     *
     * @param msg the message to process
     */
    void handleBucketResponseRpc(BucketResponse* msg);

    /**
     * handles a received Bucket timeout
     *
     * @param msg the message to process
     */
    void handleBucketTimeout(BucketCall* msg);

    /**
     * handles a received Ping request
     *
     * @param msg the message to process
     */
    void handleBroosePingRequestRpc(BroosePingCall* msg);

    /**
     * handles a received Ping response
     *
     * @param msg the message to process
     * @param rtt the round trip time of the message
     */
    void handleBroosePingResponseRpc(BroosePingResponse* msg, simtime_t rtt);

    /**
     * handles a received Ping timeout
     *
     * @param msg the message to process
     * @param dest the destination node
     * @param destKey the destination OverlayKey
     */
    void handleBroosePingTimeout(BroosePingCall* msg,
				 const TransportAddress& dest,
				 const OverlayKey& destKey);

    /**
     * updates the timestamp of a node in all buckets
     *
     * @param node node handle which should be updated
     */
    void setLastSeen (NodeHandle node);

    /**
     * adds a node to all buckets
     *
     * @param node node handle which should be added
     * @param bucket reserved
     */
    void addNode(NodeHandle node, int bucket = 0);

    /**
     * resets the counter of failed responses
     *
     * @param node node handle of the responding node
     */
    void resetFailedResponses(NodeHandle node);

    /**
     * sets the rtt to a node in all buckets
     *
     * @param node node handle to which a rtt is added/updated
     * @param rtt round trip time to the node
     */
    void setRTT(NodeHandle node, simtime_t rtt);

    friend class BrooseBucket;
};



#endif
