//
// Copyright (C) 2006 Institut fuer Telematik, Universitaet Karlsruhe (TH)
//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public License
// as published by the Free Software Foundation; either version 2
// of the License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
//

/**
 * @file ChordFingerTable.h
 * @author Markus Mauch, Ingmar Baumgart
 */

#ifndef __CHORDFINGERTABLE_H_
#define __CHORDFINGERTABLE_H_

#include <vector>
#include <map>

#include <omnetpp.h>

#include <NodeVector.h>
#include <InitStages.h>

class BaseOverlay;

typedef std::multimap<double, NodeHandle> Successors;
typedef std::pair<NodeHandle, Successors> FingerEntry;

std::ostream& operator<<(std::ostream& os, const Successors& suc);
std::ostream& operator<<(std::ostream& os, const FingerEntry& entry);

/**
 * Chord's finger table module
 *
 * This modul contains the finger table of the Chord implementation.
 *
 * @author Markus Mauch, Ingmar Baumgart
 * @see Chord
 */
class ChordFingerTable : public cSimpleModule
{
  public:

    virtual int numInitStages() const
    {
        return MAX_STAGE_OVERLAY + 1;
    }

    virtual void initialize(int stage);
    virtual void handleMessage(cMessage* msg);

    /**
     * Sets up the finger table
     *
     * Sets up the finger table and makes all fingers pointing
     * to the node itself. Should be called on startup to initialize
     * the finger table.
     *
     * @param size number of fingers
     * @param owner set all fingers to the key of node handle owner
     * @param overlay pointer to the main chord module
     */
    virtual void initializeTable(uint size, const NodeHandle& owner,
	BaseOverlay* overlay);

    /**
     * Sets a particular finger to point to node
     *
     * @param pos number of the finger to set
     * @param node set finger to this node
     */
    virtual void setFinger(uint pos, const NodeHandle& node);
    virtual void setFinger(uint pos, const Successors& nodes);

    virtual bool updateFinger(uint pos, const NodeHandle& node, double rtt);

    /**
     * Returns the NodeVector of a particular finger
     *
     * @param pos number of the finger to get
     * @return NodeVector of the particular finger(s)
     */
    virtual const NodeHandle& getFinger(uint pos);

    virtual NodeVector* getFinger(uint pos, const OverlayKey& key);

    /**
     * Returns the size of the finger table
     *
     * @return number of fingers
     */
    virtual uint getSize();

  protected:

    std::vector<FingerEntry> fingerTable; /**< the finger table vector */

    BaseOverlay* overlay; /**< pointer to the main chord module */
};

#endif
