//
// Copyright (C) 2006 Institut fuer Telematik, Universitaet Karlsruhe (TH)
//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public License
// as published by the Free Software Foundation; either version 2
// of the License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

#include "Kademlia.h"
#include "KademliaMessage_m.h"

#include <assert.h>

#include <IPAddressResolver.h>
#include <IPvXAddress.h>
#include <InterfaceTable.h>
#include <IPv4InterfaceData.h>
#include "TopologyVis.h"
#include <AbstractLookup.h>
#include <LookupListener.h>
#include <RpcMacros.h>
#include <BootstrapList.h>

#if 0
#define BUCKET_CONSISTENCY(msg) \
    do {\
        if (siblingTable->size() > 0) {\
            uint z = routingBucketIndex(siblingTable->back().key) - 1;\
            if (routingTable[z - 1] != NULL && routingTable[z - 1]->size())\
                breakpoint(#msg);\
        }\
        for (uint y = 0; y < routingTable.size(); ++y) {\
            if (routingTable[y] != NULL && routingTable[y]->size() > k)\
                breakpoint(#msg);\
        }\
    } while (false)
#else
#define BUCKET_CONSISTENCY(msg)
#endif

Define_Module(Kademlia);

std::ostream& operator<<(std::ostream& os, const KademliaBucket* n)
{
    if (n == NULL)
        os << "NULL";
    else {
        for (KademliaBucket::const_iterator i=n->begin(); i !=n->end(); i++) {
            os << *i << endl;
        }
        os << "last usage = " << n->getLastUsage()
           << ", last update = " << n->getLastUpdate();
    }
    return os;
};

class KademliaLookupListener : public LookupListener
{
private:
    Kademlia* overlay;
public:
    KademliaLookupListener(Kademlia* overlay)
    {
        this->overlay = overlay;
    }

    virtual void lookupFinished(AbstractLookup *lookup)
    {
        overlay->lookupFinished(lookup->isValid());
        delete this;
    }
};

//-----------------------------------------------------------------------------

void Kademlia::initializeOverlay(int stage)
{
    if (stage != MIN_STAGE_OVERLAY)
        return;

    // Kademlia provides KBR services
    kbr = true;

    // setup kademlia parameters
    minSiblingTableRefreshInterval = par("minSiblingTableRefreshInterval");
    minBucketRefreshInterval = par("minBucketRefreshInterval");
    maxStaleCount = par("maxStaleCount");
    pingNewSiblings = par("pingNewSiblings");

    activePing = par("activePing");

    k = par("k");
    b = par("b");
    s = par("s");

    // self-message
    bucketRefreshTimer = new cMessage("bucketRefreshTimer");

    nodesReplaced = 0;

    comparator = NULL;
    siblingTable = NULL;
}

Kademlia::~Kademlia()
{
    routingDeinit();

    replacementCache.clear();
    delete comparator;
    cancelAndDelete(bucketRefreshTimer);
}

void Kademlia::finishOverlay()
{
    // cancel timer
    cancelEvent(bucketRefreshTimer);

    simtime_t time = globalStatistics->calcMeasuredLifetime(creationTime);
    if(time == 0) return;

    globalStatistics->addStdDev("Kademlia: Nodes replaced in buckets/s",
                                nodesReplaced / time);
}

void Kademlia::joinOverlay()
{
    // remove current node handle from the bootstrap list
    if (!thisNode.key.isUnspecified()) {
        bootstrapList->removeBootstrapNode(thisNode);
    }

    // initialize routing
    routingDeinit();
    routingInit();

    const NodeHandle& handle = bootstrapList->getBootstrapNode();

    if (!handle.isUnspecified()) {
        // ping the bootstrap node to start bootstrapping
        pingNode(handle);
        //std::cout << thisNode << " bootstrapping via " << handle << std::endl;
/*
        FindNodeCall* findNodeCall = new FindNodeCall();
        findNodeCall->setLookupKey(handle.getKey());
        findNodeCall->setSrcNode(getThisNode());
        findNodeCall->setNumRedundantNodes(getMaxNumRedundantNodes());
        findNodeCall->setNumSiblings(getMaxNumSiblings());
        findNodeCall->setLength(FINDNODECALL_L(call));

        RECORD_STATS(numFindNodeSent++;
            bytesFindNodeSent += findNodeCall->byteLength());

        sendUdpRpcCall(handle, findNodeCall);
*/
    } else {
        // we're the only node in the network
        state = READY;
        setOverlayReady(true);

        // schedule bucket refresh timer
        cancelEvent(bucketRefreshTimer);
        scheduleAt(simulation.simTime(), bucketRefreshTimer);
    }
}

//-----------------------------------------------------------------------------

void Kademlia::routingInit()
{
    // set join state
    state = INIT;

    setOverlayReady(false);

    // setup comparator
    comparator = new KeyDistanceComparator<KeyXorMetric>( thisNode.getKey() );

    // calculate number of buckets: ( (2^b)-1 ) * ( keylength / b )
    numBuckets = ( (1L << b) - 1L ) * (OverlayKey::getLength() / b );

    // init routing and sibling table
    siblingTable = new KademliaBucket ( s * 5, comparator );

    // initialize pointers
    routingTable.assign(numBuckets, (KademliaBucket*)NULL);

    WATCH_VECTOR(*siblingTable);
    WATCH_VECTOR(routingTable);

    updateTooltip();
    BUCKET_CONSISTENCY(routingInit: end);
}

void Kademlia::routingDeinit()
{
    // delete buckets
    for (uint i = 0; i < routingTable.size(); i++) {
        if (routingTable[i] != NULL) {
            delete routingTable[i];
            routingTable[i] = NULL;
        }
    }
    routingTable.clear();

    // clear sibling table
    if (siblingTable != NULL) {
        delete siblingTable;
        siblingTable = NULL;
    }

    if (comparator != NULL) {
        delete comparator;
        comparator = NULL;
    }
}

int Kademlia::getMaxNumSiblings()
{
    return s;
}

int Kademlia::getMaxNumRedundantNodes()
{
    return k;
}

int Kademlia::routingBucketIndex(const OverlayKey& key)
{
    // calculate XOR distance
    OverlayKey delta = key ^ getThisNode().getKey();

    // find first subinteger that is not zero...
    int i;
    for (i = key.getLength() - b; i >= 0 && delta.getBitRange(i, b) == 0; i -= b)
        ;
    if (i < 0)
        return -1;

    assert((i / b) * ((1 << b) - 1) + (delta.getBitRange(i, b) - 1) ==
           OverlayKey::getLength() -
               getThisNode().getKey().sharedPrefixLength(key, b) - 1);

    return (i / b) * ((1 << b) - 1) + (delta.getBitRange(i, b) - 1);
}

KademliaBucket* Kademlia::routingBucket(const OverlayKey& key, bool ensure)
{
    // get bucket index
    int num = routingBucketIndex(key);
    if (num < 0)
        return NULL;

    // get bucket and allocate if necessary
    KademliaBucket* bucket = routingTable[ num ];
    if (bucket == NULL && ensure)
        bucket = routingTable[ num ] = new KademliaBucket( k, comparator );

    // return bucket
    return bucket;
}

bool Kademlia::routingAdd(const NodeHandle& handle, bool isAlive, simtime_t rtt)
{
    BUCKET_CONSISTENCY(routingAdd: start);
    // never add unspecified node handles
    if (handle.isUnspecified() || handle.key == getThisNode().getKey() )
        return false;

    // bucket index
    KademliaBucket::iterator i;
    bool result = false;

    bool needsRtt = (activePing && ((rtt == MAXTIME) ? true : false));

    // convert node handle
    KademliaBucketEntry kadHandle = handle;
    kadHandle.setRtt(rtt);
    kadHandle.setLastSeen(simulation.simTime());

    /* check if node is already a sibling -----------------------------------*/
    if ((i = siblingTable->findIterator(handle.getKey()))
         != siblingTable->end()) {
        // not alive? -> do not change routing information
        if (isAlive) {
            if (kadHandle.getRtt() != i->getRtt()) {
                siblingTable->setLastUpdate(simTime());
                if (kadHandle.getRtt() == MAXTIME)
                    kadHandle.setRtt(i->getRtt());
            }
            // refresh sibling
            (*i) = kadHandle;
        }
        BUCKET_CONSISTENCY(routingAdd: node is sibling);
        return true;
    }

    /* check if node is already in a bucket ---------------------------------*/
    KademliaBucket* bucket = routingBucket(handle.getKey(), false);
    if (bucket != NULL && (i = bucket->findIterator(handle.getKey() ) )
            != bucket->end() ) {
        // not alive? -> do not change routing information
        if (isAlive) {
            if (kadHandle.getRtt() == MAXTIME) {
                kadHandle.setRtt(i->getRtt());
            }
            // remove old handle
            bucket->erase(i);
            // re-add to tail
            bucket->push_back(kadHandle);

            bucket->setLastUpdate(simTime());

            if (needsRtt && (kadHandle.getRtt() == MAXTIME))
                pingNode(handle);
        }
        BUCKET_CONSISTENCY(routingAdd: node is in bucket);
        return true;
    }

    /* check if node can be added to the sibling list -----------------------*/
    if (siblingTable->isAddable(handle) ) {

        bool finished = false;
        int siblingPos = -1;

        // check if sibling list is full so a handle is preemted from the list
        if (siblingTable->isFull()) {
            // get handle thats about to be preempted
            KademliaBucketEntry oldHandle = siblingTable->back();
            assert(oldHandle.key != kadHandle.key);
            // add handle to the sibling list
            siblingPos = siblingTable->add(kadHandle);

            // change, so that the preempted handle is added to a bucket
            kadHandle = oldHandle;

            // return always true, since the handle has been added
            result = true;
        } else {
            // simply add the handle and stop
            siblingPos = siblingTable->add(kadHandle);

            // don't need to add kadHandle also to regular buckets
            finished = true;
        }
        assert(siblingPos > -1);

        // ping new siblings
        if ((pingNewSiblings && !isAlive) || needsRtt) {
            pingNode(handle);
        }

        siblingTable->setLastUpdate(simTime());

        updateTooltip();

        // call update() for real siblings
        if (siblingPos < getMaxNumSiblings()) {
            if (siblingTable->size() > (uint)getMaxNumSiblings()) {
                // removed old sibling
                NodeHandle& removedSibling = siblingTable->at(getMaxNumSiblings());
                deleteOverlayNeighborArrow(removedSibling);
                callUpdate(removedSibling, false);
            }
            // new sibling
            showOverlayNeighborArrow(handle, false,
                                     "m=m,50,100,50,100;o=green,1");
            callUpdate(handle, true);
        }

        if (finished) {
            BUCKET_CONSISTENCY(routingAdd: node is now sibling);
            return true;
        }
    }

    /* add node to the appropriate bucket, if not full ---------------------*/
    bucket = routingBucket(kadHandle.getKey(), true);
    if (!bucket->isFull()) {
        EV << "Kademlia::routingAdd(): Adding new node " << kadHandle
           << " to bucket " << routingBucketIndex(kadHandle.getKey()) << endl;
        bucket->push_back(kadHandle);
        bucket->setLastUpdate(simTime());
        result = true;
    } else if (isAlive) {
        // save candidate, ping head,
        KademliaBucket::iterator it = bucket->begin();
        while (it != bucket->end() &&
               replacementCache.find(*it) != replacementCache.end()) {
            ++it;
        }
        if (it != bucket->end()) {
            replacementCache.insert(std::make_pair(*it, kadHandle));
            pingNode(*it, -1,-1, NULL, NULL, NULL, -1,
                     UDP_TRANSPORT, true);
        }
    }

    if (result && needsRtt)
        pingNode(handle);

    BUCKET_CONSISTENCY(routingAdd: end);
    return result;
}

bool Kademlia::routingRemove(const OverlayKey& key)
{
    return routingTimeout(key, true);
}

bool Kademlia::routingTimeout(const OverlayKey& key, bool immediately)
{
    BUCKET_CONSISTENCY(routingTimeout: start);
    // key unspecified? yes -> ignore
    if (key.isUnspecified())
        return false;

    // bucket index
    KademliaBucket::iterator i;

    /* check if the node is one of the siblings -----------------------------*/
    if ((i = siblingTable->findIterator(key)) != siblingTable->end()) {

        i->incStaleCount();

        if (i->getStaleCount() > maxStaleCount || immediately) {
            // remove from sibling table
            NodeHandle oldSibling = *i;
            siblingTable->erase(i);

            if (siblingTable->size() < (uint)getMaxNumSiblings()) {
                // no new replacement sibling
                deleteOverlayNeighborArrow(oldSibling);
                callUpdate(oldSibling, false);
            } else if (comparator->compare(oldSibling.key,
                               siblingTable->at(getMaxNumSiblings() - 1).key) < 0) {
                // failed sibling was replaced by next closest node in siblingTable
                deleteOverlayNeighborArrow(oldSibling);
                callUpdate(oldSibling, false);

                showOverlayNeighborArrow(siblingTable->at(getMaxNumSiblings() - 1),
                                         false, "m=m,50,100,50,100;o=green,1");
                callUpdate(siblingTable->at(getMaxNumSiblings() - 1), true);
            }

            updateTooltip();

            // lost last sibling?
            if (siblingTable->size() == 0) {
                joinOverlay();
                return true;
            }

            BUCKET_CONSISTENCY(routingTimeout: is sibling);

            // try to refill with new closest contact
            refillSiblingTable();

            return true;
        }
    }

    /* check if node is already in a bucket ---------------------------------*/
    KademliaBucket* bucket = routingBucket(key, false);
    if (bucket != NULL && (i = bucket->findIterator(key) ) != bucket->end() ) {

        i->incStaleCount();
        std::map<NodeHandle, NodeHandle>::iterator it
                    = replacementCache.find(*i);
        if (i->getStaleCount() > maxStaleCount ||
            it != replacementCache.end() || immediately) {
            // remove from routing table
            bucket->erase(i);
        }
        if (it != replacementCache.end()) {
            routingAdd(it->second, true);
            nodesReplaced++;
            //EV << "node replaced" << endl;
        }
        BUCKET_CONSISTENCY(routingTimeout: is in bucket);
        return true;
    }
    BUCKET_CONSISTENCY(routingTimeout: end);
    return false;
}

void Kademlia::refillSiblingTable()
{
    if (siblingTable->size() == 0 ||
        siblingTable->isFull())
        return;

    int index = routingBucketIndex(siblingTable->back().key) - 1;
    assert(index > 0);

    while ((routingTable[index] == NULL ||
            routingTable[index]->empty()) &&
            index < (int)(OverlayKey::getLength() - 1)) {
        index++;
    }
    if (index < (int)OverlayKey::getLength() &&
            routingTable[index] != NULL && routingTable[index]->size()) {
        KademliaBucket sortedBucket(k, comparator);
        for (uint i = 0; i < routingTable[index]->size(); ++i)
            sortedBucket.add(routingTable[index]->at(i));
        siblingTable->add(sortedBucket.front());
        // no need to callUpdate(), because s < siblingTable->size(), thus
        // new sibling talbe entry is no real sibling
        routingTable[index]->
        erase(routingTable[index]->
              findIterator(sortedBucket.front().key));
        assert(siblingTable->isFull());
        BUCKET_CONSISTENCY(routingTimeout: end refillSiblingTable());
    }
}

void Kademlia::setBucketUsage(const OverlayKey& key)
{
    KademliaBucket* bucket = routingBucket(key, false);

    if (bucket)
        bucket->setLastUsage(simTime());
}

bool Kademlia::isSiblingFor(const NodeHandle& node, const OverlayKey& key,
                            int numSiblings, bool* err)
{
    if (key.isUnspecified())
        error("Kademlia::isSiblingFor(): key is unspecified!");

    if (state != READY) {
        *err = true;
        return false;
    }

    if (numSiblings > getMaxNumSiblings()) {
        opp_error("Kademlia::isSiblingFor(): numSiblings too big!");
    }

    // set default number of siblings to consider
    if (numSiblings == -1) {
        numSiblings = getMaxNumSiblings();
    }

    if (numSiblings == 0) {
        *err = false;
        return (node.getKey() == key);
    }

    if (siblingTable->isEmpty()) {
        *err = false;
        return true;
    }

    if ((thisNode.key ^ key) > (thisNode.key ^ siblingTable->back().key)) {
        *err = true;
        return false;
    }

    KeyDistanceComparator<KeyXorMetric>* comp =
        new KeyDistanceComparator<KeyXorMetric>(key);

    // create result vector
    NodeVector* result = new NodeVector(numSiblings, comp);

    for (KademliaBucket::iterator i=siblingTable->begin();
         i != siblingTable->end(); i++) {
        result->add( *i);
    }

    // add local node
    result->add(thisNode);

    *err = false;
    delete comp;

    if (result->contains(node.key)) {
        delete result;
        return true;
    } else {
        delete result;
        return false;
    }
}

bool Kademlia::handleFailedNode(const TransportAddress& failed)
{
    assert(!failed.isUnspecified());
    //std::cout << "Kademlia::handleFailedNode()" << std::endl;
    KademliaBucket::iterator i;
    // check sibling table
    for (i = siblingTable->begin(); i != siblingTable->end(); ++i) {
        if (failed == *i) break;
    }

    if (i != siblingTable->end()) {
        // remove from sibling table
        NodeHandle oldSibling = *i;
        siblingTable->erase(i);

        if (siblingTable->size() < (uint)getMaxNumSiblings()) {
            // no new replacement sibling
            deleteOverlayNeighborArrow(oldSibling);
            callUpdate(oldSibling, false);
        } else if (comparator->compare(oldSibling.key,
                           siblingTable->at(getMaxNumSiblings() - 1).key) < 0) {
            // failed sibling was replaced by next closest node in siblingTable
            deleteOverlayNeighborArrow(oldSibling);
            callUpdate(oldSibling, false);

            showOverlayNeighborArrow(siblingTable->at(getMaxNumSiblings() - 1),
                                     false, "m=m,50,100,50,100;o=green,1");
            callUpdate(siblingTable->at(getMaxNumSiblings() - 1), true);
        }

        updateTooltip();

        // try to refill with new closest contact
        refillSiblingTable();
    } else {
        // check buckets
        uint m;
        for (m = 0; m < routingTable.size(); ++m) {
            if (routingTable[m] != NULL) {
                for (i = routingTable[m]->begin(); i != routingTable[m]->end();
                     ++i) {
                    if (failed == *i) {
                        // remove from routing table
                        routingTable[m]->erase(i);
                        return (siblingTable->size() != 0);
                    }
                }
            }
        }
    }
    return (siblingTable->size() != 0);
}

bool Kademlia::forwardMessageRecursive(const TransportAddress& dest,
        BaseRouteMessage* msg)
{
}

NodeVector* Kademlia::findNode(const OverlayKey& key, int numRedundantNodes,
                               int numSiblings, BaseOverlayMessage* msg)
{
    if ((numRedundantNodes > getMaxNumRedundantNodes()) || (numSiblings
            > getMaxNumSiblings())) {

        opp_error("(Kademlia::findNode()) numRedundantNodes or numSiblings "
                  "too big!");
    }

    if (numRedundantNodes < 2) {
        throw new cRuntimeError("Kademlia::findNode(): For Kademlia "
                                "redundantNodes must be at least 2 "
                                "and lookupMerge should be true!");
    }

    // create temporary comparator
    KeyDistanceComparator<KeyXorMetric>* comp =
        new KeyDistanceComparator<KeyXorMetric>( key );

    // select result set size
    bool err;
    int resultSize = isSiblingFor(thisNode, key, numSiblings, &err) ?
        (numSiblings ? numSiblings : 1) : numRedundantNodes;

    assert(numSiblings || numRedundantNodes);

    NodeVector* result = new NodeVector(resultSize, comp);

    NodeRttVector* resultProx = NULL;
    KeyDistanceComparator<KeyPrefixMetric>* compProx = NULL;

    // add items from buckets
    int mainIndex = routingBucketIndex(key);

    for (int i = 1; !result->isFull() && i < numBuckets * 3; i++) {
        int index = mainIndex + (((i & 1) == 1) ? -1 : 1) * (i / 2);
        if (index < 0 || index >= numBuckets)
            continue;

        // add bucket to result vector
        KademliaBucket* bucket = routingTable[index];
        if (bucket!=NULL) {
            for (KademliaBucket::iterator i=bucket->begin(); i!=bucket->end(); i++) {
                result->add(*i);
            }
        }
    }

    // add nodes from sibling table
    for (KademliaBucket::iterator i = siblingTable->begin();
         i != siblingTable->end(); i++) {
        result->add(*i);
    }

    // add local node
    result->add(thisNode);

    delete comp;
    delete compProx;
    delete resultProx;

    return result;
}

//-----------------------------------------------------------------------------


void Kademlia::handleTimerEvent(cMessage* msg)
{
    if (msg == bucketRefreshTimer) {
        handleBucketRefreshTimerExpired();
    }
}

void Kademlia::handleUDPMessage(BaseOverlayMessage* msg)
{
    OverlayCtrlInfo* ctrlInfo =
            check_and_cast<OverlayCtrlInfo*>(msg->removeControlInfo());
    KademliaRoutingInfoMessage* kadRoutingInfoMsg = check_and_cast<
            KademliaRoutingInfoMessage*>(msg);

    routingAdd(kadRoutingInfoMsg->getSrcNode(), true);

    for (uint i = 0; i < kadRoutingInfoMsg->getNextHopsArraySize(); i++) {
        routingAdd(kadRoutingInfoMsg->getNextHops(i), false);
    }

    delete ctrlInfo;
    delete msg;
}

//In Kademlia this method is used to maintain the routing table.
bool Kademlia::handleRpc(BaseCallMessage* msg)
{
    RPC_SWITCH_START( msg )
    RPC_ON_CALL( Ping ) {
        // add active node
        routingAdd( _PingCall->getSrcNode(), true );
        break;
    }
    RPC_ON_CALL(FindNode)
    {
        // add active node
        routingAdd(_FindNodeCall->getSrcNode(), true);
        break;
    }
    RPC_SWITCH_END( )
    return false;
}

//In Kademlia this method is used to maintain the routing table.
void Kademlia::handleRpcResponse(BaseResponseMessage* msg,
                                 cPolymorphic* context, int rpcId,
                                 simtime_t rtt)
{
    // add node that reponded
    routingAdd(msg->getSrcNode(), true, rtt);

    RPC_SWITCH_START(msg)
        RPC_ON_RESPONSE(Ping) {
            if (state == INIT) {
                // schedule bucket refresh timer
                cancelEvent(bucketRefreshTimer);
                scheduleAt(simulation.simTime(), bucketRefreshTimer);
                state = JOIN;
            }

            // add active node
            replacementCache.erase(_PingResponse->getSrcNode());
            //routingAdd(_PingResponse->getSrcNode(), true , rtt);
        }
        RPC_ON_RESPONSE(FindNode)
        {
            // add active node
            if (defaultRoutingType == SEMI_RECURSIVE_ROUTING ||
                defaultRoutingType == FULL_RECURSIVE_ROUTING ||
                defaultRoutingType == RECURSIVE_SOURCE_ROUTING)
                rtt = MAXTIME;
            //routingAdd(_FindNodeResponse->getSrcNode(), true, rtt);
            setBucketUsage(_FindNodeResponse->getSrcNode().key);

            // add inactive nodes
            for (uint i=0; i<_FindNodeResponse->getClosestNodesArraySize(); i++)
                routingAdd(_FindNodeResponse->getClosestNodes(i), false);
            break;
        }
    RPC_SWITCH_END()
}

// In Kademlia this method is used to maintain the routing table.
void Kademlia::handleRpcTimeout(BaseCallMessage* msg,
                                const TransportAddress& dest,
                                cPolymorphic* context, int rpcId,
                                const OverlayKey& destKey)
{
    if (dest.isUnspecified()) return;

    const NodeHandle& handle = dynamic_cast<const NodeHandle&>(dest);

    RPC_SWITCH_START(msg)
        if (state == JOIN) {
            joinOverlay();
        }
    RPC_ON_CALL(Ping) {
        routingTimeout(handle.key);

        // remove node from replacementCache
        std::map<NodeHandle, NodeHandle>::iterator it
            = replacementCache.find(handle);
        if (it != replacementCache.end()) {
            replacementCache.erase(it);
        }

        break;
    }
    RPC_ON_CALL(FindNode) {
        routingTimeout(handle.key);
        setBucketUsage(handle.key);
        break;
    }
    RPC_SWITCH_END()
}

void Kademlia::lookupFinished(bool isValid)
{
    if (state != READY) {
        cancelEvent(bucketRefreshTimer);
        scheduleAt(simulation.simTime(), bucketRefreshTimer);

        state = READY;
        setOverlayReady(true);
    }
}

// handle a expired stabilize timer
/*
void Kademlia::handleStabilizeTimerExpired()
{
    // refresh sibling table
    for (int i = 0; i < (int)siblingTable->size(); i++) {
        NodeHandle& handle = siblingTable->at(i);
        FindNodeCall* findNodeCall = new FindNodeCall();
        findNodeCall->setLookupKey(handle.getKey());
        findNodeCall->setSrcNode(getThisNode());
        findNodeCall->setNumRedundantNodes(getMaxNumRedundantNodes());
        findNodeCall->setNumSiblings(getMaxNumSiblings());
        findNodeCall->setLength(FINDNODECALL_L(call));

        RECORD_STATS(numFindNodeSent++;
                     bytesFindNodeSent += findNodeCall->byteLength());

        sendUdpRpcCall(handle, findNodeCall);
    }

    // schedule next stabilization process
    cancelEvent(stabilizeTimer);
    scheduleAt(simulation.simTime() + stabilizeDelay, stabilizeTimer);
}
*/

// handle a expired bucket refresh timer
void Kademlia::handleBucketRefreshTimerExpired()
{
    // refresh buckets
    if (state != READY || (((simTime() - siblingTable->getLastUsage()) >
                            minSiblingTableRefreshInterval)))
        //std::cout << thisNode << ": lookup" << std::endl;
        createLookup()->lookup(getThisNode().getKey() +
                               OverlayKey::ONE, 0, 0, 0,
                               new KademliaLookupListener(this));

    if (state == READY && siblingTable->size()) {
        uint diff = OverlayKey::getLength() - getThisNode().getKey().
                        sharedPrefixLength(siblingTable->front().key);
        for (uint i = OverlayKey::getLength() - 1; i >= (diff - 1); i--) {
            if ((routingTable[i] == NULL) ||
                 ((simTime() - routingTable[i]->getLastUsage()) >
                   minBucketRefreshInterval)) {
                //EV << i << "bla";
                createLookup()->lookup(getThisNode().getKey() ^
                    (OverlayKey::ONE << i), 0, 0, 0,
                    new KademliaLookupListener(this));
            }
        }
        // schedule next bucket refresh process
        cancelEvent(bucketRefreshTimer);
        scheduleAt(simulation.simTime() +
                   (min(minSiblingTableRefreshInterval,
                        minBucketRefreshInterval) / 10.0),
                   bucketRefreshTimer);
    }
}

//virtual public: xor metric
OverlayKey Kademlia::distance(const OverlayKey& x, const OverlayKey& y) const
{
    return x^y;
}

void Kademlia::updateTooltip()
{
    if (ev.isGUI()) {
        std::stringstream ttString;

        // show our nodeId in a tooltip
        ttString << "This: " << thisNode << endl << "Siblings: "
                 << siblingTable->size();

        parentModule()->parentModule()->displayString().
        setTagArg("tt", 0, ttString.str().c_str());
        parentModule()->displayString().
        setTagArg("tt", 0, ttString.str().c_str());
        displayString().setTagArg("tt", 0, ttString.str().c_str());
    }
}

