//
// Copyright (C) 2006 Institut fuer Telematik, Universitaet Karlsruhe (TH)
//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public License
// as published by the Free Software Foundation; either version 2
// of the License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

#ifndef __KADEMLIA_NODE_HANDLE_H
#define __KADEMLIA_NODE_HANDLE_H

#include <NodeHandle.h>
#include <NodeVector.h>

/**
 * Kademlia BucketEntry
 *
 * @author Sebastian Mies
 * @author Bernhard Heep
 */
class KademliaBucketEntry : public NodeHandle
{
public:
    /**
     * Constructs an unspecified NodeHandle
     */
    KademliaBucketEntry()
    : NodeHandle()
    {
        staleCount = 0;
        rtt = MAXTIME;
    }

    KademliaBucketEntry(const NodeHandle& handle, simtime_t rtt = MAXTIME)
    : NodeHandle(handle)
    {
        staleCount = 0;
        this->rtt = rtt;
    }

    /**
     * Complete constructor.
     *
     * @param key The OverlayKey
     * @param ip The IPvXAddress
     * @param port The UDP-Port
     */
    KademliaBucketEntry(const OverlayKey& key,
                        const IPvXAddress& ip,
                        int port)
    : NodeHandle(key, ip, port)
    {
        staleCount = 0;
        rtt = MAXTIME;
    }

    inline simtime_t getRtt() const { return rtt; }

    inline void setRtt(simtime_t rtt) { this->rtt = rtt; }

    inline bool operator<(const KademliaBucketEntry& entry) const
    {
        return rtt < entry.rtt;
    }

    inline uint8_t getStaleCount() const { return staleCount; }

    inline void setStaleCount(uint8_t staleCount)
    {
        this->staleCount = staleCount;
    }

    inline void resetStaleCount() { this->setStaleCount(0); }

    inline void incStaleCount() { this->staleCount++; }

    inline void setLastSeen(simtime_t lastSeen) { this->lastSeen = lastSeen; }

    inline simtime_t getLastSeen() { return this->lastSeen; }

private:

    uint8_t staleCount;
    simtime_t lastSeen;
    simtime_t rtt;

    friend std::ostream& operator<<(std::ostream& os,
                                    const KademliaBucketEntry& n)
    {
        os << (NodeHandle)n << " " << n.rtt;
        return os;
    };
};

/**
 * Class for extracting the relevant OverlayKey from a type used as template
 * parameter T_key for NodeVector<T, T_key> - Version for KademliaBucketEntry.
 *
 * @author Sebastian Mies
 * @author Felix M. Palmen
 */
template <> struct KeyExtractor<KademliaBucketEntry>
{
    static const OverlayKey& key(const KademliaBucketEntry& entry)
    {
        return entry.key;
    };
};

template <> struct RttExtractor<KademliaBucketEntry>
{
    static simtime_t rtt(const KademliaBucketEntry& entry)
    {
        return entry.getRtt();
    };
};


#endif
