//
// Copyright (C) 2007 Institut fuer Telematik, Universitaet Karlsruhe (TH)
//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public License
// as published by the Free Software Foundation; either version 2
// of the License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
//

/**
 * @file Koorde.cc
 * @author Jochen Schenk
 */

#include <IPAddressResolver.h>
#include <IPvXAddress.h>
#include <InterfaceTable.h>
#include <IPv4InterfaceData.h>
#include <GlobalStatistics.h>

#include "Koorde.h"

using namespace std;

Define_Module(Koorde);

void Koorde::initializeOverlay(int stage)
{
    // because of IPAddressResolver, we need to wait until interfaces
    // are registered, address auto-assignment takes place etc.
    if(stage != MIN_STAGE_OVERLAY)
        return;

    // fetch some parameters
    deBruijnDelay = par("deBruijnDelay");
    deBruijnListSize = par("deBruijnListSize");
    shiftingBits = par("shiftingBits");
    useOtherLookup = par("useOtherLookup");
    useSucList = par("useSucList");

    // init flags
    sucDeBruijnNodeSet = false;
    preDeBruijnNodeSet = false;
    breakLookup = false;
    firstTimer = true;
    preOrSuc = false;

    // some local variables
    failedDBResponses = 0;
    deBruijnRetryTime = 20;
    deBruijnNumber = 0;
    deBruijnNodes = new NodeHandle[deBruijnListSize];

    // statistics
    deBruijnCount = 0;
    deBruijnBytesSent = 0;

    // add some watches
    WATCH(sucDeBruijnNodeSet);
    WATCH(deBruijnNumber);
    WATCH(deBruijnNode);

    //timer Messages
    deBruijn_timer = new cMessage("deBruijn_timer");

    Chord::initializeOverlay(stage);
}

void Koorde::changeState(int toState)
{
    Chord::changeState(toState);

    switch(state) {
    case INIT:
        // init de Bruijn nodes
        deBruijnNode = NodeHandle::UNSPECIFIED_NODE;
        for (int i=0; i < deBruijnListSize; i++)
            deBruijnNodes[i] = NodeHandle::UNSPECIFIED_NODE;
        updateTooltip();
        break;
    case READY:
        // init de Bruijn Protocol
        cancelEvent(deBruijn_timer);
        scheduleAt(simulation.simTime() + deBruijnDelay, deBruijn_timer);

        // since we don't need the fixfingers protocol in Koorde cancel timer
        cancelEvent(fixfingers_timer);
        break;
    default:
        break;
    }

}

void Koorde::handleTimerEvent(cMessage* msg)
{
    // catch DeBruijn timer
    if (msg->isName("deBruijn_timer"))
        handleDeBruijnTimerExpired(msg);
    else if (msg->isName("fixfingers_timer"))
        handleFixFingersTimerExpired(msg);
    else
        Chord::handleTimerEvent(msg);
}


void Koorde::handleDeBruijnTimerExpired(cMessage* msg)
{
    if (state == READY) {
        OverlayKey lookup = thisNode.key << shiftingBits;

        if (lookup.isBetweenR(thisNode.key, successorList->getSuccessor().key)
                || successorList->isEmpty()) {

            int sucNum = successorList->getSize();
            if (sucNum > deBruijnListSize)
                sucNum = deBruijnListSize;

            deBruijnNode = thisNode;
            for (int i = 0; i < sucNum; i++) {
                deBruijnNodes[i] = successorList->getSuccessor(i);
                deBruijnNumber = i+1;
            }

            updateTooltip();
        } else {
            DeBruijnCall* call = new DeBruijnCall("DeBruijnCall");
            call->setDestKey(lookup);
            call->setLength(DEBRUIJNCALL_L(call));

            // if we don't have a de Bruijn node we use our successor
            // or our predecessor as a springboard if they have their
            // flags set or we didn't have a chance to get it
            if (deBruijnNode.isUnspecified()) {
                if (firstTimer || (sucDeBruijnNodeSet && !preOrSuc)) {
                    sendRouteRpcCall(OVERLAY_COMP,
                                     successorList->getSuccessor(),
                                     call->getDestKey(), call, NULL,
                                     DEFAULT_ROUTING, deBruijnRetryTime);
                    firstTimer = false;
                } else if (preOrSuc && preDeBruijnNodeSet) {
                    sendRouteRpcCall(OVERLAY_COMP,
                                     predecessorNode,
                                     call->getDestKey(), call, NULL,
                                     DEFAULT_ROUTING, deBruijnRetryTime);
                } else {
                    sendRouteRpcCall(OVERLAY_COMP, call->getDestKey(),
                                     call, NULL, DEFAULT_ROUTING,
                                     deBruijnRetryTime);

                    EV << "[Koorde::handleDeBruijnTimerExpired() @ " << thisNode.ip
                       << " (" << thisNode.key.toString(16) << ")]\n"
                       << "    Have to route DeBruijnCall to destination key"
                       << endl;
                }
            } else {
                sendRouteRpcCall(OVERLAY_COMP, deBruijnNode,
                                 call->getDestKey(), call, NULL,
                                 DEFAULT_ROUTING, deBruijnRetryTime);
            }

        }

        cancelEvent(deBruijn_timer);
        scheduleAt(simulation.simTime() + deBruijnDelay, deBruijn_timer);
    }
}


void Koorde::handleFixFingersTimerExpired(cMessage* msg)
{
    // just in case not all timers from Chord code could be canceled
}


void Koorde::handleUDPMessage(BaseOverlayMessage* msg)
{
    Chord::handleUDPMessage(msg);
}


bool Koorde::handleRpc(BaseCallMessage* msg)
{
    if (state != READY) {
        //EV << "[Koorde::handleRpc() @ " << thisNode.ip
        //   << " (" << thisNode.key.toString(16) << ")]\n"
        //   << "    Received RPC call and state != READY!"
        //   << endl;
        //cout << "RPC in non-Ready state received!" << endl;
        return false;
    }

    // delegate messages
    RPC_SWITCH_START( msg )
    // RPC_DELEGATE( <messageName>[Call|Response], <methodToCall> )
    RPC_DELEGATE( Join, rpcJoin );
    RPC_DELEGATE( Notify, rpcNotify );
    RPC_DELEGATE( Stabilize, rpcStabilize );
    RPC_DELEGATE( Fixfingers, rpcFixfingers );
    RPC_DELEGATE( DeBruijn, handleRpcDeBruijnRequest );
    RPC_SWITCH_END( )

    return RPC_HANDLED;
}

void Koorde::handleRpcResponse(BaseResponseMessage* msg,
                               cPolymorphic* context,
                               int rpcId, simtime_t rtt)
{
    RPC_SWITCH_START(msg)
    RPC_ON_RESPONSE( Join ) {
        handleRpcJoinResponse(_JoinResponse);
        EV << "[Koorde::handleRpcResponse() @ " << thisNode.ip
           << " (" << thisNode.key.toString(16) << ")]\n"
           << "    Join RPC Response received: id=" << rpcId
           << "\n    msg=" << *_JoinResponse << " rtt=" << rtt
           << endl;
        break;
    }
    RPC_ON_RESPONSE( Notify ) {
        handleRpcNotifyResponse(_NotifyResponse);
        EV << "[Koorde::handleRpcResponse() @ " << thisNode.ip
           << " (" << thisNode.key.toString(16) << ")]\n"
           << "    Notify RPC Response received: id=" << rpcId
           << "\n    msg=" << *_NotifyResponse << " rtt=" << rtt
           << endl;
        break;
    }
    RPC_ON_RESPONSE( Stabilize ) {
        handleRpcStabilizeResponse(_StabilizeResponse);
        EV << "[Koorde::handleRpcResponse() @ " << thisNode.ip
           << " (" << thisNode.key.toString(16) << ")]\n"
           << "    Stabilize RPC Response received: id=" << rpcId
           << "\n    msg=" << *_StabilizeResponse << " rtt=" << rtt
           << endl;
        break;
    }
    RPC_ON_RESPONSE( Fixfingers ) {
        handleRpcFixfingersResponse(_FixfingersResponse);
        EV << "[Koorde::handleRpcResponse() @ " << thisNode.ip
           << " (" << thisNode.key.toString(16) << ")]\n"
           << "    Fixfingers RPC Response received: id=" << rpcId
           << "\n    msg=" << *_FixfingersResponse << " rtt=" << rtt
           << endl;
        break;
    }
    RPC_ON_RESPONSE( DeBruijn ) {
        handleRpcDeBruijnResponse(_DeBruijnResponse);
        EV << "[Koorde::handleRpcResponse() @ " << thisNode.ip
           << " (" << thisNode.key.toString(16) << ")]\n"
           << "    DeBruijn RPC Response received: id=" << rpcId
           << "\n    msg=" << *_DeBruijnResponse << " rtt=" << rtt
           << endl;
        break;
    }
    RPC_SWITCH_END( )
}

void Koorde::handleRpcTimeout(BaseCallMessage* msg, const TransportAddress& dest,
                              cPolymorphic* context, int rpcId,
                              const OverlayKey& destKey)
{
    RPC_SWITCH_START(msg)
    RPC_ON_CALL( Join ) {
        EV << "[Koorde::handleRpcTimeout() @ " << thisNode.ip
           << " (" << thisNode.key.toString(16) << ")]\n"
           << "    Join RPC Call timed out: id=" << rpcId
           << "\n    msg=" << *_JoinCall
           << endl;
        break;
    }
    RPC_ON_CALL( Notify ) {
        EV << "[Koorde::handleRpcTimeout() @ " << thisNode.ip
           << " (" << thisNode.key.toString(16) << ")]\n"
           << "    Notify RPC Call timed out: id=" << rpcId
           << "\n    msg=" << *_NotifyCall
           << endl;
        break;
    }
    RPC_ON_CALL( Stabilize ) {
        EV << "[Koorde::handleRpcTimeout() @ " << thisNode.ip
           << " (" << thisNode.key.toString(16) << ")]\n"
           << "    Stabilize RPC Call timed out: id=" << rpcId
           << "\n    msg=" << *_StabilizeCall
           << endl;
        break;
    }
    RPC_ON_CALL( Fixfingers ) {
        EV << "[Koorde::handleRpcTimeout() @ " << thisNode.ip
           << " (" << thisNode.key.toString(16) << ")]\n"
           << "    Fixfingers RPC Call timed out: id=" << rpcId
           << "\n    msg=" << *_FixfingersCall
           << endl;
        break;
    }
    RPC_ON_CALL( DeBruijn ) {
        handleDeBruijnTimeout(_DeBruijnCall);
        EV << "[Koorde::handleRpcTimeout() @ " << thisNode.ip
           << " (" << thisNode.key.toString(16) << ")]\n"
           << "    DeBruijn RPC Call timed out: id=" << rpcId
           << "\n    msg=" << *_DeBruijnCall
           << endl;
        break;
    }
    RPC_SWITCH_END( )
}


void Koorde::handleRpcJoinResponse(JoinResponse* joinResponse)
{
    Chord::handleRpcJoinResponse(joinResponse);

    // has to be canceled in Koorde
    cancelEvent(fixfingers_timer);

    // immediate deBruijn protocol
    cancelEvent(deBruijn_timer);
    scheduleAt(simulation.simTime(), deBruijn_timer);
}

void Koorde::handleRpcStabilizeResponse(StabilizeResponse* stabilizeResponse)
{
    // our successor seems to be alive
    missingSuccessorStabResponses = 0;

    // fetch the successor's predecessor
    NodeHandle predecessor = stabilizeResponse->getPreNode();

    // is the successor's predecessor a new successor for this node?
    if (successorList->isEmpty() ||
            predecessor.key.isBetween(thisNode.key,
                                      successorList->getSuccessor().key)) {
        // add the successor's predecessor to the successor list
        successorList->addSuccessor(predecessor);
        updateTooltip();
    }

    // compile NOTIFY RPC
    NotifyCall* notifyCall = new NotifyCall("NotifyCall");

    if (!deBruijnNode.isUnspecified()) {
        notifyCall->setPreDBNodeSet(true);
    } else {
        notifyCall->setPreDBNodeSet(false);
    }

    notifyCall->setLength(NOTIFYCALL_L(notifyCall));
    sendUdpRpcCall(successorList->getSuccessor(), notifyCall);
}


void Koorde::rpcNotify(NotifyCall* call)
{
    // our predecessor seems to be alive
    missingPredecessorStabRequests = 0;

    NodeHandle newPredecessor = call->getSrcNode();

    // is the new predecessor closer than the current one?
    if (predecessorNode.isUnspecified() ||
            newPredecessor.key.isBetween(predecessorNode.key, thisNode.key)) {

	if ((predecessorNode.isUnspecified()) ||
	    (newPredecessor != predecessorNode)) {
	    // send update to application if we've got a new predecessor
	    if (!predecessorNode.isUnspecified()) {
		callUpdate(predecessorNode, false);
	    }
	    callUpdate(newPredecessor, true);

	    // set up new predecessor
	    predecessorNode = newPredecessor;
	}

        preDeBruijnNodeSet = call->getPreDBNodeSet();
        firstTimer = false;
        updateTooltip();
    }

    // needed if the predecessor doesn't change
    if (newPredecessor == predecessorNode)
        preDeBruijnNodeSet = call->getPreDBNodeSet();

    // compile NOTIFY response
    NotifyResponse* notifyResponse = new NotifyResponse("NotifyResponse");

    int sucNum = successorList->getSize();

    if (deBruijnNode.isUnspecified()) {
        notifyResponse->setSucDBNodeSet(false);
    } else {
        notifyResponse->setSucDBNodeSet(true);
    }

    notifyResponse->setSucNum(sucNum);
    notifyResponse->setSucNodeArraySize(sucNum);

    for (int k = 0; k < sucNum; k++) {
        notifyResponse->setSucNode(k, successorList->getSuccessor(k));
    }

    notifyResponse->setLength(NOTIFYRESPONSE_L(notifyResponse));

    sendRpcResponse(call, notifyResponse);
}


void Koorde::handleRpcNotifyResponse(NotifyResponse* notifyResponse)
{
    if (successorList->getSuccessor() != notifyResponse->getSrcNode()) {
        EV << "[Koorde::handleRpcNotifyResponse() @ " << thisNode.ip
           << " (" << thisNode.key.toString(16) << ")]\n"
           << "    The srcNode of the received NotifyResponse is not our current successor"
           << endl;
        return;
    }

    // replace our successor list by our successor's successor list
    successorList->updateList(notifyResponse);

    // store if the de bruijn node of my successor is set, also
    // reset firstTimer flag
    sucDeBruijnNodeSet = notifyResponse->getSucDBNodeSet();
    firstTimer = false;

    updateTooltip();
}



void Koorde::handleRpcDeBruijnRequest(DeBruijnCall* deBruijnCall)
{
    // The key lies between thisNode and its predecessor and
    // because routing the message to the predecessor of a key
    // is near to impossible we set the deBruijnNodes here
    // and the information is as actual as the predecessor pointer
    if (deBruijnCall->getDestKey().isBetweenR(predecessorNode.key,
            thisNode.key)) {
        DeBruijnResponse* deBruijnResponse =
            new DeBruijnResponse("DeBruijnResponse");

        deBruijnResponse->setDBNode(predecessorNode);

        int sucNum = successorList->getSize() + 1;
        deBruijnResponse->setSucNum(sucNum);
        deBruijnResponse->setSucNodeArraySize(sucNum);

        deBruijnResponse->setSucNode(0, thisNode);
        for (int k = 1; k < sucNum; k++) {
            deBruijnResponse->setSucNode(k, successorList->getSuccessor(k-1));
        }
        deBruijnResponse->setLength(DEBRUIJNRESPONSE_L(deBruijnResponse));

        sendRpcResponse(deBruijnCall, deBruijnResponse);
    } else if (deBruijnCall->getDestKey().isBetweenR(thisNode.key,
               successorList->getSuccessor().key)) {
        error("Koorde::handleRpcDeBruijnRequest() - unknown error.");
    } else {
        error("Koorde::handleRpcDeBruijnRequest() - "
              "Request couldn't be delivered!");
    }
}


void Koorde::handleRpcDeBruijnResponse(DeBruijnResponse* deBruijnResponse)
{
    int sucNum = deBruijnResponse->getSucNum();
    if (sucNum > deBruijnListSize)
        sucNum = deBruijnListSize;

    for (int i = 0; i < sucNum; i++) {
        deBruijnNodes[i] = deBruijnResponse->getSucNode(i);
        deBruijnNumber = i+1;
    }

    deBruijnNode = deBruijnResponse->getDBNode();

    updateTooltip();
}

void Koorde::handleDeBruijnTimeout(DeBruijnCall* deBruijnCall)
{
    if (failedDBResponses == stabilizeRetry) {
        failedDBResponses = 0;
        deBruijnNode = NodeHandle::UNSPECIFIED_NODE;
        preOrSuc = !preOrSuc;
    } else {
        failedDBResponses++;
    }

    updateTooltip();
    cancelEvent(deBruijn_timer);
    scheduleAt(simulation.simTime(), deBruijn_timer);
}

NodeVector* Koorde::findNode(const OverlayKey& key,
			    int numRedundantNodes,
			    int numSiblings,
			    BaseOverlayMessage* msg)
{
    NodeVector* nextHop = new NodeVector();
    KoordeFindNodeExtMessage *findNodeExt = NULL;

    if (state != READY)
        return nextHop;

    if (msg != NULL) {
        if (!msg->hasObject("findNodeExt")) {
            findNodeExt = new KoordeFindNodeExtMessage("findNodeExt");

            OverlayKey routeKey = findStartKey(thisNode.key,
                                     successorList->getSuccessor().key, key);
            int matchingBits = findLongestMatch(routeKey, key);
            findNodeExt->setRouteKey(routeKey);
            findNodeExt->setStep(matchingBits + 1);
            findNodeExt->setOtherLookup(false);
            findNodeExt->setLength(KOORDEFINDNODEEXTMESSAGE_L);
            findNodeExt->setSrcNode(thisNode);
            msg->addObject( findNodeExt );
        }
        findNodeExt = (KoordeFindNodeExtMessage*) msg->getObject("findNodeExt");
    }

    if (key.isUnspecified()) {
        error("Koorde::findNode() - direct Messaging is no longer in use.");
    } else if (key.isBetweenR(predecessorNode.key, thisNode.key)) {
        // the message is destined for this node
        nextHop->push_back(thisNode);
    } else if (key.isBetweenR(thisNode.key, successorList->getSuccessor().key)){
        // the message destined for our successor
        nextHop->push_back(successorList->getSuccessor());
    } else {
        // if useOtherLookup is enabled we try to use
        // our successor list to get to the key
        if (useOtherLookup) {
            NodeHandle tmpNode = walkSuccessorList(key);
            if (tmpNode != successorList->getSuccessor(
		                             successorList->getSize()-1)) {
                findNodeExt->setOtherLookup(true);
                nextHop->push_back(tmpNode);
            } else if (findNodeExt->getOtherLookup() == true) {
                nextHop->push_back(tmpNode);
            } else {
                NodeHandle tmpHandle = findDeBruijnHop(key, findNodeExt);
                if (tmpHandle != thisNode || breakLookup) {
                    nextHop->push_back(tmpHandle);
                    breakLookup = false;
                } else {
                    return findNode(key, numRedundantNodes, numSiblings, msg);
                }
            }
        } else {
            // find next hop using either the de Bruijn node and
            // its successors or our own successors
            NodeHandle tmpHandle = findDeBruijnHop(key, findNodeExt);
            if (tmpHandle != thisNode || breakLookup) {
                nextHop->push_back(tmpHandle);
                breakLookup = false;
            } else {
                return findNode(key, numRedundantNodes, numSiblings, msg);
            }
        }
    }
    return nextHop;
}

NodeHandle Koorde::findDeBruijnHop(const OverlayKey& destKey,
                        KoordeFindNodeExtMessage* findNodeExt)
{
    // check if the route key falls in our responsibility or
    // else forward the message to our successor
    if (findNodeExt->getRouteKey().isBetweenR(thisNode.key,
            successorList->getSuccessor().key)) {
        if ((unsigned int)findNodeExt->getStep() > destKey.getLength())
            error("Broose::findDeBruijnHop - Bounding error: "
                  "trying to get non existing bit out of overlay key!");

        // updating the route key
        OverlayKey add = OverlayKey(destKey.getBit(destKey.getLength() -
                                                   findNodeExt->getStep()));
        for (int i = 1; i < shiftingBits; i++) {
            add = (add << 1) + OverlayKey(destKey.getBit(destKey.getLength() - findNodeExt->getStep() - i));
        }

        OverlayKey routeKey = (findNodeExt->getRouteKey()<<shiftingBits) + add;
        findNodeExt->setRouteKey(routeKey);
        findNodeExt->setStep(findNodeExt->getStep() + shiftingBits);

        if(deBruijnNode.isUnspecified()) {
            breakLookup = true;
            if (useSucList)
                return walkSuccessorList(findNodeExt->getRouteKey());
            else
                return successorList->getSuccessor();
        }

        // check if the new route key falls between our
        // de Bruijn node and its successor
        if (deBruijnNumber > 0) {
            if (findNodeExt->getRouteKey().isBetweenR(deBruijnNode.key,
                    deBruijnNodes[0].key)) {
                return deBruijnNode;
            } else {
                // otherwise check if the route key falls between
                // our de Bruijn successors
                NodeHandle nextHop = walkDeBruijnList(findNodeExt->
						      getRouteKey());
                return nextHop;
            }
        } else {
            return deBruijnNode;
        }
    } else {
        breakLookup = true;
        // if optimization is set search the successor list and
        // de bruijn node to find "good" next hop
        if (useSucList) {
            if (!deBruijnNode.isUnspecified()) {
                return walkSuccessorList(findNodeExt->getRouteKey());
            } else {
                NodeHandle tmpHandle = walkSuccessorList(
		                                 findNodeExt->getRouteKey());

                //todo: optimization - check complete deBruijnList
                if (deBruijnNode.key.isBetween(tmpHandle.key,
                                               findNodeExt->getRouteKey())) {
                    return deBruijnNode;
                } else {
                    return tmpHandle;
                }
            }
        } else
            return successorList->getSuccessor();
    }
}


const NodeHandle& Koorde::walkDeBruijnList(const OverlayKey& key)
{
    if (deBruijnNumber == 0)
        return NodeHandle::UNSPECIFIED_NODE;

    for (int i = 0; i < deBruijnNumber-1; i++) {
        if (key.isBetweenR(deBruijnNodes[i].key,deBruijnNodes[i+1].key)) {
            return deBruijnNodes[i];
        }
    }

    return deBruijnNodes[deBruijnNumber-1];
}

const NodeHandle& Koorde::walkSuccessorList(const OverlayKey& key)
{
    for (unsigned int i = 0; i < successorList->getSize()-1; i++) {
        if (key.isBetweenR(successorList->getSuccessor(i).key,
                           successorList->getSuccessor(i+1).key)) {
            return successorList->getSuccessor(i);
        }
    }

    return successorList->getSuccessor(successorList->getSize()-1);
}

void Koorde::updateTooltip()
{
    //
    // Updates the tooltip display strings.
    //

    if (ev.isGUI()) {
        std::stringstream ttString;

        // show our predecessor, successor and de Bruijn node in tooltip
        ttString << "Pred "<< predecessorNode << endl << "This  "
		 << thisNode << endl
		 << "Suc   " << successorList->getSuccessor() << endl
		 << "DeBr " << deBruijnNode << endl;
        ttString << "List ";

        for (unsigned int i = 0; i < successorList->getSize(); i++) {
            ttString << successorList->getSuccessor(i).ip << " ";
        }

        ttString << endl;
        ttString << "DList ";

        for (int i = 0; i < deBruijnNumber; i++) {
            ttString << deBruijnNodes[i].ip << " ";
        }

        ttString << endl;

        parentModule()->parentModule()->displayString().setTagArg(
            "tt", 0, ttString.str().c_str());
        parentModule()->displayString().setTagArg("tt", 0,
                ttString.str().c_str());
        displayString().setTagArg("tt", 0, ttString.str().c_str());

        // draw an arrow to our current successor
        showOverlayNeighborArrow(successorList->getSuccessor(), true,
                                 "m=m,50,0,50,0;o=red,1");
    }
}


void Koorde::finishOverlay()
{
    // destroy self timer messages
    cancelEvent(deBruijn_timer);

    delete deBruijn_timer;

    // statistics
    simtime_t time = globalStatistics->calcMeasuredLifetime(creationTime);

    if(time != 0) {
	globalStatistics->addStdDev("Koorde: Sent DEBRUIJN Messages/s", deBruijnCount / time);
	globalStatistics->addStdDev("Koorde: Sent DEBRUIJN Bytes/s", deBruijnBytesSent / time);
    }

    Chord::finishOverlay();
}

void Koorde::recordOverlaySentStats(BaseOverlayMessage* msg)
{
    Chord::recordOverlaySentStats(msg);

    BaseOverlayMessage* innerMsg;

    if (msg->getType() == OVERLAYROUTE)
        innerMsg = dynamic_cast<BaseOverlayMessage*>(msg->encapsulatedMsg());
    else
        innerMsg = msg;

    switch (innerMsg->getType()) {

    case RPC: {
            if ((dynamic_cast<DeBruijnCall*>(innerMsg) != NULL) ||
                    (dynamic_cast<DeBruijnResponse*>(innerMsg) != NULL)) {
                RECORD_STATS(deBruijnCount++; deBruijnBytesSent +=
                                 msg->byteLength());
            }
            break;
        }
    }
}

OverlayKey Koorde::findStartKey(const OverlayKey& startKey,
				const OverlayKey& endKey,
				const OverlayKey& destKey)
{
    OverlayKey diffKey, newStart, tmpDest, newKey, powKey;
    int nBits;

    if (startKey == endKey)
        return startKey;

    diffKey = endKey - startKey;
    nBits = diffKey.log_2();

    while ((startKey.getLength() - nBits) % shiftingBits != 0) {
        nBits--;
    }

    newStart = startKey >> nBits;
    newStart = newStart << nBits;

    tmpDest = destKey >> (destKey.getLength() - nBits);
    newKey = tmpDest + newStart;

    // is the new constructed route key bigger than our start key return it
    if (newKey.isBetweenR(startKey, endKey)) {
        return newKey;
    }

    // If the part of the destination key smaller than the one of
    // the orginal key add pow(nBits) (this is the first bit where
    // the start key and end key differ) to the new constructed key
    // and check if it's beetween start and end key.

    newKey += powKey.pow2(nBits);
    if(newKey.isBetweenR(startKey, endKey)) {
        return newKey;
    }

    // this part should not be called
    error("Koorde::findStartKey()");
    return OverlayKey::UNSPECIFIED_KEY;
}

int Koorde::findLongestMatch(const OverlayKey& srcKey,
			     const OverlayKey& destKey)
{
    int keyLength = srcKey.getLength();
    bool equal = true;

    for (int i = keyLength; i>0; i--) {
        equal = true;
        for(int j = 1; j <= i; j++) {
            if (srcKey.getBit(i - j) != destKey.getBit(destKey.getLength() - j)) {
                equal = false;
                break;
            }
        }

        if (equal)
            return i;

    }
    return 0;
}

void Koorde::findFriendModules()
{
    successorList = check_and_cast<ChordSuccessorList*>
                    (parentModule()->submodule("successorList"));
}

void Koorde::initializeFriendModules()
{
    // initialize successor list
    successorList->initializeList(par("successorListSize"), thisNode, this);
}

void Koorde::predecessorIsDead()
{
    preDeBruijnNodeSet = false;
    //todo: this info could be used to remove predecessor node
    //      from de bruijn successor list or even as our de bruijn node
}

void Koorde::successorIsDead()
{
    sucDeBruijnNodeSet = false;
    //todo: this info could be used to remove successor node
    //      from de bruijn successor list or even as our de bruijn node
}

