//
// Copyright (C) 2006 Institut fuer Telematik, Universitaet Karlsruhe (TH)
//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public License
// as published by the Free Software Foundation; either version 2
// of the License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
//

/**
 * @file Pastry.cc
 * @author Felix Palmen
 * @author Gerhard Petruschat
 * @author Bernhard Heep
 */

#include <IPAddressResolver.h>
#include <IPvXAddress.h>
#include <InterfaceTable.h>
#include <IPv4InterfaceData.h>
#include <RpcMacros.h>
#include <InitStages.h>
#include <SimpleNetConfigurator.h>
#include <NeighborCache.h>
#include <GlobalStatistics.h>
#include <BootstrapList.h>
#include <assert.h>

#include "BasePastry.h"

void BasePastry::purgeVectors(void)
{
    // purge pending state messages
    if (!stReceived.empty()) {
        for (std::vector<PastryStateMsgHandle>::iterator it =
                 stReceived.begin(); it != stReceived.end(); it++) {
            // check whether one of the pointers is a duplicate of stateCache
            if (it->msg == stateCache.msg) {
                stateCache.msg = NULL;
                stateCache.prox = NULL;
            }
            delete it->msg;
            delete it->prox;
        }
        stReceived.clear();
    }

    // purge notify list:
    notifyList.clear();

    // purge Queue for messages to be forwarded in recursive mode:
    for (std::map<TransportAddress, BaseRouteMessage*>::iterator i =
        recFwdQueue.begin(); i != recFwdQueue.end(); i++) {
        delete i->second;
    }
    recFwdQueue.clear();

    // purge Queue for processing multiple STATE messages:
    while (! stateCacheQueue.empty()) {
        delete stateCacheQueue.front().msg;
        stateCacheQueue.pop();
    }

    // delete cached state message:
    if (stateCache.msg) {
        delete stateCache.msg;
        stateCache.msg = NULL;
        delete stateCache.prox;
        stateCache.prox = NULL;
    }

    // purge vector of waiting sendState messages:
    if (! sendStateWait.empty()) {
        for (std::vector<PastrySendState*>::iterator it =
                 sendStateWait.begin(); it != sendStateWait.end(); it++) {
            if ( (*it)->isScheduled() ) cancelEvent(*it);
            delete *it;
        }
        sendStateWait.clear();
    }
}

void BasePastry::baseInit()
{

    bitsPerDigit = par("bitsPerDigit");
    numberOfLeaves = par("numberOfLeaves");
    numberOfNeighbors = par("numberOfNeighbors");
    joinTimeoutAmount = par("joinTimeout");
    readyWaitAmount = par("readyWait");
    repairTimeout = par("repairTimeout");
    enableNewLeafs = par("enableNewLeafs");
    optimizeLookup = par("optimizeLookup");
    optimisticForward = par("optimisticForward");
    avoidDuplicates = par("avoidDuplicates");
    sendStateWaitAmount = par("sendStateWaitAmount");
    partialJoinPath = par("partialJoinPath");
    timeoutPing = par("pingTimeout");
    pingRetries = par("pingRetries");
    useRegularNextHop = par("useRegularNextHop");
    alwaysSendUpdate = par("alwaysSendUpdate");

    if (!neighborCache->isEnabled()) {
        throw new cRuntimeError("NeighborCache is disabled, which is mandatory "
                                "for Pastry/Bamboo. Activate it by setting "
                                "\"**.neighborCache.enableNeighborCache "
                                "= true\" in your omnetpp.ini!");
    }
    // coordinate-based routing (CBR)
    coordBasedRouting = par("coordBasedRouting");
    if(coordBasedRouting &&
       !dynamic_cast<SimpleNetConfigurator*>(underlayConfigurator)) {
        opp_error("Coordinate-based routing currently only"
                  " possible in SimpleUnderlay!");
    }
    numCoordDigits = par("numCoordDigits");
    CBRstartAtDigit = par("CBRstartAtDigit");
    CBRstopAtDigit = par("CBRstopAtDigit");

    if (numberOfLeaves % 2) {
        EV << "Pastry: Warning: numberOfLeaves must be even - adding 1.";
        numberOfLeaves++;
    }

    routingTable = check_and_cast<PastryRoutingTable*>
        (parentModule()->submodule("pastryRoutingTable"));
    leafSet = check_and_cast<PastryLeafSet*>
        (parentModule()->submodule("pastryLeafSet"));
    neighborhoodSet = check_and_cast<PastryNeighborhoodSet*>
        (parentModule()->submodule("pastryNeighborhoodSet"));

    stateCache.msg = NULL;
    stateCache.prox = NULL;

    rowToAsk = 0;

    // initialize statistics
    joins = 0;
    joinTries = 0;
    joinPartial = 0;
    joinSeen = 0;
    joinReceived = 0;
    joinSent = 0;
    stateSent = 0;
    stateReceived = 0;
    repairReqSent = 0;
    repairReqReceived = 0;
    stateReqSent = 0;
    stateReqReceived = 0;

    joinBytesSeen = 0;
    joinBytesReceived = 0;
    joinBytesSent = 0;
    stateBytesSent = 0;
    stateBytesReceived = 0;
    repairReqBytesSent = 0;
    repairReqBytesReceived = 0;
    stateReqBytesSent = 0;
    stateReqBytesReceived = 0;

    totalLookups = 0;
    responsibleLookups = 0;
    routingTableLookups = 0;
    closerNodeLookups = 0;
    closerNodeLookupsFromNeighborhood = 0;

    leafsetReqSent = 0;
    leafsetReqBytesSent = 0;
    leafsetReqReceived = 0;
    leafsetReqBytesReceived = 0;
    leafsetSent = 0;
    leafsetBytesSent = 0;
    leafsetReceived = 0;
    leafsetBytesReceived = 0;

    routingTableReqSent = 0;
    routingTableReqBytesSent = 0;
    routingTableReqReceived = 0;
    routingTableReqBytesReceived = 0;
    routingTableSent = 0;
    routingTableBytesSent = 0;
    routingTableReceived = 0;
    routingTableBytesReceived = 0;

    WATCH(joins);
    WATCH(joinTries);
    WATCH(joinSeen);
    WATCH(joinBytesSeen);
    WATCH(joinReceived);
    WATCH(joinBytesReceived);
    WATCH(joinSent);
    WATCH(joinBytesSent);
    WATCH(stateSent);
    WATCH(stateBytesSent);
    WATCH(stateReceived);
    WATCH(stateBytesReceived);
    WATCH(repairReqSent);
    WATCH(repairReqBytesSent);
    WATCH(repairReqReceived);
    WATCH(repairReqBytesReceived);
    WATCH(stateReqSent);
    WATCH(stateReqBytesSent);
    WATCH(stateReqReceived);
    WATCH(stateReqBytesReceived);
    WATCH(lastStateChange);

    WATCH(leafsetReqSent);
    WATCH(leafsetReqBytesSent);
    WATCH(leafsetReqReceived);
    WATCH(leafsetReqBytesReceived);
    WATCH(leafsetSent);
    WATCH(leafsetBytesSent);
    WATCH(leafsetReceived);
    WATCH(leafsetBytesReceived);

    WATCH(routingTableReqSent);
    WATCH(routingTableReqBytesSent);
    WATCH(routingTableReqReceived);
    WATCH(routingTableReqBytesReceived);
    WATCH(routingTableSent);
    WATCH(routingTableBytesSent);
    WATCH(routingTableReceived);
    WATCH(routingTableBytesReceived);

}

void BasePastry::baseChangeState(int toState)
{
    switch (toState) {
    case INIT:

        state = INIT;

        if (!thisNode.key.isUnspecified())
            bootstrapList->removeBootstrapNode(thisNode);

        if (joinTimeout->isScheduled()) cancelEvent(joinTimeout);
        if (readyWait->isScheduled()) cancelEvent(readyWait);

        purgeVectors();

        bootstrapNode = bootstrapList->getBootstrapNode();

        routingTable->initializeTable(bitsPerDigit, repairTimeout,
                                      thisNode);
        leafSet->initializeSet(numberOfLeaves, bitsPerDigit,
                               repairTimeout, thisNode,
                               this);
        neighborhoodSet->initializeSet(numberOfNeighbors, bitsPerDigit,
                                       thisNode);

        updateTooltip();
        lastStateChange = simTime();

        parentModule()->parentModule()->bubble("entering INIT state");

        break;


    case DISCOVERY:

        break;

    case JOINING_2:
        state = JOINING_2;

        // bootstrapNode must be obtained before calling this method,
        // for example by calling changeState(INIT)

        if (bootstrapNode.isUnspecified()) {
            // no existing pastry network -> first node of a new one
            changeState(READY);
            return;
        }

        cancelEvent(joinTimeout);
        scheduleAt(simTime()+joinTimeoutAmount, joinTimeout);

        updateTooltip();
        parentModule()->parentModule()->bubble("entering JOIN state");

        RECORD_STATS(joinTries++);

        break;

    case READY:
        assert(state != READY);
        state = READY;

         //bootstrapList->registerBootstrapNode(thisNode);

        // if we are the first node in the network, there's nothing else
        // to do
        if (bootstrapNode.isUnspecified()) {
            RECORD_STATS(joinTries++);
            RECORD_STATS(joins++);
            setOverlayReady(true);
            return;
        }

        parentModule()->parentModule()->bubble("entering READY state");
        updateTooltip();
        RECORD_STATS(joins++);

        break;
    }
    setOverlayReady(state == READY);
}


void BasePastry::newLeafs(void)
{
    if (! enableNewLeafs) return;

    PastryNewLeafsMessage* msg = leafSet->getNewLeafsMessage();
    if (msg) {
        send(msg, "to_app");
        EV << "Pastry: newLeafs() called." << endl;
    }
}
void BasePastry::setOwnNodeID()
{
    thisNode.key = OverlayKey::random();
    //cbr
}


void BasePastry::changeState(int toState)
{

}


bool BasePastry::forwardMessageRecursive(const TransportAddress& dest,
                                         BaseRouteMessage* msg)
{
    if (dest == thisNode) {
        return true;
    }
    PastryMessage* pmsg
        = dynamic_cast<PastryMessage*>(msg->encapsulatedMsg());
    if ( pmsg && (dest != thisNode) ) {
        if (pmsg->getPastryMsgType() == PASTRY_MSG_JOIN) {
            PastryJoinMessage* jmsg = check_and_cast<PastryJoinMessage*>(pmsg);
            RECORD_STATS(joinSeen++; joinBytesSeen += jmsg->byteLength());
//            assert(jmsg->getJoinHopCount() ==
//                   check_and_cast<OverlayCtrlInfo*>(msg->controlInfo())
//                   ->getHopCount());
            sendStateTables(jmsg->getSendStateTo(), PASTRY_STATE_JOIN,
                            check_and_cast<OverlayCtrlInfo*>(msg->controlInfo())
                                               ->getHopCount(),
                            /*jmsg->getJoinHopCount(),*/ false);
            //jmsg->setJoinHopCount(jmsg->getJoinHopCount() + 1);
        }
    }

    if (recFwdQueue.find(dest) != recFwdQueue.end()) {
        // rare case, other message for same next hop is pending:
        // send the message and hope it isn't lost.
        return true;
    }

    assert(msg->controlInfo());

//    bool send = false;
//    if (optimisticForward) {
//        // forward now:
//        send = true;
//        if (! avoidDuplicates) {
//            // and keep copy for possible retry:
//            recFwdQueue[dest] = static_cast<BaseRouteMessage*>(msg->dup());
//        }
//    } else {
//        // keep message in queue (forward later)
//        recFwdQueue[dest] = msg;
//    }
//    pingNode(dest, timeoutPing, pingRetries, "PING next hop");
//
//    return send;

    if (optimisticForward) {
        // forward now:
        return true;
    } else {
        if (routeMsgAcks) throw new cRuntimeError("!optimisticForward && "
                                                  "routeMsgAcks");
        // keep message in queue (forward later)
        recFwdQueue[dest] = msg;
        pingNode(dest, timeoutPing, pingRetries, NULL,
                 "PING next hop");

        return false;
    }
}

void BasePastry::doRoutingTableMaintenance()
{

  if (sendPullFlag) {
    //static uint rowToAsk = 1;
    uint rrows = routingTable->getLastRow();
    if (++rowToAsk >= rrows) rowToAsk = 0;
    const TransportAddress& ask4row = routingTable->getRandomNode(rowToAsk);

    if ((!ask4row.isUnspecified()) && (ask4row != thisNode)) {

        PastryRoutingRowRequestMessage* msg = new PastryRoutingRowRequestMessage("ROWREQ");

        msg->setPastryMsgType(PASTRY_MSG_ROWREQ);
        msg->setStatType(MAINTENANCE_STAT);
        msg->setSendStateTo(thisNode);
        msg->setRow(rowToAsk+1);
        msg->setLength(PASTRYRTREQ_L(msg));
        RECORD_STATS(routingTableReqSent++; routingTableReqBytesSent += msg->byteLength());
        EV << "Bamboo: Sending  Message to Node in Row" << rowToAsk << endl;
        sendMessageToUDP(ask4row, msg);
    }
  } else {


    for (int i = 0; i < routingTable->getLastRow(); i++) {
        const TransportAddress& ask4row = routingTable->getRandomNode(i);

        if ((!ask4row.isUnspecified()) && (ask4row != thisNode)) {

            PastryRoutingRowRequestMessage* msg = new PastryRoutingRowRequestMessage("ROWREQ");

            msg->setPastryMsgType(PASTRY_MSG_ROWREQ);
            msg->setStatType(MAINTENANCE_STAT);
            msg->setSendStateTo(thisNode);
            msg->setRow(i+1);
            msg->setLength(PASTRYRTREQ_L(msg));
            RECORD_STATS(routingTableReqSent++; routingTableReqBytesSent += msg->byteLength());
            sendMessageToUDP(ask4row, msg);
        } else EV << "could not send Message to Node in Row" << i << endl;

    }
  }
}

void BasePastry::pingResponse(PingResponse* msg, cPolymorphic* context,
                              int rpcId, simtime_t rtt)
{
    const NodeHandle& src = msg->getSrcNode();
    assert(!src.isUnspecified());

    // a node with the an equal ID has responded
    if ((src.key == thisNode.key) && (src.ip != thisNode.ip)) {
        EV << "a node with the an equal ID has responded, rejoining" << endl;
        delete context;
        joinOverlay();
        return;
    }

    //const NodeHandle* node;
    if (context != NULL && stateCache.msg && stateCache.prox) {
        PingContext* pingContext = check_and_cast<PingContext*>(context);
        if (pingContext->nonce != stateCache.nonce) {
            throw new cRuntimeError("response doesn't fit stateCache");
        }
        switch (pingContext->stateObject) {
            case ROUTINGTABLE: {
                /*node = &(stateCache.msg->getRoutingTable(pingContext->index));
                if((node->isUnspecified()) || (*node != src)) {
                    std::cout << simTime() << " " << thisNode.ip << " rt: state from "
                              << stateCache.msg->getSender().ip << " *** failed: node "
                              << node->ip << " src " << src.ip << std::endl;
                    break;
                }*/
                *(stateCache.prox->pr_rt.begin() + pingContext->index) = rtt;
                break;
            }
            case LEAFSET: {
                /*node = &(stateCache.msg->getLeafSet(pingContext->index));
                if ((node->isUnspecified()) || (*node != src)) {
                    std::cout << simTime() << " " << thisNode.ip << " ls: state from "
                              << stateCache.msg->getSender().ip << " *** failed: node "
                              << node->ip << " src " << src.ip << std::endl;
                    break;
                }*/
                *(stateCache.prox->pr_ls.begin() + pingContext->index) = rtt;
                break;
            }
            case NEIGHBORHOODSET: {
                /*node = &(stateCache.msg->getNeighborhoodSet(pingContext->index));
                if((node->isUnspecified()) || (*node != src)) {
                    std::cout << simTime() << " " << thisNode.ip << " ns: state from "
                              << stateCache.msg->getSender().ip << " *** failed: node "
                              << node->ip << " src " << src.ip << std::endl;
                    break;
                }*/
                *(stateCache.prox->pr_ns.begin() + pingContext->index) = rtt;
                break;
            }
            default: {
                throw new cRuntimeError("wrong state object type!");
            }
        }
    }

    delete context;

    checkProxCache();

    std::map<TransportAddress, BaseRouteMessage*>::iterator pos;
    if ((pos = recFwdQueue.find(src)) != recFwdQueue.end()) {
        // send message
        if (!optimisticForward) {
            sendMessageToUDP(pos->first, pos->second);
        } else if (!avoidDuplicates) {
            delete pos->second;
        }
        recFwdQueue.erase(pos);
    }
}

void BasePastry::prePing(const PastryStateMessage* stateMsg)
{
    uint rt_size = stateMsg->getRoutingTableArraySize();
    uint ls_size = stateMsg->getLeafSetArraySize();
    uint ns_size = stateMsg->getNeighborhoodSetArraySize();

    for (uint i = 0; i < rt_size + ls_size + ns_size; i++) {
        const NodeHandle* node;
        if (i < rt_size) {
            node = &(stateMsg->getRoutingTable(i));
        }
        else if (i < (rt_size + ls_size) ) {
            node = &(stateMsg->getLeafSet(i - rt_size));
        }
        else {
            node = &(stateMsg->getNeighborhoodSet(i - rt_size - ls_size));
        }
        if ((node->isUnspecified()) || (*node == thisNode)) {
            continue;
        }
        if (node->key == thisNode.key) {
            /*
            cerr << "Pastry Warning: Other node with same key found, "
                "restarting!" << endl;
            opp_error("TODO: Other node with same key found...");
            joinOverlay(); //segfault
            return;
            */
            continue;
        }
        pingNode(*node, timeoutPing, pingRetries, NULL,
                 "PING received state");
    }
}

void BasePastry::pingNodes(void)
{
    if (stateCache.msg == NULL) throw new cRuntimeError("no state msg");

    assert(stateCache.prox == NULL);
    stateCache.prox = new PastryStateMsgProximity();

    uint rt_size = stateCache.msg->getRoutingTableArraySize();
    //stateCache.prox->pr_rt.clear();
    stateCache.prox->pr_rt.resize(rt_size, PASTRY_PROX_UNDEF);

    uint ls_size = stateCache.msg->getLeafSetArraySize();
    //stateCache.prox->pr_ls.clear();
    stateCache.prox->pr_ls.resize(ls_size, PASTRY_PROX_UNDEF);

    uint ns_size = stateCache.msg->getNeighborhoodSetArraySize();
    //stateCache.prox->pr_ns.clear();
    stateCache.prox->pr_ns.resize(ns_size, PASTRY_PROX_UNDEF);

    std::vector< std::pair<const NodeHandle*, PingContext*> > nodesToPing;
    // set prox state
    for (uint i = 0; i < rt_size + ls_size + ns_size; i++) {
        const NodeHandle* node;
        std::vector<simtime_t>::iterator proxPos;
        PingContext* pingContext = NULL;
        StateObject stateObject;
        uint index;
        if (stateCache.msg == NULL) break;
        if (i < rt_size) {
            node = &(stateCache.msg->getRoutingTable(i));
            proxPos = stateCache.prox->pr_rt.begin() + i;
            stateObject = ROUTINGTABLE;
            index = i;
        } else if ( i < (rt_size + ls_size) ) {
            node = &(stateCache.msg->getLeafSet(i - rt_size));
            proxPos = stateCache.prox->pr_ls.begin() + (i - rt_size);
            stateObject = LEAFSET;
            index = i - rt_size;
        } else {
            node = &(stateCache.msg->getNeighborhoodSet(i - rt_size - ls_size));
            proxPos = stateCache.prox->pr_ns.begin() + (i - rt_size - ls_size);
            stateObject = NEIGHBORHOODSET;
            index = i - rt_size - ls_size;
        }
        // proximity is undefined for unspecified nodes:
        if (!node->isUnspecified()) {
            if (*node == thisNode) {
                *proxPos = 0;
            } else {
                *proxPos = PASTRY_PROX_PENDING;
                pingContext = new PingContext(stateObject, index,
                                              stateCache.nonce);
                nodesToPing.push_back(std::make_pair(node, pingContext));
            }
        }
    }
    assert(nodesToPing.size() != 0);
    // ping nodes
    for (uint i = 0; i < nodesToPing.size(); ++i) {
        pingNode(*nodesToPing[i].first, timeoutPing, pingRetries,
                 nodesToPing[i].second, "PING received state");
    }

}

void BasePastry::determineAliveTable(const PastryStateMessage* stateMsg)
{
    uint rt_size = stateMsg->getRoutingTableArraySize();
    aliveTable.pr_rt.clear();
    aliveTable.pr_rt.resize(rt_size, 1);

    uint ls_size = stateMsg->getLeafSetArraySize();
    aliveTable.pr_ls.clear();
    aliveTable.pr_ls.resize(ls_size, 1);

    uint ns_size = stateMsg->getNeighborhoodSetArraySize();
    aliveTable.pr_ns.clear();
    aliveTable.pr_ns.resize(ns_size, 1);

    for (uint i = 0; i < rt_size + ls_size + ns_size; i++) {
        const TransportAddress* node;
        std::vector<simtime_t>::iterator tblPos;
        if (i < rt_size) {
            node = &(stateMsg->getRoutingTable(i));
            tblPos = aliveTable.pr_rt.begin() + i;
        } else if ( i < (rt_size + ls_size) ) {
            node = &(stateMsg->getLeafSet(i - rt_size));
            tblPos = aliveTable.pr_ls.begin() + (i - rt_size);
        } else {
            node = &(stateMsg->getNeighborhoodSet(i - rt_size - ls_size));
            tblPos = aliveTable.pr_ns.begin() + (i - rt_size - ls_size);
        }
        if (neighborCache->getAddressRtt(*node) == NEIGHBORSTATE_TIMEOUT) {
            *tblPos = PASTRY_PROX_INFINITE;
        }
    }
}

void BasePastry::sendStateTables(const TransportAddress& destination, int type, ...)
{
    if (destination.ip == thisNode.ip)
        opp_error("Pastry: trying to send state to self!");

    int hops = 0;
    bool last = false;
    simtime_t timestamp = 0;

    if ((type == PASTRY_STATE_JOIN) || (type == PASTRY_STATE_UPDATE)) {
        // additional parameters needed:
        va_list ap;
        va_start(ap, type);
        if (type == PASTRY_STATE_JOIN) {
            hops = va_arg(ap, int);
            last = static_cast<bool>(va_arg(ap, int));
        } else {
            timestamp = va_arg(ap, simtime_t);
        }
        va_end(ap);
    }

    // create new state msg and set special fields for some types:
    PastryStateMessage* stateMsg;
    if (type == PASTRY_STATE_JOIN) {
        stateMsg = new PastryStateMessage("STATE (Join)");
        stateMsg->setJoinHopCount(hops);
        stateMsg->setLastHop(last);
        stateMsg->setTimestamp(simTime());
    } else if (type == PASTRY_STATE_UPDATE) {
        stateMsg = new PastryStateMessage("STATE (Update)");
        EV << "Pastry: sending state (update) to " << destination << endl;
        stateMsg->setTimestamp(timestamp);
    } else if (type == PASTRY_STATE_REPAIR) {
        stateMsg = new PastryStateMessage("STATE (Repair)");
        stateMsg->setTimestamp(timestamp);
        EV << "Pastry: sending state (repair) to " << destination << endl;
    } else {
        stateMsg = new PastryStateMessage("STATE");
        EV << "Pastry: sending state (standard) to " << destination << endl;
        stateMsg->setTimestamp(simTime());
    }

    // fill in standard content:
    stateMsg->setPastryMsgType(PASTRY_MSG_STATE);
    stateMsg->setStatType(MAINTENANCE_STAT);
    stateMsg->setPastryStateMsgType(type);
    stateMsg->setSender(thisNode);
    routingTable->dumpToStateMessage(stateMsg);
    leafSet->dumpToStateMessage(stateMsg);
    neighborhoodSet->dumpToStateMessage(stateMsg);

    // send...
    stateMsg->setLength(PASTRYSTATE_L(stateMsg));
    RECORD_STATS(stateSent++; stateBytesSent += stateMsg->byteLength());
    sendMessageToUDP(destination, stateMsg);
}

void BasePastry::sendStateDelayed(const TransportAddress& destination)
{
    PastrySendState* selfMsg = new PastrySendState("sendStateWait");
    selfMsg->setDest(destination);
    sendStateWait.push_back(selfMsg);
    scheduleAt(simTime() + sendStateWaitAmount, selfMsg);
}

void BasePastry::pingTimeout(PingCall* msg,
                             const TransportAddress& dest,
                             cPolymorphic* context,
                             int rpcId)
{
    EV << "[BasePastry::pingTimeout() @ " << thisNode.ip
       << " (" << thisNode.key.toString(16) << ")]\n"
       << "    Ping timeout occurred (" << dest.ip << ")"
       << endl;

    // handle failed node
    if (state == READY) {
        handleFailedNode(dest);
        updateTooltip();

        // this could initiate a re-join, exit the handler in that
        // case because all local data was erased:
        if (state != READY) {
            delete context;
            return;
        }
    }

    if (context && stateCache.msg && stateCache.prox &&
        msg->isName("PING received state")) { //TODO string compare is inefficient
        PingContext* pingContext = check_and_cast<PingContext*>(context);
        if (pingContext->nonce != stateCache.nonce) {
            throw new cRuntimeError("timeout/call doesn't fit stateCache");
        }
        //const NodeHandle* node;
        switch (pingContext->stateObject) {
            case ROUTINGTABLE: {
                /*if (pingContext->index >=
                    stateCache.msg->getRoutingTableArraySize()) {
                    std::cout << "*** FAILED ***" << std::endl;
                    break;
                }
                node = &(stateCache.msg->getRoutingTable(pingContext->index));
                if((node->isUnspecified()) || (dest != *node)) {
                    std::cout << msg->getNonce() << " " << simTime() << " " << thisNode.ip << " rt: state from "
                    << stateCache.msg->getSender().ip << " *** failed: node "
                    << node->ip << " failed dest " << dest.ip << std::endl;
                    break;
                }*/
                *(stateCache.prox->pr_rt.begin() + pingContext->index) =
                    PASTRY_PROX_INFINITE;
                break;
            }
            case LEAFSET: {
                /*if (pingContext->index >=
                    stateCache.msg->getLeafSetArraySize()) {
                    std::cout << "*** FAILED ***" << std::endl;
                    break;
                }
                node = &(stateCache.msg->getLeafSet(pingContext->index));
                if((node->isUnspecified()) || (dest != *node)) {
                    std::cout << msg->getNonce() << " " << simTime() << " " << thisNode.ip << " ls: state from "
                    << stateCache.msg->getSender().ip << " *** failed: node "
                    << node->ip << " failed dest " << dest.ip << std::endl;
                    break;
                }*/
                *(stateCache.prox->pr_ls.begin() + pingContext->index) =
                    PASTRY_PROX_INFINITE;
                break;
            }
            case NEIGHBORHOODSET: {
                /*if (pingContext->index >=
                    stateCache.msg->getNeighborhoodSetArraySize()) {
                    std::cout << "*** FAILED ***" << std::endl;
                    break;
                }
                node = &(stateCache.msg->getNeighborhoodSet(pingContext->index));
                if((node->isUnspecified()) || (dest != *node)) {
                    std::cout << msg->getNonce() << " " << simTime() << " " << thisNode.ip << " ns: state from "
                    << stateCache.msg->getSender().ip << " *** failed: node "
                    << node->ip << " failed dest " << dest.ip << std::endl;
                    break;
                }*/
                *(stateCache.prox->pr_ns.begin() + pingContext->index) =
                    PASTRY_PROX_INFINITE;
                break;
            }
        }
        checkProxCache();
    }

    delete context;

    if (msg->isName("PING next hop")) {//TODO string compare is inefficient
        // handle forward queue entry
        std::map<TransportAddress, BaseRouteMessage*>::iterator pos;
        if ((pos = recFwdQueue.find(dest)) == recFwdQueue.end()) return;
        BaseRouteMessage* rmsg = pos->second;
        recFwdQueue.erase(pos);
        if (dynamic_cast<PastryJoinMessage*>(rmsg->encapsulatedMsg()) &&
                rmsg->getDestKey() == thisNode.key) {
            std::cout << "join timeout -> rejoin!";
            changeState(INIT);
            return;
        }
        assert(rmsg->controlInfo());
        sendToKey(rmsg->getDestKey(), rmsg);
    }
}

void BasePastry::sendRequest(const TransportAddress& ask, int type)
{
    std::string msgName("Req: ");
    switch (type) {
    case PASTRY_REQ_REPAIR:
        if (ask.isUnspecified())
            opp_error("Pastry::sendRequest(): asked for repair from "
                  "unspecified node!");
        msgName += "Repair";
        break;

    case PASTRY_REQ_STATE:
        if (ask.isUnspecified())
            opp_error("Pastry::sendRequest(): asked for state from "
                  "unspecified node!");
        msgName += "State";
        break;

    case PASTRY_REQ_LEAFSET:
        if (ask.isUnspecified())
            opp_error("Pastry::sendRequest(): asked for leafset from "
                  "unspecified node!");
        msgName += "Leafset";
        break;
    }
    PastryRequestMessage* msg = new PastryRequestMessage(msgName.c_str());
    msg->setPastryMsgType(PASTRY_MSG_REQ);
    msg->setPastryReqType(type);
    msg->setStatType(MAINTENANCE_STAT);
    msg->setSendStateTo(thisNode);
    msg->setLength(PASTRYREQ_L(msg));
    sendMessageToUDP(ask, msg);

    switch (type) {
    case PASTRY_REQ_REPAIR:
        RECORD_STATS(repairReqSent++; repairReqBytesSent += msg->byteLength());
        break;

    case PASTRY_REQ_STATE:
        RECORD_STATS(stateReqSent++; stateReqBytesSent += msg->byteLength());
        break;

    case PASTRY_REQ_LEAFSET:
        RECORD_STATS(leafsetReqSent++; leafsetReqBytesSent += msg->byteLength());
        break;
    }
}


void BasePastry::sendLeafset(const TransportAddress& tell, bool pull)
{
    if (tell.isUnspecified())
        opp_error("Pastry::sendLeafset(): send leafset to "
                  "unspecified node!");
    //if (pull) PastryLeafsetPullMessage* msg = new PastryLeafsetPullMessage("Leafset Pull");
    PastryLeafsetMessage* msg = new PastryLeafsetMessage("Leafset");
    if (pull) msg->setPastryMsgType(PASTRY_MSG_LEAFSET_PULL);
    else msg->setPastryMsgType(PASTRY_MSG_LEAFSET);
    msg->setStatType(MAINTENANCE_STAT);
    msg->setSender(thisNode);
    msg->setSendStateTo(thisNode);
    leafSet->dumpToStateMessage(msg);
    msg->setLength(PASTRYLEAFSET_L(msg));
    RECORD_STATS(leafsetSent++; leafsetBytesSent += msg->byteLength());
    sendMessageToUDP(tell, msg);


}

void BasePastry::sendRoutingRow(const TransportAddress& tell, int row)
{
    if (tell.isUnspecified())
        opp_error("Pastry::sendRoutingTable(): asked for routing Table from "
                  "unspecified node!");

    PastryRoutingRowMessage* msg = new PastryRoutingRowMessage("Routing Row");
    msg->setPastryMsgType(PASTRY_MSG_RROW);
    msg->setStatType(MAINTENANCE_STAT);
    //msg->setSendStateTo(thisNode);
    msg->setSender(thisNode);
    msg->setRow(row);
    routingTable->dumpRowToMessage(msg, row);
    msg->setLength(PASTRYRTABLE_L(msg));
    RECORD_STATS(routingTableSent++; routingTableBytesSent += msg->byteLength());
    sendMessageToUDP(tell, msg);
}

void BasePastry::handleRequestMessage(PastryRequestMessage* msg)
{
        uint reqtype = msg->getPastryReqType();
        if (reqtype == PASTRY_REQ_REPAIR) {
            RECORD_STATS(repairReqReceived++; repairReqBytesReceived +=
                         msg->byteLength());
            if (state == READY)
                sendStateTables(msg->getSendStateTo(),
                                PASTRY_STATE_REPAIR);
            else
                EV << "Pastry: received repair request before reaching "
                   << "READY state, dropping message!" << endl;
            delete msg;
        }
        else if (reqtype == PASTRY_REQ_STATE) {
            RECORD_STATS(stateReqReceived++; stateReqBytesReceived +=
                     msg->byteLength());
            if (state == READY)
                sendStateTables(msg->getSendStateTo());
            else
                EV << "Pastry: received state request before reaching "
                   << "READY state, dropping message!" << endl;
            delete msg;
        }
        else if (PASTRY_REQ_LEAFSET) {
            RECORD_STATS(leafsetReqReceived++; leafsetReqBytesReceived +=
            msg->byteLength());
            if (state == READY) {
                 sendLeafset(msg->getSendStateTo());
            }
            else
                EV << "Pastry: received leafset request before reaching "
                   << "READY state, dropping message!" << endl;
            delete msg;
        }

}

void BasePastry::handleLeafsetMessage(PastryLeafsetMessage* msg, bool mergeSender)
{
    uint lsSize = msg->getLeafSetArraySize();
    PastryStateMessage* stateMsg;

    stateMsg = new PastryStateMessage("STATE");
    stateMsg->setTimestamp(msg->getTimestamp());
    stateMsg->setPastryMsgType(PASTRY_MSG_STATE);
    stateMsg->setStatType(MAINTENANCE_STAT);
    stateMsg->setSender(msg->getSender());
    stateMsg->setLeafSetArraySize(lsSize);
    stateMsg->setNeighborhoodSetArraySize(0);
    stateMsg->setRoutingTableArraySize(0);

    for (uint i = 0; i < lsSize; i++) {
        stateMsg->setLeafSet(i, msg->getLeafSet(i));
    }

    if (mergeSender) {
        stateMsg->setLeafSetArraySize(lsSize+1);
        stateMsg->setLeafSet(lsSize, msg->getSender());
    }

    handleStateMessage(stateMsg);
    delete msg;
}




bool BasePastry::isSiblingFor(const NodeHandle& node,
                          const OverlayKey& key,
                          int numSiblings,
                          bool* err)
{
    if (key.isUnspecified())
        error("Pastry::isSiblingFor(): key is unspecified!");

    if ((numSiblings == 1) && (node == thisNode)) {
        if (leafSet->isClosestNode(key)) {
            *err = false;
            return true;
        } else {
            *err = false;
            return false;
        }
    }

    NodeVector* result =  leafSet->createSiblingVector(key, numSiblings);

    if (result == NULL) {
        *err = true;
        return false;
    }

    if (result->contains(node.key)) {
        delete result;
        *err = false;
        return true;
    } else {
        delete result;
        *err = true;
        return false;
    }

    /*
      const NodeHandle& dest = leafSet->getDestinationNode(key);
      if (!dest.isUnspecified()) {
      *err = false;
      return true;
      } else {

      *err = true;
      return false;
      }
    */
}


void BasePastry::handleAppMessage(BaseOverlayMessage* msg)
{
    delete msg;
}

void BasePastry::updateTooltip()
{
    if (ev.isGUI()) {
        std::stringstream ttString;

        // show our predecessor and successor in tooltip
        ttString << leafSet->getPredecessor() << endl << thisNode << endl
                 << leafSet->getSuccessor();

        parentModule()->parentModule()->displayString().
            setTagArg("tt", 0, ttString.str().c_str());
        parentModule()->displayString().
            setTagArg("tt", 0, ttString.str().c_str());
        displayString().setTagArg("tt", 0, ttString.str().c_str());

        // draw arrows:
        showOverlayNeighborArrow(leafSet->getSuccessor(), true,
                                 "m=m,50,0,50,0;o=red,1");
        showOverlayNeighborArrow(leafSet->getPredecessor(), false,
                                 "m=m,50,100,50,100;o=green,1");

    }
}

void BasePastry::finishOverlay()
{
    // remove this node from the bootstrap list
    if (!thisNode.key.isUnspecified()) bootstrapList->removeBootstrapNode(thisNode);

    purgeVectors();

    // collect statistics
    simtime_t time = globalStatistics->calcMeasuredLifetime(creationTime);
    if(time == 0) return;

    globalStatistics->addStdDev("Pastry: Successful Joins/s", joins / time);
    globalStatistics->addStdDev("Pastry: overall join tries/s", joinTries / time);
    globalStatistics->addStdDev("Pastry: joins with missing replies from routing path/s",
                                joinPartial / time);
    globalStatistics->addStdDev("Pastry: JOIN Messages seen/s", joinSeen / time);
    globalStatistics->addStdDev("Pastry: bytes of JOIN Messages seen/s", joinBytesSeen / time);
    globalStatistics->addStdDev("Pastry: JOIN Messages received/s", joinReceived / time);
    globalStatistics->addStdDev("Pastry: bytes of JOIN Messages received/s",
                                joinBytesReceived / time);
    globalStatistics->addStdDev("Pastry: JOIN Messages sent/s", joinSent / time);
    globalStatistics->addStdDev("Pastry: bytes of JOIN Messages sent/s", joinBytesSent / time);
    globalStatistics->addStdDev("Pastry: STATE Messages sent/s", stateSent / time);
    globalStatistics->addStdDev("Pastry: bytes of STATE Messages sent/s", stateBytesSent / time);
    globalStatistics->addStdDev("Pastry: STATE Messages received/s", stateReceived / time);
    globalStatistics->addStdDev("Pastry: bytes of STATE Messages received/s",
                                stateBytesReceived / time);
    globalStatistics->addStdDev("Pastry: REPAIR Requests sent/s", repairReqSent / time);
    globalStatistics->addStdDev("Pastry: bytes of REPAIR Requests sent/s",
                                repairReqBytesSent / time);
    globalStatistics->addStdDev("Pastry: REPAIR Requests received/s", repairReqReceived / time);
    globalStatistics->addStdDev("Pastry: bytes of REPAIR Requests received/s",
                                repairReqBytesReceived / time);
    globalStatistics->addStdDev("Pastry: STATE Requests sent/s", stateReqSent / time);
    globalStatistics->addStdDev("Pastry: bytes of STATE Requests sent/s", stateReqBytesSent / time);
    globalStatistics->addStdDev("Pastry: STATE Requests received/s", stateReqReceived / time);
    globalStatistics->addStdDev("Pastry: bytes of STATE Requests received/s",
                                stateReqBytesReceived / time);

    globalStatistics->addStdDev("Pastry: bytes of STATE Requests received/s",
                                stateReqBytesReceived / time);

    globalStatistics->addStdDev("Pastry: total number of lookups", totalLookups);
    globalStatistics->addStdDev("Pastry: responsible lookups", responsibleLookups);
    globalStatistics->addStdDev("Pastry: lookups in routing table", routingTableLookups);
    globalStatistics->addStdDev("Pastry: lookups using closerNode()", closerNodeLookups);
    globalStatistics->addStdDev("Pastry: lookups using closerNode() with result from "
                                "neighborhood set", closerNodeLookupsFromNeighborhood);
    globalStatistics->addStdDev("Pastry: LEAFSET Requests sent/s", leafsetReqSent / time);
    globalStatistics->addStdDev("Pastry: bytes of LEAFSET Requests sent/s", leafsetReqBytesSent / time);
    globalStatistics->addStdDev("Pastry: LEAFSET Requests received/s", leafsetReqReceived / time);
    globalStatistics->addStdDev("Pastry: bytes of LEAFSET Requests received/s",
                                leafsetReqBytesReceived / time);
    globalStatistics->addStdDev("Pastry: LEAFSET Messages sent/s", leafsetSent / time);
    globalStatistics->addStdDev("Pastry: bytes of LEAFSET Messages sent/s", leafsetBytesSent / time);
    globalStatistics->addStdDev("Pastry: LEAFSET Messages received/s", leafsetReceived / time);
    globalStatistics->addStdDev("Pastry: bytes of LEAFSET Messages received/s",
                                leafsetBytesReceived / time);
    globalStatistics->addStdDev("Pastry: ROUTING TABLE Requests sent/s", routingTableReqSent / time);
    globalStatistics->addStdDev("Pastry: bytes of ROUTING TABLE Requests sent/s", routingTableReqBytesSent / time);
    globalStatistics->addStdDev("Pastry: ROUTING TABLE Requests received/s", routingTableReqReceived / time);
    globalStatistics->addStdDev("Pastry: bytes of ROUTING TABLE Requests received/s",
                                routingTableReqBytesReceived / time);
    globalStatistics->addStdDev("Pastry: ROUTING TABLE Messages sent/s", routingTableSent / time);
    globalStatistics->addStdDev("Pastry: bytes of ROUTING TABLE Messages sent/s", routingTableBytesSent / time);
    globalStatistics->addStdDev("Pastry: ROUTING TABLE Messages received/s", routingTableReceived / time);
    globalStatistics->addStdDev("Pastry: bytes of ROUTING TABLE Messages received/s",
                                routingTableBytesReceived / time);

}

int BasePastry::getMaxNumSiblings()
{
    return (int) floor(numberOfLeaves/4.0);
}

int BasePastry::getMaxNumRedundantNodes()
{
    return 10;//TODO
}

NodeVector* BasePastry::findNode(const OverlayKey& key,
                                 int numRedundantNodes,
                                 int numSiblings,
                                 BaseOverlayMessage* msg)
{
    if ((numRedundantNodes > getMaxNumRedundantNodes()) ||
        (numSiblings > getMaxNumSiblings())) {

        opp_error("(Pastry::findNode()) numRedundantNodes or numSiblings "
                  "too big!");
    }
    RECORD_STATS(totalLookups++);

    NodeVector* nextHops = new NodeVector(numRedundantNodes);

    const NodeHandle* next;
    PastryFindNodeExtData* findNodeExt = NULL;
    if (msg && msg->hasObject("findNodeExt")) {
        findNodeExt =
            check_and_cast<PastryFindNodeExtData*>(msg->
                                                   getObject("findNodeExt"));
    }

    if (state != READY) {
        return nextHops;
    } else if (key.isUnspecified() || leafSet->isClosestNode(key)) {
        RECORD_STATS(responsibleLookups++);
        nextHops->add(thisNode);
    } else {
        // Send state tables on any JOIN message we see:
        if (findNodeExt) {
            const TransportAddress& stateRecipient =
                findNodeExt->getSendStateTo();
            if (!stateRecipient.isUnspecified()) {
                RECORD_STATS(joinSeen++);
                sendStateTables(stateRecipient, PASTRY_STATE_JOIN,
                                findNodeExt->getJoinHopCount(), false);
            }
        }

        next = &(leafSet->getDestinationNode(key));

        if (next->isUnspecified()) {
            next = &(routingTable->lookupNextHop(key));
            if (!next->isUnspecified()) {
                RECORD_STATS(routingTableLookups++);
            }
        } else {
            RECORD_STATS(responsibleLookups++);
        }

        if (next->isUnspecified()) {
            RECORD_STATS(closerNodeLookups++);
            // call findCloserNode() on all state objects
            if (optimizeLookup) {
                const NodeHandle* tmp;
                next = &(routingTable->findCloserNode(key, true));
                tmp = &(neighborhoodSet->findCloserNode(key, true));

                if ((! tmp->isUnspecified()) &&
                    (leafSet->isCloser(*tmp, key, *next))) {
                    RECORD_STATS(closerNodeLookupsFromNeighborhood++);
                    next = tmp;
                }

                tmp = &(leafSet->findCloserNode(key, true));
                if ((! tmp->isUnspecified()) &&
                    (leafSet->isCloser(*tmp, key, *next))) {
                    RECORD_STATS(closerNodeLookupsFromNeighborhood--);
                    next = tmp;
                }
            } else {
                next = &(routingTable->findCloserNode(key));

                if (next->isUnspecified()) {
                    RECORD_STATS(closerNodeLookupsFromNeighborhood++);
                    next = &(neighborhoodSet->findCloserNode(key));
                }

                if (next->isUnspecified()) {
                    RECORD_STATS(closerNodeLookupsFromNeighborhood--);
                    next = &(leafSet->findCloserNode(key));
                }
            }
        }

        if (!next->isUnspecified()) {
            if (findNodeExt) {
                findNodeExt->setJoinHopCount(findNodeExt->getJoinHopCount() + 1);
            }
            nextHops->add(*next);
        }
    }

    bool err;

    /* if we're a sibling, return all numSiblings */
    if (isSiblingFor(thisNode, key, numSiblings, &err)) {
        if (err == false) {
            delete nextHops;
            return  leafSet->createSiblingVector(key, numSiblings);
        }
    }

    //TODO routingType, too slow, ...
    if (/*(nextHops->size() > 0) &&*/ (numRedundantNodes > 1)) {
        //memleak... comp should be a ptr and deleted in NodeVector::~NodeVector()...
        KeyDistanceComparator<KeyRingMetric> comp(key);
        //KeyDistanceComparator<KeyRingMetric>* comp =
        //new KeyDistanceComparator<KeyRingMetric>( key );
        NodeVector* additionalHops = new NodeVector( numRedundantNodes, &comp );

        routingTable->findCloserNodes(key, additionalHops);
        leafSet->findCloserNodes(key, additionalHops);
        neighborhoodSet->findCloserNodes(key, additionalHops);

        if (useRegularNextHop && (*additionalHops)[0] != (*nextHops)[0]) {
            for (uint i = 0; i < additionalHops->size(); i++) {
                if ((*additionalHops)[i] != (*nextHops)[0])
                    nextHops->push_back((*additionalHops)[i]);
            }
            delete additionalHops;
        } else {
            delete nextHops;
            return additionalHops;
        }
    }
    return nextHops;
}
AbstractLookup* BasePastry::createLookup(RoutingType routingType,
                                         const BaseOverlayMessage* msg,
                                         const cObject* dummy,
                                         bool appLookup)
{
    assert(dummy == NULL);
    PastryFindNodeExtData* findNodeExt =
        new PastryFindNodeExtData("findNodeExt");

    if (msg) {
        const PastryMessage* pmsg =
            dynamic_cast<const PastryMessage*>(msg->encapsulatedMsg());
        if ((pmsg) && (pmsg->getPastryMsgType() == PASTRY_MSG_JOIN)) {
            const PastryJoinMessage* jmsg =
                check_and_cast<const PastryJoinMessage*>(pmsg);
            findNodeExt->setSendStateTo(jmsg->getSendStateTo());
            findNodeExt->setJoinHopCount(1);
        }
    }
    findNodeExt->setLength(PASTRYFINDNODEEXTDATA_L);

    AbstractLookup* newLookup = BaseOverlay::createLookup(routingType,
                                                          msg, findNodeExt,
                                                          appLookup);

    delete findNodeExt;
    return newLookup;
}

bool stateMsgIsSmaller(const PastryStateMsgHandle& hnd1,
                       const PastryStateMsgHandle& hnd2)
{
    return (hnd1.msg->getJoinHopCount() < hnd2.msg->getJoinHopCount());
}

std::ostream& operator<<(std::ostream& os, const PastryStateMsgProximity pr)
{
    os << "PastryStateMsgProximity {" << endl;
    os << "  pr_rt {" << endl;
    for (std::vector<simtime_t>::const_iterator i = pr.pr_rt.begin();
         i != pr.pr_rt.end(); ++i) {
        os << "    " << *i << endl;
    }
    os << "  }" << endl;
    os << "  pr_ls {" << endl;
    for (std::vector<simtime_t>::const_iterator i = pr.pr_ls.begin();
         i != pr.pr_ls.end(); ++i) {
        os << "    " << *i << endl;
    }
    os << "  }" << endl;
    os << "  pr_ns {" << endl;
    for (std::vector<simtime_t>::const_iterator i = pr.pr_ns.begin();
         i != pr.pr_ns.end(); ++i) {
        os << "    " << *i << endl;
    }
    os << "  }" << endl;
    os << "}" << endl;
    return os;
}
