//
// Copyright (C) 2006 Institut fuer Telematik, Universitaet Karlsruhe (TH)
//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public License
// as published by the Free Software Foundation; either version 2
// of the License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
//

/**
 * @file Pastry.cc
 * @author Felix Palmen
 * @author Gerhard Petruschat
 * @author Bernhard Heep
 */

#include <IPAddressResolver.h>
#include <IPvXAddress.h>
#include <InterfaceTable.h>
#include <IPv4InterfaceData.h>
#include <RpcMacros.h>
#include <InitStages.h>
#include <GlobalStatistics.h>
#include <SimpleNetConfigurator.h>

#include "Pastry.h"


Define_Module(Pastry);

Pastry::~Pastry()
{
    // destroy self timer messages
    cancelAndDelete(joinTimeout);
    cancelAndDelete(readyWait);
    cancelAndDelete(joinUpdateWait);
    cancelAndDelete(secondStageWait);
    cancelAndDelete(ringCheck);
    if (useDiscovery) cancelAndDelete(discoveryTimeout);
    if (periodicMaintenance) cancelAndDelete(repairTaskTimeout);

    purgeVectors();
}

void Pastry::initializeOverlay(int stage)
{
    if ( stage != MIN_STAGE_OVERLAY )
        return;

    // Pastry provides KBR services
    kbr = true;

    baseInit();

    useSecondStage = par("useSecondStage");
    useDiscovery = par("useDiscovery");
    periodicMaintenance = par("periodicMaintenance");
    secondStageWaitAmount = par("secondStageWait");
    discoveryTimeoutAmount = par("discoveryTimeoutAmount");
    repairTaskTimeoutAmount = par("repairTaskTimeoutAmount");
    sendStateAtLeafsetRepair = par("sendStateAtLeafsetRepair");
    ringCheckInterval = par("ringCheckInterval");

    overrideOldPastry = par("overrideOldPastry");
    overrideNewPastry = par("overrideNewPastry");

    if (overrideOldPastry) {
        useSecondStage = true;
        useDiscovery = false;
        periodicMaintenance = false;
        sendStateAtLeafsetRepair = true;
    }

    if (overrideNewPastry) {
        useSecondStage = false;
        useDiscovery = true;
        periodicMaintenance = true;
        discoveryTimeoutAmount = 0.4;
        repairTaskTimeoutAmount = 60;
        sendStateAtLeafsetRepair = false;
    }

    joinTimeout = new cMessage("joinTimeout");
    readyWait = new cMessage("readyWait");
    secondStageWait = new cMessage("secondStageWait");
    joinUpdateWait = new cMessage("joinUpdateWait");
    ringCheck = new cMessage("ringCheck");

    if (useDiscovery) discoveryTimeout = new cMessage("discoveryTimeout");
    if (periodicMaintenance) repairTaskTimeout = new cMessage("repairTaskTimeout");

    sendPullFlag = false;

}

void Pastry::joinOverlay()
{
    changeState(INIT);

    if (bootstrapNode.isUnspecified()) {
        // no existing pastry network -> first node of a new one
        changeState(READY);
    } else {
        // join existing pastry network
        nearNode = bootstrapNode;
        if (useDiscovery) changeState(DISCOVERY);
        else changeState(JOINING_2);
    }
}

void Pastry::changeState(int toState)
{
    if (ringCheck->isScheduled()) cancelEvent(ringCheck);
    baseChangeState(toState);

    switch (toState) {
    case INIT:
        purgeVectors();
        break;

    case DISCOVERY:
        state = DISCOVERY;
        //nearNode = bootstrapNode;
        nearNodeRtt = strToSimtime("10s");
        sendRequest(bootstrapNode, PASTRY_REQ_LEAFSET);
        depth = -1;

        break;

    case JOINING_2:

        joinHopCount = 0;

        {
            PastryJoinMessage* msg = new PastryJoinMessage("JOIN-Request");
            msg->setPastryMsgType(PASTRY_MSG_JOIN);
            msg->setStatType(MAINTENANCE_STAT);
            //msg->setJoinHopCount(1);
            msg->setSendStateTo(thisNode);
            msg->setLength(PASTRYJOIN_L(msg));
            RECORD_STATS(joinSent++; joinBytesSent += msg->byteLength());
            std::vector<TransportAddress> sourceRoute;
            sourceRoute.push_back(nearNode);
            sendToKey(thisNode.key, msg, 0/*1*/, sourceRoute);

        }


        break;

    case READY:
        if (ringCheckInterval > 0) {
            scheduleAt(simTime()+ringCheckInterval, ringCheck);
        }

        // determine list of all known nodes as notifyList
        notifyList.clear();
        leafSet->dumpToVector(notifyList);
        routingTable->dumpToVector(notifyList);
        sort(notifyList.begin(), notifyList.end());
        notifyList.erase(unique(notifyList.begin(), notifyList.end()),
                         notifyList.end());

        // schedule update
        cancelEvent(joinUpdateWait);
        scheduleAt(simTime() + sendStateWaitAmount, joinUpdateWait);

        // schedule second stage
        if (useSecondStage) {
            cancelEvent(secondStageWait);
            scheduleAt(simTime() + secondStageWaitAmount, secondStageWait);
        }

        // schedule routing table maintenance task
        if (periodicMaintenance) {
            cancelEvent(repairTaskTimeout);
            scheduleAt(simTime() + repairTaskTimeoutAmount, repairTaskTimeout);
            //leaf2ask = NULL;
        }

        break;
    }
}

void Pastry::handleTimerEvent(cMessage* msg)
{

    if (msg->isName("joinTimeout")) {
        EV << "Pastry: join timeout expired, restarting..." << endl;
        //std::cout << simTime() << " " << thisNode.ip << " " << stReceived.size() <<" Pastry: join timeout expired, restarting..." << std::endl;
        joinOverlay();
    } else if (msg->isName("readyWait")) {
        if (partialJoinPath) {
            RECORD_STATS(joinPartial++);
            sort(stReceived.begin(), stReceived.end(), stateMsgIsSmaller);

            // start pinging the nodes found in the first state message:
            stReceivedPos = stReceived.begin();
            stateCache = *stReceivedPos;
            EV << "Pastry: joining despite some missing STATE messages."
               << endl;
            pingNodes();
        } else {
            EV << "Pastry: timeout waiting for missing state messages in JOIN "
                "state, restarting..." << endl;
            //std::cout << thisNode.ip << "Pastry: timeout waiting for missing state messages in JOIN "
            //                "state, restarting..." << std::endl;
            joinOverlay();
        }
    } else if (msg->isName("joinUpdateWait")) {
        EV << "Pastry: sending state updates to all nodes." << endl;
        doJoinUpdate();
    } else if (msg->isName("secondStageWait")) {
        EV << "Pastry: sending STATE requests to all nodes in second stage "
            "of initialization." << endl;
        doSecondStage();
    } else if (msg->isName("ringCheck")) {
        if (state == READY) {
            // ping direct neighbors on the ring:
            const NodeHandle& pred = leafSet->getPredecessor();
            const NodeHandle& succ = leafSet->getSuccessor();
            if (! pred.isUnspecified()) {
                pingNode(pred, timeoutPing, pingRetries,
                         NULL, "PING ring check");
            }
            if (! succ.isUnspecified()) {
                pingNode(succ, timeoutPing, pingRetries,
                         NULL, "PING ring check");
            }
        }
        scheduleAt(simTime()+ringCheckInterval, ringCheck);
    } else if (msg->isName("sendStateWait")) {
        PastrySendState* sendStateMsg = check_and_cast<PastrySendState*>(msg);

        std::vector<PastrySendState*>::iterator pos =
            std::find(sendStateWait.begin(), sendStateWait.end(),
                      sendStateMsg);
        if (pos != sendStateWait.end()) sendStateWait.erase(pos);

        sendStateTables(sendStateMsg->getDest());
        delete sendStateMsg;
    }

    else if (msg->isName("discoveryTimeout")) {

        if ((depth == 0) && (nearNodeImproved)) depth++; //repeat last step if closer node was found
        if ((depth == 0) || (pingedNodes < 1)) changeState(JOINING_2);
        else {
            PastryRoutingRowRequestMessage* msg = new PastryRoutingRowRequestMessage("ROWREQ");
            msg->setPastryMsgType(PASTRY_MSG_ROWREQ);
            msg->setStatType(MAINTENANCE_STAT);
            msg->setSendStateTo(thisNode);
            msg->setRow(depth);
            msg->setLength(PASTRYRTREQ_L(msg));
            RECORD_STATS(routingTableReqSent++; routingTableReqBytesSent += msg->byteLength());
            sendMessageToUDP(nearNode, msg);
        }
    }

    else if (msg->isName("repairTaskTimeout")) {
        EV << "Pastry: starting routing table maintenance" << endl;
        doRoutingTableMaintenance();
        scheduleAt(simTime()+repairTaskTimeoutAmount, repairTaskTimeout);
    }

}

void Pastry::handleUDPMessage(BaseOverlayMessage* msg)
{
    PastryMessage* pastryMsg = check_and_cast<PastryMessage*>(msg);
    uint type = pastryMsg->getPastryMsgType();

    if (debugOutput) {
        EV << "Pastry: incoming message of type ";
        switch(type) {
        case PASTRY_MSG_STD:
            EV << "PASTRY_MSG_STD";
            break;
        case PASTRY_MSG_JOIN:
            EV << "PASTRY_MSG_JOIN";
            break;
        case PASTRY_MSG_STATE:
            EV << "PASTRY_MSG_STATE";
            break;
        case PASTRY_MSG_LEAFSET:
            EV << "PASTRY_MSG_LEAFSET";
            break;
        case PASTRY_MSG_ROWREQ:
            EV << "PASTRY_MSG_ROWREQ";
            break;
        case PASTRY_MSG_RROW:
            EV << "PASTRY_MSG_RROW";
            break;
        case PASTRY_MSG_REQ:
            EV << "PASTRY_MSG_REQ";
            break;
        default:
            EV << "UNKNOWN (" << type <<")";
            break;
        }
        EV << endl;
    }

    switch (type) {
    case PASTRY_MSG_STD:
        opp_error("Pastry received PastryMessage of unknown type!");
        break;

    case PASTRY_MSG_JOIN: {
        PastryJoinMessage* jmsg =
            check_and_cast<PastryJoinMessage*>(pastryMsg);
        RECORD_STATS(joinReceived++; joinBytesReceived +=
                     jmsg->byteLength());
        if (state != READY) {
            if (jmsg->getSendStateTo() == thisNode) {
                EV << "[Pastry::handleUDPMessage() @ " << thisNode.ip
                   << " (" << thisNode.key.toString(16) << ")]\n"
                   << "    PastryJoinMessage received by originator!";
            } else {
                EV << "Pastry: received join message before reaching "
                   << "READY state, dropping message!" << endl;
            }
        }
        else if (jmsg->getSendStateTo() == thisNode) {
            EV << "[Pastry::handleUDPMessage() @ " << thisNode.ip
               << " (" << thisNode.key.toString(16) << ")]\n"
               << "    PastryJoinMessage gets dropped because it is "
               << "outdated and has been received by originator!"
               << endl;
        } else {
            OverlayCtrlInfo* overlayCtrlInfo
                = check_and_cast<OverlayCtrlInfo*>(jmsg->controlInfo());

            uint joinHopCount =  overlayCtrlInfo->getHopCount();
            //std::cout << "-> " << jmsg->getSendStateTo().ip << " "
            //          << joinHopCount << " ";
            if ((joinHopCount > 1) &&
                ((defaultRoutingType == ITERATIVE_ROUTING) ||
                 (defaultRoutingType == EXHAUSTIVE_ITERATIVE_ROUTING)))
                 joinHopCount--;
            //std::cout << joinHopCount << std::endl;

            sendStateTables(jmsg->getSendStateTo(),
                            PASTRY_STATE_JOIN, joinHopCount, true);
        }

        delete jmsg;
    }
        break;

    case PASTRY_MSG_LEAFSET: {
        PastryLeafsetMessage* lmsg =
            check_and_cast<PastryLeafsetMessage*>(pastryMsg);
        RECORD_STATS(leafsetReceived++; leafsetBytesReceived +=
            lmsg->byteLength());

        if (state == DISCOVERY) {
            uint lsSize = lmsg->getLeafSetArraySize();
            const NodeHandle* node;
            pingedNodes = 0;

            for (uint i = 0; i < lsSize; i++) {
                node = &(lmsg->getLeafSet(i));
                // unspecified nodes not considered
                if ( !(node->isUnspecified()) ) {
                    pingNode(*node, discoveryTimeoutAmount, 0,
                             NULL, "PING received leafs for nearest node");
                    pingedNodes++;
               }
            }

            EV << "Pastry: received leafset, waiting for pings" << endl;
            scheduleAt(simTime()+discoveryTimeoutAmount, discoveryTimeout);
            delete lmsg;
        }

        else if (state == READY) {
            handleLeafsetMessage(lmsg, false);

        } else {
            delete lmsg;
        }
    }
        break;

    case PASTRY_MSG_ROWREQ: {
        PastryRoutingRowRequestMessage* rtrmsg =
            check_and_cast<PastryRoutingRowRequestMessage*>(pastryMsg);
        RECORD_STATS(routingTableReqReceived++; routingTableReqBytesReceived +=
            rtrmsg->byteLength());
        if (state == READY)
            if (rtrmsg->getRow() == -1)
                sendRoutingRow(rtrmsg->getSendStateTo(), routingTable->getLastRow());
                else if (rtrmsg->getRow() > routingTable->getLastRow())
                    EV << "Pastry: received request for nonexistent routing"
                       << "table row, dropping message!" << endl;
                else sendRoutingRow(rtrmsg->getSendStateTo(), rtrmsg->getRow());
        else
            EV << "Pastry: received routing table request before reaching "
               << "READY state, dropping message!" << endl;
        delete rtrmsg;
    }
        break;

    case PASTRY_MSG_RROW: {
        PastryRoutingRowMessage* rtmsg =
            check_and_cast<PastryRoutingRowMessage*>(pastryMsg);
        RECORD_STATS(routingTableReceived++; routingTableBytesReceived +=
            rtmsg->byteLength());

        if (state == DISCOVERY) {
            //int rtSize = rtmsg->getRoutingTableArraySize();
            uint nodesPerRow = routingTable->nodesPerRow;
            const NodeHandle* node;
            if (depth == -1) depth = rtmsg->getRow();
            pingedNodes = 0;
            nearNodeImproved = false;

            if (depth > 0) {
                for (uint i = 0; i < nodesPerRow; i++) {
                    node = &(rtmsg->getRoutingTable(i));
                    // unspecified nodes not considered
                    if ( !(node->isUnspecified()) ) {
                        // we look for best connection here, so Timeout is short and there are no retries
                        pingNode(*node, discoveryTimeoutAmount, 0,
                                 NULL, "PING received routing table for nearest node");
                        pingedNodes++;
                    }
                }
                depth--;
            }
            EV << "Pastry: received routing table, waiting for pings" << endl;
            scheduleAt(simTime()+discoveryTimeoutAmount, discoveryTimeout);

        }

        else if (state == READY) {

            uint nodesPerRow = routingTable->nodesPerRow;
            PastryStateMessage* stateMsg;


            stateMsg = new PastryStateMessage("STATE");
            stateMsg->setTimestamp(rtmsg->getTimestamp());
            stateMsg->setPastryMsgType(PASTRY_MSG_STATE);
            stateMsg->setStatType(MAINTENANCE_STAT);
            stateMsg->setSender(rtmsg->getSender());
            stateMsg->setLeafSetArraySize(0);
            stateMsg->setNeighborhoodSetArraySize(0);
            stateMsg->setRoutingTableArraySize(nodesPerRow);

            for (uint i = 0; i < nodesPerRow; i++) {
                stateMsg->setRoutingTable(i, rtmsg->getRoutingTable(i));
            }

            handleStateMessage(stateMsg);
        }

        delete rtmsg;
    }
        break;

    case PASTRY_MSG_REQ: {
        PastryRequestMessage* lrmsg =
            check_and_cast<PastryRequestMessage*>(pastryMsg);
        handleRequestMessage(lrmsg);
    }
        break;

    case PASTRY_MSG_STATE: {
        PastryStateMessage* stateMsg =
            check_and_cast<PastryStateMessage*>(msg);
        RECORD_STATS(stateReceived++; stateBytesReceived +=
                     stateMsg->byteLength());
        handleStateMessage(stateMsg);
    }
        break;
    }
}

void Pastry::doJoinUpdate(void)
{
    // send "update" state message to all nodes who sent us their state
    // during INIT, remove these from notifyList so they don't get our
    // state twice
    std::vector<TransportAddress>::iterator nListPos;
    if (!stReceived.empty()) {
        for (std::vector<PastryStateMsgHandle>::iterator it =
                 stReceived.begin(); it != stReceived.end(); ++it) {
            sendStateTables(it->msg->getSender(), PASTRY_STATE_UPDATE,
                            it->msg->getTimestamp());
            nListPos = find(notifyList.begin(), notifyList.end(),
                            it->msg->getSender());
            if (nListPos != notifyList.end()) {
                notifyList.erase(nListPos);
            }
            delete it->msg;
            delete it->prox;
        }
        stReceived.clear();
    }

    // send a normal STATE message to all remaining known nodes
    for (std::vector<TransportAddress>::iterator it =
             notifyList.begin(); it != notifyList.end(); it++) {
        if (*it != thisNode) sendStateTables(*it);
    }
    notifyList.clear();

    updateTooltip();
}

void Pastry::doSecondStage(void)
{
    parentModule()->parentModule()->bubble("entering SECOND STAGE");
    // "second stage" for locality:
    notifyList.clear();
    routingTable->dumpToVector(notifyList);
    neighborhoodSet->dumpToVector(notifyList);
    sort(notifyList.begin(), notifyList.end());
    notifyList.erase(unique(notifyList.begin(), notifyList.end()),
                     notifyList.end());
    for (std::vector<TransportAddress>::iterator it = notifyList.begin();
         it != notifyList.end(); it++) {
        if (*it == thisNode) continue;
        EV << "second stage: requesting state from " << *it << endl;
        sendRequest(*it, PASTRY_REQ_STATE);
    }
    notifyList.clear();
}

bool Pastry::handleFailedNode(const TransportAddress& failed)
{
    if (state != READY) {
        return false;
    }
    //std::cout << thisNode.ip << " is handling failed node: " << failed.ip << std::endl;
    if (failed.isUnspecified())
        opp_error("Pastry::handleFailedNode(): failed is unspecified!");

    const TransportAddress& lsAsk = leafSet->failedNode(failed);
    const TransportAddress& rtAsk = routingTable->failedNode(failed);
    neighborhoodSet->failedNode(failed);

    if (! lsAsk.isUnspecified()) {
        if (sendStateAtLeafsetRepair) sendRequest(lsAsk, PASTRY_REQ_REPAIR);
        else sendRequest(lsAsk, PASTRY_REQ_LEAFSET);
    }
    if (! rtAsk.isUnspecified() &&
        (lsAsk.isUnspecified() ||
         lsAsk != rtAsk)) sendRequest(rtAsk, PASTRY_REQ_REPAIR);

    newLeafs();

    if (lsAsk.isUnspecified() && (! leafSet->isValid())) {
        EV << "Pastry: lost connection to the network, trying to re-join."
           << endl;
        std::cout << "Pastry: lost connection to the network, trying to re-join."
                  << std::endl;
        joinOverlay();
        return false;
    }

    return true;
}

void Pastry::checkProxCache(void)
{
    // no cached STATE message?
    if (!(stateCache.msg && stateCache.prox)) return;

    // no entries in stateCache.prox?
    if (stateCache.prox->pr_rt.empty()) return;
    if (stateCache.prox->pr_ls.empty()) return;
    if (stateCache.prox->pr_ns.empty()) return;

    //debug
    /*for (uint i = 0; i < stateCache.prox->pr_rt.size(); ++i) {
        if (stateCache.prox->pr_rt[i] == -3)
            EV << stateCache.msg->getRoutingTable(i).ip << " ";
    }
    for (uint i = 0; i < stateCache.prox->pr_ls.size(); ++i) {
        if (stateCache.prox->pr_ls[i] == -3)
            EV << stateCache.msg->getLeafSet(i).ip << " ";
    }
    for (uint i = 0; i < stateCache.prox->pr_ns.size(); ++i) {
        if (stateCache.prox->pr_ns[i] == -3)
            EV << stateCache.msg->getNeighborhoodSet(i).ip << " ";
    }
    EV << endl;*/

    // some entries not yet determined?
    if (find(stateCache.prox->pr_rt.begin(), stateCache.prox->pr_rt.end(),
             PASTRY_PROX_PENDING) != stateCache.prox->pr_rt.end()) return;
    if (find(stateCache.prox->pr_ls.begin(), stateCache.prox->pr_ls.end(),
             PASTRY_PROX_PENDING) != stateCache.prox->pr_ls.end()) return;
    if (find(stateCache.prox->pr_ns.begin(), stateCache.prox->pr_ns.end(),
             PASTRY_PROX_PENDING) != stateCache.prox->pr_ns.end()) return;

    EV << "[Pastry::checkProxCache() @ " << thisNode.ip
       << " (" << thisNode.key.toString(16) << ")]\n"
       << "    all proximities for current STATE message from "
       << stateCache.msg->getSender().ip
       << " collected!"
       << endl;
/*
    //debug
    if (stateCache.prox != NULL) {
        std::vector<PastryStateMsgHandle>::iterator it;
        for (it = stReceived.begin(); it != stReceived.end(); ++it) {
            if (it->prox == NULL) {
                EV << ". " << endl;
                continue;
            }
            for (uint i = 0; i < it->prox->pr_rt.size(); ++i) {
                EV << it->prox->pr_rt[i] << " ";
            }
            for (uint i = 0; i < it->prox->pr_ls.size(); ++i) {
                EV << it->prox->pr_ls[i] << " ";
            }
            for (uint i = 0; i < it->prox->pr_ns.size(); ++i) {
                EV << it->prox->pr_ns[i] << " ";
            }
            EV << endl;
        }
        EV << endl;
    } else EV << "NULL" << endl;
*/

    simtime_t now = simTime();

    if (state == JOINING_2) {
        // save pointer to proximity vectors:
        stReceivedPos->prox = stateCache.prox;

        // collected proximities for all STATE messages?
        if (++stReceivedPos == stReceived.end()) {
            EV << "[Pastry::checkProxCache() @ " << thisNode.ip
               << " (" << thisNode.key.toString(16) << ")]\n"
               << "    proximities for all STATE messages collected!"
               << endl;
            stateCache.msg = NULL;
            stateCache.prox = NULL;
            if (debugOutput) {
                EV << "Pastry: [JOIN] starting to build own state from "
                   << stReceived.size() << " received state messages..."
                   << endl;
                EV << "Pastry: [JOIN] initializing NeighborhoodSet from "
                << stReceived.front().msg->getJoinHopCount() << ". hop"
                << endl;
            }
            if (!neighborhoodSet->mergeState( stReceived.front().msg,
                                             *stReceived.front().prox )) {
                EV << "Pastry: Error initializing own neighborhoodSet while "
                    "joining! Restarting ..." << endl;
                std::cout << "Pastry: Error initializing own neighborhoodSet while "
                             "joining! Restarting ..." << std::endl;
                joinOverlay();
                return;
            }
            if (debugOutput) {
                EV << "Pastry: [JOIN] initializing LeafSet from "
                << stReceived.back().msg->getJoinHopCount() << ". hop" << endl;
            }
            if (!leafSet->mergeState( stReceived.back().msg,
                                     *stReceived.back().prox )) {
                EV << "Pastry: Error initializing own leafSet while "
                    "joining! Restarting ..." << endl;
                std::cout << "Pastry: Error initializing own leafSet while "
                                    "joining! Restarting ..." << std::endl;
                joinOverlay();
                return;
            }
            if (debugOutput) {
                EV << "Pastry: [JOIN] initializing RoutingTable from all hops"
                   << endl;
            }
            if (!routingTable->initStateFromHandleVector(stReceived)) {
                EV << "Pastry: Error initializing own routingTable "
                    "while joining! Restarting ..." << endl;
                std::cout << "Pastry: Error initializing own routingTable "
                             "while joining! Restarting ..." << std::endl;
                joinOverlay();
                return;
            }
            lastStateChange = now;
            newLeafs();
            changeState(READY);
            EV << "changeState(READY) called" << endl;

        } else {
            EV << "[Pastry::checkProxCache() @ " << thisNode.ip
               << " (" << thisNode.key.toString(16) << ")]\n"
               << "    NOT all proximities for all STATE messages collected!"
               << endl;
            for (uint i = 0; i < stReceived.size(); ++i) {
                if (stReceived[i].msg == stReceivedPos->msg) EV << " *";
                EV << stReceived[i].msg;
            }
            EV << endl;

            // process next state message in vector:
            if (stReceivedPos->msg == NULL)
                throw new cRuntimeError("stReceivedPos->msg = NULL");
            stateCache = *stReceivedPos;
            if (stateCache.msg == NULL)
                throw new cRuntimeError("msg = NULL");
            pingNodes();
        }
    } else {
        // state == READY
        if (stateCache.msg->getPastryStateMsgType() == PASTRY_STATE_REPAIR) {
            // try to repair routingtable based on repair message:
            const TransportAddress& askRt =
                routingTable->repair(stateCache.msg, *stateCache.prox);
            if (! askRt.isUnspecified()) {
                sendRequest(askRt, PASTRY_REQ_REPAIR);
            }

            // while not really known, it's safe to assume that a repair
            // message changed our state:
            lastStateChange = now;
        } else {
            if (stateCache.outdatedUpdate) {
                // send another STATE message on outdated state update:
                sendStateDelayed(stateCache.msg->getSender());

            } else {//hack

                // merge info in own state tables
                // except leafset (was already handled in handleStateMessage)
                if (neighborhoodSet->mergeState(stateCache.msg, *stateCache.prox))
                    lastStateChange = now;
                EV << "Merging nodes into routing table" << endl;
                if (routingTable->mergeState(stateCache.msg, *stateCache.prox)) {
                    lastStateChange = now;
                    EV << "merged nodes into routing table" << endl;
                }
            }
        }
        updateTooltip();

        // if state message was not an update, send one back:
        if (stateCache.msg->getPastryStateMsgType() != PASTRY_STATE_UPDATE &&
            (alwaysSendUpdate || lastStateChange == now)) {//hack
            sendStateTables(stateCache.msg->getSender(), PASTRY_STATE_UPDATE,
                            stateCache.msg->getTimestamp());
        }

        delete stateCache.msg;
        stateCache.msg = NULL;
        delete stateCache.prox;
        stateCache.prox = NULL;

        // process next queued message:
        if (! stateCacheQueue.empty()) {
            stateCache = stateCacheQueue.front();
            stateCacheQueue.pop();
            pingNodes();
        }
    }
}

void Pastry::handleStateMessage(PastryStateMessage* msg)
{
    if (debugOutput) {
        EV << "Pastry::handleStateMessage() new STATE message to process "
           << static_cast<void*>(msg) << " in state " <<
            ((state == READY)?"READY":((state == JOINING_2)?"JOIN":"INIT"))
           << endl;
        if (state == JOINING_2) {
            EV << " ***   own joinHopCount:  " << joinHopCount << endl;
            EV << " ***   already received:  " << stReceived.size() << endl;
            EV << " ***   last-hop flag:     " << msg->getLastHop() << endl;
            EV << " ***   msg joinHopCount:  "
               << msg->getJoinHopCount() << endl;
        }
    }
    if (state == INIT) {
        EV << "Pastry: can't handle state messages until at least reaching "
            "JOIN state." << endl;
        delete msg;
        return;
    }

    PastryStateMsgHandle handle(msg);

    // in JOIN state, store all received state Messages, need them later:
    if (state == JOINING_2) {
        //std::cout << simTime() << " " << thisNode.ip << " "
        //          << msg->getJoinHopCount()
        //          << (msg->getLastHop() ? " *" : "") << std::endl;

        if (joinHopCount && stReceived.size() == joinHopCount) {
            EV << "Pastry: Warning: dropping state message received after "
                "all needed state messages were collected in JOIN state."
               << endl;
            delete msg;
            return;
        }

        stReceived.push_back(handle);
        prePing(msg);

        if (msg->getLastHop()) {
            if (joinTimeout->isScheduled()) cancelEvent(joinTimeout);
            if (msg->getSender().key == thisNode.key) {
                EV << "Pastry: Error: OverlayKey already in use, restarting!"
                   << endl;
                std::cout << "Pastry: Error: OverlayKey already in use, restarting!"
                                   << std::endl;
                joinOverlay();
                return;
            }

            if (joinHopCount) {
                EV << "Pastry: Error: received a second `last' state message! "
                    "Restarting ..." << endl;
                std::cout << thisNode.ip << "Pastry: Error: received a second `last' state message! "
                                    "Restarting ..." << std::endl;
                joinOverlay();
                return;
            }

            joinHopCount = msg->getJoinHopCount();

            if (stReceived.size() < joinHopCount) {
                // some states still missing:
                cancelEvent(readyWait);
                scheduleAt(simTime()+readyWaitAmount, readyWait);
            }
        }

        if (joinHopCount) {
            if (stReceived.size() > joinHopCount) {
                EV << "Pastry: Error: too many state messages received in "
                      "JOIN state! (" << stReceived.size()
                   << " > " << joinHopCount << ") Restarting ..." << endl;
                //std::cout << " failed!" << std::endl;
                joinOverlay();
                return;
            }
            if (stReceived.size() == joinHopCount) {
                // all state messages are here, sort by hopcount:
                sort(stReceived.begin(), stReceived.end(),
                     stateMsgIsSmaller);

                // start pinging the nodes found in the first state message:
                stReceivedPos = stReceived.begin();
                stateCache = *stReceivedPos;
                EV << "Pastry: have all STATE messages, now pinging nodes."
                   << endl;
                pingNodes();

                // cancel timeout:
                if (readyWait->isScheduled()) cancelEvent(readyWait);
            } else {
                //TODO occasionally, here we got a wrong hop count in
                // iterative mode due to more than one it. lookup during join
                // procedure
                EV << "[Pastry::handleStateMessage() @ " << thisNode.ip
                   << " (" << thisNode.key.toString(16) << ")]\n"
                   << "    Still need some STATE messages."
                   << endl;
            }

        }
        return;
    }

    if (debugOutput) {
        EV << "Pastry: handling STATE message" << endl;
        EV << "        type: " << ((msg->getPastryStateMsgType()
                                    == PASTRY_STATE_UPDATE) ? "update"
                                                            :"standard")
           << endl;
        if (msg->getPastryStateMsgType() == PASTRY_STATE_UPDATE) {
            EV << "        msg timestamp:      " <<
                msg->getTimestamp() << endl;
            EV << "        last state change:  " <<
                lastStateChange << endl;
        }
    }

    if (((msg->getPastryStateMsgType() == PASTRY_STATE_UPDATE))
            && (msg->getTimestamp() <= lastStateChange)) {
        // if we received an update based on our outdated state,
        // mark handle for retrying later:
        EV << "Pastry: outdated state from " << msg->getSender() << endl;
        handle.outdatedUpdate = true;
    }

    // determine aliveTable to prevent leafSet from merging nodes that are
    // known to be dead:
    determineAliveTable(msg);

    if (msg->getPastryStateMsgType() == PASTRY_STATE_REPAIR) {
        // try to repair leafset based on repair message right now
        const TransportAddress& askLs = leafSet->repair(msg, aliveTable);
        if (! askLs.isUnspecified()) {
            sendRequest(askLs, PASTRY_REQ_REPAIR);
        }

        // while not really known, it's safe to assume that a repair
        // message changed our state:
        lastStateChange = simTime();
        newLeafs();
    } else if (leafSet->mergeState(msg, aliveTable)) {
        // merged state into leafset right now
        lastStateChange = simTime();
        newLeafs();
        updateTooltip();
    }
    // in READY state, only ping nodes to get proximity metric:
    if (!stateCache.msg) {
        // no state message is processed right now, start immediately:
        stateCache = handle;
        pingNodes();
    } else {
        // enqueue message for later processing:
        stateCacheQueue.push(handle);
        prePing(msg);
    }
}
