//
// Copyright (C) 2006 Institut fuer Telematik, Universitaet Karlsruhe (TH)
//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public License
// as published by the Free Software Foundation; either version 2
// of the License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
//

/**
* @file ConnectivityProbe.cc
* @author Helge Backhaus
*/

#include "ConnectivityProbe.h"

Define_Module(ConnectivityProbe);

VTopologyNode::VTopologyNode(int moduleID)
{
    this->moduleID = moduleID;
    visited = false;
}

Vast* VTopologyNode::module() const
{
    return check_and_cast<Vast*>(simulation.module(moduleID));
}

void ConnectivityProbe::initialize()
{
    probeIntervall = par("connectivityProbeIntervall");
    plotIntervall = par("visualizeNetworkIntervall");
    probeTimer = new cMessage("probeTimer");
    plotTimer = new cMessage("plotTimer");

    if(probeIntervall > 0.0) {
        scheduleAt(simulation.simTime() + probeIntervall, probeTimer);

        cOV_NodeCount.setName("node count");
        cOV_MaximumComponent.setName("largest connected component");
        cOV_MaxConnectivity.setName("connectivity");
        cOV_ZeroMissingNeighbors.setName("nodes with 0 missing neighbors");
        cOV_OneMissingNeighbor.setName("nodes with 1 missing neighbor");
        cOV_TwoMissingNeighbor.setName("nodes with 2 missing neighbor");
        cOV_ThreeMissingNeighbor.setName("nodes with 3 missing neighbor");
        cOV_FourMissingNeighbor.setName("nodes with 4 missing neighbor");
        cOV_FiveMissingNeighbors.setName("nodes with 5 missing neighbors");
        cOV_SeveralMissingNeighbors.setName("nodes with more than 5 missing neighbors");
        cOV_AverageMissingNeighbors.setName("average missing neighbors per node");
        cOV_MaxMissingNeighbors.setName("largest missing neighbors count");
    }

    if(plotIntervall > 0.0) {
        scheduleAt(simulation.simTime() + plotIntervall, plotTimer);
    }
}

void ConnectivityProbe::handleMessage(cMessage* msg)
{
    // fill topology with all VAST modules
    extractTopology();

    if(Topology.size() == 0) {
        return;
    }

    // catch self timer messages
    if(msg->isName("probeTimer")) {
        //reset timer
        cancelEvent(probeTimer);
        scheduleAt(simulation.simTime() + probeIntervall, msg);

        unsigned int maxComponent = 0;
        unsigned int connectivity = 0;
        for(VTopology::iterator itTopology = Topology.begin(); itTopology != Topology.end(); ++itTopology) {
            unsigned int count = getComponentSize(itTopology->second.module()->getThisNode().key);
            if(count > maxComponent) {
                maxComponent = count;
            }
            if(count == Topology.size()) {
                ++connectivity;
            }
            resetTopologyNodes();
        }

        cOV_NodeCount.record((double)Topology.size());
        cOV_MaximumComponent.record((double)maxComponent);
        cOV_MaxConnectivity.record((double)connectivity);

        int mnZero, mnOne, mnTwo, mnThree, mnFour, mnFive, mnSeveral, mnMax;
        double mnAverage = 0.0;
        double AOIWidthSqr = Topology.begin()->second.module()->getAOI();
        AOIWidthSqr *= AOIWidthSqr;
        mnZero = mnOne = mnTwo = mnThree = mnFour = mnFive = mnSeveral = mnMax = 0;

        for(VTopology::iterator itTopology = Topology.begin(); itTopology != Topology.end(); ++itTopology) {
            NodeSet SiteHandles;
            Vector2D vastPosition = itTopology->second.module()->getPosition();
            int missing = 0;
            for(VTopology::iterator itI = Topology.begin(); itI != Topology.end(); ++itI) {
                if(itI != itTopology && vastPosition.distanceSqr(itI->second.module()->getPosition()) <= AOIWidthSqr) {
                    SiteHandles.insert(itI->second.module()->getThisNode());
                }
            }

            for(NodeSet::iterator itSiteHandles = SiteHandles.begin(); itSiteHandles != SiteHandles.end(); ++itSiteHandles) {
                if(itTopology->second.module()->getList()->Sites.find(*itSiteHandles) == itTopology->second.module()->getList()->Sites.end()) {
                    ++missing;
                }
            }

            switch(missing) {
                case 0:
                    ++mnZero;
                break;
                case 1:
                    ++mnOne;
                break;
                case 2:
                    ++mnTwo;
                break;
                case 3:
                    ++mnThree;
                break;
                case 4:
                    ++mnFour;
                break;
                case 5:
                    ++mnFive;
                break;
                default:
                    ++mnSeveral;
            }
            mnAverage += missing;
            if(mnMax < missing) {
                mnMax = missing;
            }
        }
        mnAverage /= (double)Topology.size();

        cOV_ZeroMissingNeighbors.record((double)mnZero);
        cOV_OneMissingNeighbor.record((double)mnOne);
        cOV_TwoMissingNeighbor.record((double)mnTwo);
        cOV_ThreeMissingNeighbor.record((double)mnThree);
        cOV_FourMissingNeighbor.record((double)mnFour);
        cOV_FiveMissingNeighbors.record((double)mnFive);
        cOV_SeveralMissingNeighbors.record((double)mnSeveral);
        cOV_AverageMissingNeighbors.record(mnAverage);
        cOV_MaxMissingNeighbors.record((double)mnMax);
    }
    else if(msg->isName("plotTimer")) {
         //reset timer
        cancelEvent(plotTimer);
        scheduleAt(simulation.simTime() + plotIntervall, msg);

        int range = par("areaDimension");;
        std::stringstream oss;
        std::string filename;
        int simTime, stellen = 1;
        simTime = (int)simulation.simTime();
        oss << "plot";
        for(int i=0; i<6; i++) {
            if(!(simTime / stellen)) {
                oss << "0";
            }
            stellen *= 10;
        }
        oss << simTime;

        // write plot file
        filename = oss.str() + ".plt";
        pltNetwork.open(filename.c_str(), std::ios::out);
        pltNetwork << "set noarrow" << endl;
        pltNetwork << "set style line 1 linetype 7 linecolor rgb \"red\"" << endl;
        pltNetwork << "set style line 2 linetype 7 linecolor rgb \"green\"" << endl;
        pltNetwork << "set style line 3 linetype 7 linecolor rgb \"blue\"" << endl;
        pltNetwork << "set style line 4 linetype 7 linecolor rgb \"black\"" << endl;
        pltNetwork << "set xrange [0:" << range << "]" << endl;
        pltNetwork << "set yrange [0:" << range << "]" << endl;
        for(VTopology::iterator itTopology = Topology.begin(); itTopology != Topology.end(); ++itTopology) {
            for(SiteMap::iterator itSites = itTopology->second.module()->getList()->Sites.begin(); itSites != itTopology->second.module()->getList()->Sites.end(); ++itSites) {
                pltNetwork << "set arrow from " << itTopology->second.module()->getPosition().x << "," << itTopology->second.module()->getPosition().y <<
                              " to " << itSites->second.coord.x << "," << itSites->second.coord.y << " ls ";
                if(itSites->second.type & NEIGHBOR) {
                    if(itSites->second.type & ENCLOSING) {
                        pltNetwork << "3" << endl;
                    }
                    else {
                        pltNetwork << "1" << endl;
                    }
                }
                else if(itSites->second.type & BOUNDARY) {
                    if(itSites->second.type & ENCLOSING) {
                        pltNetwork << "4" << endl;
                    }
                    else {
                        pltNetwork << "2" << endl;
                    }
                }
            }
        }
        filename = oss.str() + ".dat";
        pltNetwork << "plot '" << filename << "' using 1:2 with points pointtype 7" << endl;
        pltNetwork.close();

        // write point data file
        pltData.open(filename.c_str(), std::ios::out);
        for(VTopology::iterator itTopology = Topology.begin(); itTopology != Topology.end(); ++itTopology) {
            pltData << itTopology->second.module()->getPosition().x << "\t" << itTopology->second.module()->getPosition().y << endl;
        }
        pltData.close();
    }
    Topology.clear();
}

void ConnectivityProbe::extractTopology()
{
    for(int i=0; i<=simulation.lastModuleId(); i++) {
        cModule* module = simulation.module(i);
        if(module && dynamic_cast<Vast*>(module)) {
            Vast* vast = check_and_cast<Vast*>(module);
            if(vast->getState() == BaseOverlay::READY) {
                VTopologyNode temp(i);
                Topology.insert(std::make_pair(vast->getThisNode().key, temp));
            }
        }
    }
}

void ConnectivityProbe::resetTopologyNodes()
{
    for(VTopology::iterator itTopology = Topology.begin(); itTopology != Topology.end(); ++itTopology) {
        itTopology->second.visited = false;
    }
}

unsigned int ConnectivityProbe::getComponentSize(OverlayKey key)
{
    VTopology::iterator itEntry = Topology.find(key);
    if(itEntry != Topology.end() && itEntry->second.visited == false) {
        int count = 1;
        itEntry->second.visited = true;
        Vast* vast = itEntry->second.module();
        for(SiteMap::iterator itSites = vast->getList()->Sites.begin(); itSites != vast->getList()->Sites.end(); ++itSites) {
            count += getComponentSize(itSites->first.key);
        }
        return count;
    }
    return 0;
}

ConnectivityProbe::~ConnectivityProbe()
{
    // destroy self timer messages
    cancelAndDelete(probeTimer);
    cancelAndDelete(plotTimer);
}
