//
// Copyright (C) 2006 Institut fuer Telematik, Universitaet Karlsruhe (TH)
//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public License
// as published by the Free Software Foundation; either version 2
// of the License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
//

/**
 * @file Vast.cc
 * @author Helge Backhaus
 */

#include "Vast.h"

Define_Module(Vast);

#include <NotifierConsts.h>

#include <BootstrapOracle.h>
#include <GlobalStatistics.h>
#include <BootstrapList.h>

void Vast::initializeOverlay(int stage)
{
    // because of IPAddressResolver, we need to wait until interfaces are registered,
    // address auto-assignment takes place etc.
    if(stage != MIN_STAGE_OVERLAY) return;

    // fetch parameters
    debugVoronoiOutput = par("debugVastOutput");
    ignoreFalseKeys = par("ignoreFalseKeys");
    AOI_size = par("AOIWidth");
    joinTimeout = par("joinTimeout");
    pingTimeout = par("pingTimeout");

    // find friend modules
    neighborsList = check_and_cast<NeighborsList*>(parentModule()->submodule("neighborsList"));

    // set node key
    thisNode.key = OverlayKey::random();

    // initialize neighborsList
    neighborsList->setDebug(debugVoronoiOutput);

    // self-messages
    join_timer = new cMessage("join_timer");
    ping_timer = new cMessage("ping_timer");
    sec_timer = new cMessage("sec_timer");

    // statistics
    joinRequestBytesSent = 0;
    joinAcknowledgeBytesSent = 0;
    nodeMoveBytesSent = 0;
    newNeighborsBytesSent = 0;
    nodeLeaveBytesSent = 0;
    enclosingNeighborsRequestBytesSent = 0;
    pingBytesSent = 0;
    pongBytesSent = 0;
    discardNodeBytesSent = 0;

    maxBytesPerSecondSent = 0;
    averageBytesPerSecondSent = 0;
    bytesPerSecond = 0;

    // watch some variables
    WATCH(position);
    WATCH(thisNode);
    WATCH(AOI_size);

    WATCH(joinRequestBytesSent);
    WATCH(joinAcknowledgeBytesSent);
    WATCH(nodeMoveBytesSent);
    WATCH(newNeighborsBytesSent);
    WATCH(nodeLeaveBytesSent);
    WATCH(enclosingNeighborsRequestBytesSent);
    WATCH(pingBytesSent);
    WATCH(pongBytesSent);
    WATCH(discardNodeBytesSent);

    WATCH(maxBytesPerSecondSent);
    WATCH(bytesPerSecond);

    // set initial state
    changeState(INIT);
    changeState(JOINING);
}

void Vast::changeState(int state)
{
    switch(state) {
        case INIT: {
            this->state = INIT;
            bootstrapOracle->removePeer(thisNode);
        } break;
        case JOINING: {
            this->state = JOINING;
            scheduleAt(simulation.simTime(), join_timer);
            scheduleAt(simulation.simTime() + 1.0, sec_timer);
        } break;
        case READY: {
            this->state = READY;
            cancelEvent(join_timer);
            neighborsList->initializeList(position, thisNode, AOI_size);
            bootstrapOracle->registerPeer(thisNode);
            scheduleAt(simulation.simTime() + pingTimeout, ping_timer);
            // tell the application we are ready
            CompReadyMessage* readyMsg = new CompReadyMessage("OVERLAY_READY");
            readyMsg->setReady(true);
            readyMsg->setComp(getThisCompType());
            sendToApp(readyMsg);
            GameAPIResizeAOIMessage* gameMsg = new GameAPIResizeAOIMessage("RESIZE_AOI");
            gameMsg->setCommand(RESIZE_AOI);
            gameMsg->setAOIsize(AOI_size);
            sendToApp(gameMsg);
        } break;
    }
    setBootstrapedIcon();
    // debug output
    if(debugOutput) {
        ev << "VAST: Node " << thisNode.ip << " entered ";
        switch(state) {
            case INIT: ev << "INIT"; break;
            case JOINING: ev << "JOINING"; break;
            case READY: ev << "READY"; break;
        }
        ev << " state." << endl;
    }
}

void Vast::handleTimerEvent(cMessage* msg)
{
    if(msg->isName("join_timer")) {
        //reset timer
        cancelEvent(join_timer);
        scheduleAt(simulation.simTime() + joinTimeout, msg);
        // handle event
        processJoinTimer();
    }
    else if(msg->isName("ping_timer")) {
        //reset timer
        cancelEvent(ping_timer);
        scheduleAt(simulation.simTime() + pingTimeout, msg);
        // handle event
        processPingTimer();
    }
    else if(msg->isName("sec_timer")) {
        //reset timer
        cancelEvent(sec_timer);
        scheduleAt(simulation.simTime() + 1, msg);
        // handle event
        processSecTimer();
    }
}

void Vast::handleAppMessage(cMessage* msg)
{
    if(dynamic_cast<GameAPIMessage*>(msg)) {
        GameAPIMessage* gameAPIMsg = check_and_cast<GameAPIMessage*>(msg);
        // debug output
        if(debugOutput) ev << "VAST: Node " << thisNode.ip << " received " << gameAPIMsg->name() << " from application." << endl;
        switch(gameAPIMsg->getCommand()) {
            case MOVEMENT_INDICATION: {
                GameAPIPositionMessage* gameAPIPosMsg = check_and_cast<GameAPIPositionMessage*>(msg);
                if(state == JOINING) {
                    handleJoin(gameAPIPosMsg);
                    delete msg;
                }
                else if(state == READY) {
                    handleMove(gameAPIPosMsg);
                    delete msg;
                }
            } break;
            case GAMEEVENT_CHAT:
            case GAMEEVENT_SNOW:
            case GAMEEVENT_FROZEN:
                handleEvent( gameAPIMsg );
                delete msg;
                break;
            //case EVENT: {
                // insert handling for additional application messages like events etc. here
            //} break;
            default: {
                delete msg;
            }
        }
    }
    else delete msg;
}

void Vast::handleUDPMessage(BaseOverlayMessage* msg)
{
    if(state == INIT) {
        delete msg;
        return;
    }
    if(dynamic_cast<VastMessage*>(msg)) {
        VastMessage* vastMsg = check_and_cast<VastMessage*>(msg);
        if(ignoreFalseKeys || vastMsg->getDestKey() == thisNode.key) {
            // debug output
            if(debugOutput) ev << "VAST: Node " << thisNode.ip << " received " << vastMsg->name() << " from " << vastMsg->getSourceNode().ip << "." << endl;
            bool doUpdate = true;
            if(state == READY) {
                switch(vastMsg->getCommand()) {
                    case JOIN_REQUEST: {
                        handleJoinRequest(vastMsg);
                        doUpdate = false;
                    } break;
                    case NODE_MOVE: {
                        VastMoveMessage* vastMoveMsg = check_and_cast<VastMoveMessage*>(msg);
                        handleNodeMove(vastMoveMsg);
                    } break;
                    case NEW_NEIGHBORS: {
                        VastListMessage* vastListMsg = check_and_cast<VastListMessage*>(msg);
                        handleNewNeighbors(vastListMsg);
                    } break;
                    case NODE_LEAVE: {
                        VastListMessage* vastListMsg = check_and_cast<VastListMessage*>(msg);
                        handleNodeLeave(vastListMsg);
                    } break;
                    case ENCLOSING_NEIGHBORS_REQUEST: {
                        handleEnclosingNeighborsRequest(vastMsg);
                    } break;
                    case PING: {
                        handlePing(vastMsg);
                    } break;
                    case PONG: {
                        handlePong(vastMsg);
                    } break;
                    case DISCARD_NODE: {
                        VastDiscardMessage* vastDiscardMsg = check_and_cast<VastDiscardMessage*>(msg);
                        handleDiscardNode(vastDiscardMsg);
                    } break;
                    case VAST_EVENT: {
                        sendToApp( vastMsg->decapsulate() );
                        doUpdate = false;
                        delete vastMsg;
                    } break;
                }
                // update timestamp
                if(doUpdate) {
                    neighborsList->updateTimeStamp(vastMsg->getSourceNode());
                    delete msg;
                }
            }
            else if(state == JOINING && vastMsg->getCommand() == JOIN_ACKNOWLEDGE) {
                VastListMessage* vastListMsg = check_and_cast<VastListMessage*>(msg);
                handleJoinAcknowledge(vastListMsg);
                delete msg;
            }
            else delete msg;
        }
        else {
            sendDiscardNode(vastMsg);
            delete msg;
        }
    }
    else delete msg;
}

void Vast::receiveChangeNotification(int category, cPolymorphic *details)
{
    if(category == NF_OVERLAY_NODE_GRACEFUL_LEAVE && state == READY) {
        int i;
        // debug output
        if(debugOutput) {
            ev << "VAST: Node " << thisNode.ip << " is leaving the overlay." << endl;
        }

        // generate node leave messages
        VastListMessage *vastListMsg = new VastListMessage("NODE_LEAVE");
        vastListMsg->setCommand(NODE_LEAVE);
        vastListMsg->setSourceNode(thisNode);
        vastListMsg->setPos(position);
        // fill neighbors list
        vastListMsg->setNeighborNodeArraySize(neighborsList->getSize());
        vastListMsg->setNeighborPosArraySize(neighborsList->getSize());
        i = 0;
        SiteMap::iterator itSites = neighborsList->Sites.begin();
        while(itSites != neighborsList->Sites.end()) {
            if(itSites->second.type & ENCLOSING) {
                vastListMsg->setNeighborNode(i, itSites->second.addr);
                vastListMsg->setNeighborPos(i, itSites->second.coord);
                ++i;
            }
            ++itSites;
        }
        vastListMsg->setNeighborNodeArraySize(i);
        vastListMsg->setNeighborPosArraySize(i);
        vastListMsg->setRequestEnclosingNeighbors(false);

        vastListMsg->setLength(VASTLIST_L(vastListMsg));
        if(vastListMsg->getNeighborNodeArraySize() > 0) {
            itSites = neighborsList->Sites.begin();
            while(itSites != neighborsList->Sites.end()) {
                VastListMessage *vastCopyMsg = new VastListMessage(*vastListMsg);
                // prevent notification board from taking the message
               take(vastCopyMsg);
               sendMessage(vastCopyMsg, itSites->second.addr);
               ++itSites;
            }
        }
        delete vastListMsg;

        CompReadyMessage* readyMsg = new CompReadyMessage("OVERLAY_FINISHED");
        readyMsg->setReady(false);
        readyMsg->setComp(getThisCompType());
        take(readyMsg);
        sendToApp(readyMsg);

        changeState(INIT);
    }
}

void Vast::processJoinTimer()
{
    GameAPIMessage *sgcMsg = new GameAPIMessage("MOVEMENT_REQUEST");
    sgcMsg->setCommand(MOVEMENT_REQUEST);
    sendToApp(sgcMsg);
}

void Vast::processPingTimer()
{
    bool abnormalLeave = false;
    bool boundaryLeave = false;
    SiteMap::iterator itSites = neighborsList->Sites.begin();
    while(itSites != neighborsList->Sites.end()) {
        if(itSites->second.tstamp < 0.0) { // node is dropped cause no pong has been received see below
            abnormalLeave = true;
            if(!(itSites->second.type & NEIGHBOR)) boundaryLeave = true;
            itSites->second.type = UNDEF;
        }
        else if(itSites->second.tstamp < simulation.simTime() - pingTimeout) { // node showed no activity for some time request pong and mark it to be dropped next time
            VastMessage *vastMsg = new VastMessage("PING");
            vastMsg->setCommand(PING);
            vastMsg->setSourceNode(thisNode);
            vastMsg->setPos(position);
            vastMsg->setLength(VAST_L(vastMsg));
            sendMessage(vastMsg, itSites->second.addr);
            itSites->second.tstamp = -1.0;
        }
        ++itSites;
    }
    if(abnormalLeave) {
        synchronizeApp();
        neighborsList->removeNeighbors();
        if(boundaryLeave) {
            itSites = neighborsList->Sites.begin();
            while(itSites != neighborsList->Sites.end()) {
                if(itSites->second.type & BOUNDARY) {
                    VastMessage *vastMsg = new VastMessage("ENCLOSING_NEIGHBORS_REQUEST");
                    vastMsg->setCommand(ENCLOSING_NEIGHBORS_REQUEST);
                    vastMsg->setSourceNode(thisNode);
                    vastMsg->setPos(position);
                    vastMsg->setLength(VAST_L(vastMsg));
                    sendMessage(vastMsg, itSites->second.addr);
                }
                ++itSites;
            }
        }
        neighborsList->buildVoronoi();
        //neighborsList->removeNeighbors(); should be superfluous
    }
}

void Vast::processSecTimer()
{
    if(bytesPerSecond > maxBytesPerSecondSent) {
        maxBytesPerSecondSent = bytesPerSecond;
    }
    averageBytesPerSecondSent += bytesPerSecond;
    bytesPerSecond = 0;
}

void Vast::handleJoin(GameAPIPositionMessage *sgcMsg)
{
    NodeHandle joinNode = bootstrapList->getBootstrapNode();
    position = sgcMsg->getPosition();
    // check if this is the only node in the overlay
    if(joinNode.isUnspecified()) {
        changeState(READY);
    }
    else {
        VastMessage *vastMsg = new VastMessage("JOIN_REQUEST");
        vastMsg->setCommand(JOIN_REQUEST);
        vastMsg->setSourceNode(thisNode);
        vastMsg->setPos(position);
        vastMsg->setLength(VAST_L(vastMsg));
        sendMessage(vastMsg, joinNode);
    }
}

void Vast::handleMove(GameAPIPositionMessage* sgcMsg)
{
    Vector2D pos = sgcMsg->getPosition();
    // test if new position is legal
    if(!neighborsList->validatePosition(pos)) {
        GameAPIMessage *gameMsg = new GameAPIMessage("MOVEMENT_REQUEST");
        gameMsg->setCommand(MOVEMENT_REQUEST);
        sendToApp(gameMsg);
        return;
    }
    // send position update to neighbors
    SiteMap::iterator itSites = neighborsList->Sites.begin();
    while(itSites != neighborsList->Sites.end()) {
        VastMoveMessage *vastMoveMsg = new VastMoveMessage("NODE_MOVE");
        vastMoveMsg->setCommand(NODE_MOVE);
        vastMoveMsg->setSourceNode(thisNode);
        vastMoveMsg->setPos(position);
        vastMoveMsg->setNewPos(pos);
        vastMoveMsg->setIs_boundary(itSites->second.type & BOUNDARY);
        vastMoveMsg->setLength(VASTMOVE_L(vastMoveMsg));
        sendMessage(vastMoveMsg, itSites->second.addr);
        ++itSites;
    }
    // set new position
    position = pos;
    neighborsList->changePosition(pos);
    // update voronoi
    neighborsList->buildVoronoi();
    synchronizeApp();
    neighborsList->removeNeighbors();
}

void Vast::handleEvent( GameAPIMessage* msg )
{
    // send evbent to neighbors
    SiteMap::iterator itSites = neighborsList->Sites.begin();
    while(itSites != neighborsList->Sites.end()) {
        VastEventMessage *vastMsg = new VastEventMessage("EVENT");
        vastMsg->setCommand(VAST_EVENT);
        vastMsg->encapsulate( (cMessage*) msg->dup() );
        // FIXME: Message length!
        sendMessage(vastMsg, itSites->second.addr);
        ++itSites;
    }
}

void Vast::handleJoinRequest(VastMessage *vastMsg)
{
    Site *forwardSite = NULL;
    double min_dist;
    // start with this node
    min_dist = distance_square(neighborsList->thisSite, vastMsg);
    forwardSite = &neighborsList->thisSite;
    // iterate through all neighbors
    SiteMap::iterator itSites = neighborsList->Sites.begin();
    while(itSites != neighborsList->Sites.end()) {
        if(distance_square(itSites->second, vastMsg) < min_dist) {
            min_dist = distance_square(itSites->second, vastMsg);
            forwardSite = &itSites->second;
        }
        ++itSites;
    }
    // do nothing and let node retry with new position if current position is illegal
    if(min_dist == 0.0) {
        delete vastMsg;
    }
    else {
        // send an acknowledge or forward request if any of our neighbors is closer to joining node
        if(forwardSite->type & THIS) {
            VastListMessage *vastListMsg = new VastListMessage("JOIN_ACKNOWLEDGE");
            vastListMsg->setCommand(JOIN_ACKNOWLEDGE);
            vastListMsg->setSourceNode(thisNode);
            vastListMsg->setPos(position);
            // fill neighbors list
            vastListMsg->setNeighborNodeArraySize(neighborsList->getSize());
            vastListMsg->setNeighborPosArraySize(neighborsList->getSize());
            int i = 0;
            itSites = neighborsList->Sites.begin();
            while(itSites != neighborsList->Sites.end()) {
                vastListMsg->setNeighborNode(i, itSites->second.addr);
                vastListMsg->setNeighborPos(i, itSites->second.coord);
                ++i;
                ++itSites;
            }
            vastListMsg->setNeighborNodeArraySize(i);
            vastListMsg->setNeighborPosArraySize(i);
            vastListMsg->setRequestEnclosingNeighbors(false);

            vastListMsg->setLength(VASTLIST_L(vastListMsg));
            sendMessage(vastListMsg, vastMsg->getSourceNode());
            delete vastMsg;
        }
        else {
            sendMessage(vastMsg, forwardSite->addr);
        }
    }
}

void Vast::handleJoinAcknowledge(VastListMessage *vastListMsg)
{
    Vector2D p, p_join;
    p_join = vastListMsg->getPos();
    // add acceptor node
    changeState(READY);
    neighborsList->addNode(p_join, vastListMsg->getSourceNode());
    // add new neighbors
    for(unsigned int i=0; i<vastListMsg->getNeighborNodeArraySize(); i++) {
        p = vastListMsg->getNeighborPos(i);
        neighborsList->addNode(p, vastListMsg->getNeighborNode(i));
    }
    // update voronoi with new neighbors
    neighborsList->buildVoronoi();
    synchronizeApp();
    neighborsList->removeNeighbors();
    // contact new neighbors
    SiteMap::iterator itSites = neighborsList->Sites.begin();
    while(itSites != neighborsList->Sites.end()) {
        VastMoveMessage *vastMoveMsg = new VastMoveMessage("NODE_MOVE");
        vastMoveMsg->setCommand(NODE_MOVE);
        vastMoveMsg->setSourceNode(thisNode);
        vastMoveMsg->setPos(p_join);
        vastMoveMsg->setNewPos(position);
        vastMoveMsg->setIs_boundary(itSites->second.type & BOUNDARY);
        vastMoveMsg->setLength(VASTMOVE_L(vastMoveMsg));
        sendMessage(vastMoveMsg, itSites->second.addr);
        ++itSites;
    }
}

void Vast::handleNodeMove(VastMoveMessage *vastMoveMsg)
{
    RECORD_STATS(
            globalStatistics->addStdDev(
                "Vast: MoveDelay",
                simulation.simTime() - vastMoveMsg->creationTime()
                );
            );

    Vector2D old_p, new_p;
    old_p = vastMoveMsg->getPos();
    new_p = vastMoveMsg->getNewPos();
    neighborsList->addNode(new_p, vastMoveMsg->getSourceNode());
    // update voronoi with new neighbor detection or without
    if(vastMoveMsg->getIs_boundary()) {
        neighborsList->buildVoronoi(old_p, new_p);
        synchronizeApp(vastMoveMsg);
        neighborsList->removeNeighbors();
        // send new neighbors
        VastListMessage *vastListMsg = new VastListMessage("NEW_NEIGHBORS");
        vastListMsg->setCommand(NEW_NEIGHBORS);
        vastListMsg->setSourceNode(thisNode);
        vastListMsg->setPos(position);

        vastListMsg->setNeighborNodeArraySize(neighborsList->getSize());
        vastListMsg->setNeighborPosArraySize(neighborsList->getSize());

        int i = 0;
        SiteMap::iterator itSites = neighborsList->Sites.begin();
        while(itSites != neighborsList->Sites.end()) {
            if(itSites->second.type & NEW) {
                vastListMsg->setNeighborNode(i, itSites->second.addr);
                vastListMsg->setNeighborPos(i, itSites->second.coord);
                ++i;
            }
            ++itSites;
        }
        vastListMsg->setNeighborNodeArraySize(i);
        vastListMsg->setNeighborPosArraySize(i);
        vastListMsg->setRequestEnclosingNeighbors(true);

        vastListMsg->setLength(VASTLIST_L(vastListMsg));
        if(vastListMsg->getNeighborNodeArraySize() > 0) {
            sendMessage(vastListMsg, vastMoveMsg->getSourceNode());
        }
        else {
            delete vastListMsg;
        }
    }
    else {
        neighborsList->buildVoronoi();
        synchronizeApp(vastMoveMsg);
        neighborsList->removeNeighbors();
    }
}

void Vast::handleNewNeighbors(VastListMessage *vastListMsg)
{
    Vector2D p;
    // add new neighbors
    for(unsigned int i=0; i<vastListMsg->getNeighborNodeArraySize(); i++) {
        p = vastListMsg->getNeighborPos(i);
        neighborsList->addNode(p, vastListMsg->getNeighborNode(i));

        if(vastListMsg->getRequestEnclosingNeighbors()) {
            VastMessage *vastMsg = new VastMessage("ENCLOSING_NEIGHBORS_REQUEST");
            vastMsg->setCommand(ENCLOSING_NEIGHBORS_REQUEST);
            vastMsg->setSourceNode(thisNode);
            vastMsg->setPos(position);
            vastMsg->setLength(VAST_L(vastMsg));
            sendMessage(vastMsg, vastListMsg->getNeighborNode(i));
        }
    }
    // update voronoi with new neighbors
    neighborsList->buildVoronoi();
    synchronizeApp();
    neighborsList->removeNeighbors();
}

void Vast::handleNodeLeave(VastListMessage *vastListMsg)
{
    Vector2D p;
    neighborsList->removeNode(vastListMsg->getSourceNode());
    // add possible new neighbors
    for(unsigned int i=0; i<vastListMsg->getNeighborNodeArraySize(); i++) {
        p = vastListMsg->getNeighborPos(i);
        neighborsList->addNode(p, vastListMsg->getNeighborNode(i));
    }
    // update voronoi with new neighbors
    neighborsList->buildVoronoi();
    synchronizeApp();
    neighborsList->removeNeighbors();
}

void Vast::handleEnclosingNeighborsRequest(VastMessage *vastMsg)
{
    // send new neighbors
    VastListMessage *vastListMsg = new VastListMessage("NEW_NEIGHBORS");
    vastListMsg->setCommand(NEW_NEIGHBORS);
    vastListMsg->setSourceNode(thisNode);
    vastListMsg->setPos(position);

    vastListMsg->setNeighborNodeArraySize(neighborsList->getSize());
    vastListMsg->setNeighborPosArraySize(neighborsList->getSize());

    int i = 0;
    SiteMap::iterator itSites = neighborsList->Sites.begin();
    while(itSites != neighborsList->Sites.end()) {
        if((itSites->second.type & ENCLOSING) && itSites->second.addr != vastMsg->getSourceNode()) {
            vastListMsg->setNeighborNode(i, itSites->second.addr);
            vastListMsg->setNeighborPos(i, itSites->second.coord);
            ++i;
        }
        ++itSites;
    }
    vastListMsg->setNeighborNodeArraySize(i);
    vastListMsg->setNeighborPosArraySize(i);
    vastListMsg->setRequestEnclosingNeighbors(false);

    vastListMsg->setLength(VASTLIST_L(vastListMsg));
    if(vastListMsg->getNeighborNodeArraySize() > 0) {
        sendMessage(vastListMsg, vastMsg->getSourceNode());
    }
    else {
        delete vastListMsg;
    }
}

void Vast::handlePing(VastMessage *vastMsg)
{
    VastMessage *vastPongMsg = new VastMessage("PONG");
    vastPongMsg->setCommand(PONG);
    vastPongMsg->setSourceNode(thisNode);
    vastPongMsg->setPos(position);
    vastPongMsg->setLength(VAST_L(vastPongMsg));
    sendMessage(vastPongMsg, vastMsg->getSourceNode());
}

void Vast::handlePong(VastMessage *vastMsg)
{
    Vector2D p;
    // replace entry cause it was probably outdated
    p = vastMsg->getPos();
    neighborsList->addNode(p, vastMsg->getSourceNode());
    // update voronoi
    neighborsList->buildVoronoi();
    synchronizeApp();
    neighborsList->removeNeighbors();
}

void Vast::handleDiscardNode(VastDiscardMessage *vastMsg)
{
    // discard outdated entry
    neighborsList->removeNode(vastMsg->getDiscardNode());
    // update voronoi
    neighborsList->buildVoronoi();
    synchronizeApp();
    neighborsList->removeNeighbors();
}

void Vast::sendDiscardNode(VastMessage *vastMsg)
{
    NodeHandle discardNode;
    discardNode.ip = thisNode.ip;
    discardNode.key = vastMsg->getDestKey();
    // send message
    VastDiscardMessage *vastDiscardMsg = new VastDiscardMessage("DISCARD_NODE");
    vastDiscardMsg->setCommand(DISCARD_NODE);
    vastDiscardMsg->setSourceNode(thisNode);
    vastDiscardMsg->setPos(position);
    vastDiscardMsg->setDiscardNode(discardNode);
    // debug output
    if(debugOutput) ev << "VAST: Node " << thisNode.ip << " received message for an outdated node from " << vastMsg->getSourceNode().ip << "." << endl;
    vastDiscardMsg->setLength(VASTDISCARD_L(vastDiscardMsg));
    sendMessage(vastDiscardMsg, vastMsg->getSourceNode());
}

void Vast::synchronizeApp(VastMoveMessage *vastMoveMsg)
{
    GameAPIListMessage *sgcMsg = new GameAPIListMessage("NEIGHBOR_UPDATE");
    sgcMsg->setCommand(NEIGHBOR_UPDATE);

    sgcMsg->setRemoveNeighborArraySize(neighborsList->getSize());
    sgcMsg->setAddNeighborArraySize(neighborsList->getSize() + 1);
    sgcMsg->setNeighborPositionArraySize(neighborsList->getSize() + 1);

    int remSize, addSize;
    remSize = addSize = 0;
    SiteMap::iterator itSites = neighborsList->Sites.begin();
    while(itSites != neighborsList->Sites.end()) {
        if(itSites->second.type == UNDEF) {
            sgcMsg->setRemoveNeighbor(remSize, itSites->second.addr);
            ++remSize;
        }
        else if(!itSites->second.isAdded) {
            sgcMsg->setAddNeighbor(addSize, itSites->second.addr);
            sgcMsg->setNeighborPosition(addSize, itSites->second.coord);
            itSites->second.isAdded = true;
            ++addSize;
        }
        ++itSites;
    }

    if(vastMoveMsg &&
       neighborsList->Sites.find(vastMoveMsg->getSourceNode()) != neighborsList->Sites.end() &&
       neighborsList->Sites.find(vastMoveMsg->getSourceNode())->second.isAdded) {
        sgcMsg->setAddNeighbor(addSize, vastMoveMsg->getSourceNode());
        sgcMsg->setNeighborPosition(addSize, vastMoveMsg->getNewPos());
        ++addSize;
    }

    sgcMsg->setRemoveNeighborArraySize(remSize);
    sgcMsg->setAddNeighborArraySize(addSize);
    sgcMsg->setNeighborPositionArraySize(addSize);

    if(sgcMsg->getAddNeighborArraySize() || sgcMsg->getRemoveNeighborArraySize()) {
        sendToApp(sgcMsg);
    }
    else {
        delete sgcMsg;
    }
}

void Vast::sendToApp(cMessage *msg)
{
    // debug output
    if(debugOutput) ev << "VAST: Node " << thisNode.ip << " sending " << msg->name() << " to application." << endl;
    send(msg, "to_app");
}

void Vast::sendMessage(VastMessage *vastMsg, NodeHandle destAddr)
{
    // collect statistics
    switch(vastMsg->getCommand()) {
        case JOIN_REQUEST: {
            joinRequestBytesSent += vastMsg->byteLength();
        } break;
        case JOIN_ACKNOWLEDGE: {
            joinAcknowledgeBytesSent += vastMsg->byteLength();
        } break;
        case NODE_MOVE: {
            nodeMoveBytesSent += vastMsg->byteLength();
        } break;
        case NEW_NEIGHBORS: {
            newNeighborsBytesSent += vastMsg->byteLength();
        } break;
        case NODE_LEAVE: {
            nodeLeaveBytesSent += vastMsg->byteLength();
        } break;
        case ENCLOSING_NEIGHBORS_REQUEST: {
            enclosingNeighborsRequestBytesSent += vastMsg->byteLength();
        } break;
        case PING: {
            pingBytesSent += vastMsg->byteLength();
        } break;
        case PONG: {
            pongBytesSent += vastMsg->byteLength();
        } break;
        case DISCARD_NODE: {
            discardNodeBytesSent += vastMsg->byteLength();
        } break;
    }
    bytesPerSecond += vastMsg->byteLength();

    // debug output
    if(debugOutput) ev << "VAST: Node " << thisNode.ip << " sending " << vastMsg->name() << " to " << destAddr.ip << "." << endl;
    vastMsg->setDestKey(destAddr.key);
    sendMessageToUDP(destAddr, vastMsg);
}

double Vast::distance_square(Site s, VastMessage *msg)
{
    return s.coord.distanceSqr(msg->getPos());
}

void Vast::setBootstrapedIcon()
{
    if(ev.isGUI()) {
        if(state == READY) {
            parentModule()->parentModule()->displayString().setTagArg("i2", 1, "green");
            displayString().setTagArg("i", 1, "green");
        }
        else if(state == JOINING) {
            parentModule()->parentModule()->displayString().setTagArg("i2", 1, "yellow");
            displayString().setTagArg("i", 1, "yellow");
        }
        else {
            parentModule()->parentModule()->displayString().setTagArg("i2", 1, "red");
            displayString().setTagArg("i", 1, "red");
        }
    }
}

void Vast::finishOverlay()
{
    bootstrapOracle->removePeer(thisNode);

    simtime_t time = globalStatistics->calcMeasuredLifetime(creationTime);
    if(time == 0) return;

    // collect statistics
    globalStatistics->addStdDev("Vast: JOIN_REQUEST bytes sent/s", joinRequestBytesSent/time);
    globalStatistics->addStdDev("Vast: JOIN_ACKNOWLEDGE bytes sent/s", joinAcknowledgeBytesSent/time);
    globalStatistics->addStdDev("Vast: NODE_MOVE bytes sent/s", nodeMoveBytesSent/time);
    globalStatistics->addStdDev("Vast: NEW_NEIGHBORS bytes sent/s", newNeighborsBytesSent/time);
    globalStatistics->addStdDev("Vast: NODE_LEAVE bytes sent/s", nodeLeaveBytesSent/time);
    globalStatistics->addStdDev("Vast: ENCLOSING_NEIGHBORS_REQUEST bytes sent/s", enclosingNeighborsRequestBytesSent/time);
    globalStatistics->addStdDev("Vast: PING bytes sent/s", pingBytesSent/time);
    globalStatistics->addStdDev("Vast: PONG bytes sent/s", pongBytesSent/time);
    globalStatistics->addStdDev("Vast: DISCARD_NODE bytes sent/s", discardNodeBytesSent/time);

    globalStatistics->addStdDev("max bytes/second sent", maxBytesPerSecondSent);
    globalStatistics->addStdDev("average bytes/second sent", averageBytesPerSecondSent / time);
}

double Vast::getAOI()
{
    Enter_Method_Silent();
    return AOI_size;
}

Vector2D Vast::getPosition()
{
    Enter_Method_Silent();
    return position;
}

NeighborsList* Vast::getList()
{
    Enter_Method_Silent();
    return neighborsList;
}

Vast::~Vast()
{
    // destroy self timer messages
    cancelAndDelete(join_timer);
    cancelAndDelete(ping_timer);
    cancelAndDelete(sec_timer);
}
