#!/usr/bin/python

# Global Imports
import re, os, math, logging, optparse

# Class Definitions
class Experiment: # stores data for each experiment
    """ class to store data for each experiment """
    
    # prepare regular expressions
    reSemi = re.compile(r";") # semicolon
    reLine = re.compile(r"(\S*)\s*=\s*(.*?)[\r\n]") # split between =
    reName = re.compile(r"^\s*\**\.*([\w\.]*)") # prettyprint name
    reValue = re.compile(r"^\s*\"?([\w\.]*)\"?") # prettyprint value
    reComb = re.compile(r"\$\$") # combined parameter for two lines
        
    reRuns = re.compile(r"^\s*runs\s*=\s*(\d+)") # number of runs
    
    def __init__(self, expLines, expName):
        # initialization
        self.name = expName # name of the experiment
        self.fileNames = [] # list of file names
        self.lines = [] # list of lists of lines for ini files
        self.parameterNames = [] # list of parameter names
        self.parameterValues =  [] # list of lists of parameter values
        self.runs = 0 # number of runs for the experiment
        
        self.currentLines = expLines[:]
        self.allLines = expLines
        
        # parse the lines
        self.generate (0, expName)
        
        if self.runs == 0:
            self.runs = self.defaultRuns
        
        # cleanup
        del self.currentLines
        del self.allLines        
        
    def generate(self, index, fileName):
        """ parse the lines and generate parameter combinations """
        
        if index >= len(self.allLines): # all lines processed
            self.fileNames.append(fileName)
            self.lines.append(self.currentLines[:])
            return
        
        line = self.allLines[index]
        
        if self.reSemi.search(line): # semicolon found, parse parameters
            matches = self.reLine.search(line)
            name = matches.group(1)
            pName = self.reName.search(name).group(1) # pretty name for output
            values = self.reSemi.split(matches.group(2))
            
            if not pName in self.parameterNames: # add parameter to global parameter list
                self.parameterNames.append(pName)
                pValues = []
                for value in values:
                    pValues.append(self.reValue.search(value).group(1))
                self.parameterValues.append(pValues)
                
            for value in values:
                self.currentLines[index] = name + "=" + value + "\n"
                pValue = self.reValue.search(value).group(1) # pretty value for output
                self.generate (index + 1, fileName + "-" + pName + "=" + pValue)
                
        elif self.reRuns.search(line): # line contains number of runs
            if not self.runs: # check if number of runs already set
                self.runs = int(self.reRuns.search(line).group(1))
            self.generate(index + 1, fileName)                
            
        elif self.reComb.search(line):
            lastValue = self.reLine.search(self.currentLines[index-1]).group(2)
            thisName = self.reLine.search(line).group(1)
            self.currentLines[index] = thisName + "=" + lastValue + "\n"
            self.generate(index + 1, fileName)
            
        else:
            # nothing to do, go into next recursion step
            self.generate(index + 1, fileName)
    
class IniParser: # parses the ini file and creates experiments
    """parses the ini file and generates experiments"""
    
    def __init__(self, iniFile):
        self.iniFile = iniFile
        
    def parse(self):
        """parse ini file"""
        
        logging.info("parsing ini file")
        
        # compile regular expressions
        empty = re.compile(r"^\s*$|^\#") # empty or comment line
        sect = re.compile(r"^\s*\[(\w*)\]") # new section
        inval = re.compile(r"^\s*\[") # invalid section
        
        expLines = {}
        
        # parse file
        file = open (iniFile, "r")
        
        for line in file:
            if empty.search(line): # skip empty lines
                continue 
            
            if sect.search(line): # new section starting
                currentSect = sect.search(line).group(1)
                logging.info("-- parsing section [" + currentSect + "]")
                if not currentSect in expLines:
                    expLines[currentSect] = []
                continue
                
            elif inval.search(line): # invalid section
                logging.error("invalid characters found in section " + line)
                quit()
            
            if currentSect == "": # invalid file format
                logging.error("lines found before a section started in ini file")
                quit()
            
            expLines[currentSect].append(line)
            
        file.close()
        
        logging.info("-- generating experiments")
        
        experiments = []
        for name in expLines.keys():
            if name == "Global":
                continue
            lines = expLines[name][:]
            if "Global" in expLines: # add Global lines to each experiment
                lines.extend(expLines["Global"])
                
            experiments.append(Experiment(lines, name))
        
        logging.info("Done")
        return experiments
        
        
class BaseWorker: # parent class for "worker" classes (for common methods)
    def remove(self, file): # delete file if it exists
        if os.access(file, os.F_OK):
            os.remove(file)

class RemoveOversimFiles(BaseWorker): # delete files created by Oversim
    
    def delete(self, experiments):
        logging.info("Deleting original sca and vec files")
        
        for exp in experiments: # loop through experiments
            logging.info("-- " + exp.name)
            
            for index, fileName in enumerate(exp.fileNames): # loop through parameter combinations
                
                for run in range(1, exp.runs+1): # loop through number of runs
                    baseFileName = fileName + "-run=" + str(run)
                    self.remove(baseFileName + ".sca")
                    self.remove(baseFileName + ".vec")
                    self.remove(baseFileName + ".debug")
                    
        logging.info("Done")
        
        
class GenerateIniFiles(BaseWorker): # generates ini files
    
    def generate(self, experiments):
        """ generate ini files """
        logging.info("Generate ini files")
        
        for exp in experiments: # loop through experiments
            logging.info("-- " + exp.name)
            
            for index, fileName in enumerate(exp.fileNames): # loop through parameter combinations

                for run in range(1, exp.runs+1): # loop through number of runs
                    baseFileName = fileName + "-run=" + str(run)
                    logging.info("---- " + baseFileName)
                    
                    # write ini file
                    file = open(baseFileName + ".ini", "w")
                    
                    file.write("[Run " + str(run) + "]\n")
                    file.write("description = " + baseFileName + "\n")
                    file.write("output-scalar-file = " + baseFileName + ".sca\n")
                    file.write("output-vector-file = " + baseFileName + ".vec\n")
                    for line in exp.lines[index]:
                        file.write(line)
                        
                    file.close()
                    
        logging.info("Done")
        
    def delete(self, experiments):
        """ delete ini files """
        logging.info("Deleting ini files")
        
        for exp in experiments: # loop through experiments
            logging.info("-- " + exp.name)
            
            for index, fileName in enumerate(exp.fileNames): # loop through parameter combinations
                
                for run in range(1, exp.runs+1): # loop through number of runs
                    baseFileName = fileName + "-run=" + str(run)
                    self.remove(baseFileName + ".ini")
                    
        logging.info("Done")

    
class CollectStatistics(BaseWorker): # collects statistic data
    
    # compile regular expressions
    reVec = re.compile(r"^vector") # new vector starting
    reVecName = re.compile(r'(\d*)  "(.*)"  (".*")  (\d*)\n$') # parse name if vector
    reVecLine = re.compile(r"^([0-9]*)\t([0-9.]*)\t([0-9.]*)\n$") # parse data line in vec file
    
    reRemDigits = re.compile(r"\[[0-9]*\]") # find digits in square brackets
    
    reScaRun = re.compile(r"^run|^\s*$") # new run starting or empty line
    reScaLine = re.compile(r'scalar\s*("[^"]*")\s*("[^"]*")\s*(-?[\d.]*e?-?\d*)') # parse data line in sca file
    
    def generate(self, experiments):
        """ collect statistics from sca and vec files """
        logging.info("collecting statistics")
        
        for exp in experiments: # loop through experiments
            logging.info("-- " + exp.name)
            self.globalVecBaskets = {}
            self.globalScaStat = {}
            for index, fileName in enumerate(exp.fileNames): # loop through parameter combinations
                logging.info("---- " + fileName)
                self.currVecBaskets = {}
                self.currScaStat = {}
                for run in range(1, exp.runs+1): # loop through number of runs
                    logging.info("------ Run " + str(run) + "/" + str(exp.runs))
                    baseFileName = fileName + "-run=" + str(run)
                    self.processSca(baseFileName)
                    self.processVec(baseFileName)
                    
                self.printScaStats(fileName, self.currScaStat)
                self.printVecStats(fileName, self.currVecBaskets)
                
            logging.info("---- Global statistics")
            self.printScaStats(exp.name, self.globalScaStat)
            self.printVecStats(exp.name, self.globalVecBaskets)
            
            self.makeScaDat(exp)
            
        logging.info("done")
    
    def delete(self, experiments):
        """ delete generated files """
        logging.info("deleting dat files")
        
        for exp in experiments:
            logging.info("-- " + exp.name)
            self.remove(exp.name + ".sca")
            self.remove(exp.name + ".vec.dat")
            for fileName in exp.fileNames:
                logging.info("---- " + fileName)
                self.remove(fileName + ".sca")
                self.remove(fileName + ".vec.dat")
                
            for index, pName in enumerate(exp.parameterNames):
                values = exp.parameterValues[index]
                for value in values: # loop through values
                    param = pName + "=" + value # parameter-value string in fileName
                    for fileName in exp.fileNames: # search files that contain param
                        if not fileName.find(param) < 0:
                            stripped = fileName.replace(param, "")
                            logging.info("---- " + stripped)
                            self.remove(stripped + ".sca.dat")
                            
        logging.info("done")
        
        
    def processSca(self, fileName):
        """ parse sca file """
        import scipy, scipy.stats # import scipy packages
        
        fileName = fileName + ".sca"
        
        if not os.access(fileName, os.F_OK):
            logging.warn("Unable to parse file " + fileName + ": file not found")
            return
            
        file = open(fileName, "r")
        
        for line in file:
            if self.reScaRun.search(line):
                continue
                
            line = self.reRemDigits.sub("[]", line)
            matches = self.reScaLine.search(line)
            name = matches.group(1) + "\t" + matches.group(2)
            value = float(matches.group(3))
            
            if not name in self.currScaStat:
                self.currScaStat[name] = []
            if not name in self.globalScaStat:
                self.globalScaStat[name] = []
                
            self.currScaStat[name].append(value)
            self.globalScaStat[name].append(value)
            
        file.close()
        
    def processVec(self, fileName):
        """ parse vec file """
        import scipy, scipy.stats # import scipy packages
        
        fileName = fileName + ".vec"
        
        vectorNames = {}
        baskets = {}
        
        if not os.access(fileName, os.F_OK):
            logging.warn("Unable to parse file " + fileName + ": file not found")
            return
        
        # parse file
        file = open(fileName, "r")
        
        for line in file:
            if self.reVec.search(line): # new vector starting
                line = self.reRemDigits.sub("[]", line)
                matches = self.reVecName.search(line)
                vectorNames[matches.group(1)] = matches.group(2) + " " + matches.group(3)
                
            elif self.reVecLine.search(line): # data line
                matches = self.reVecLine.search(line)
                index = math.floor(float(matches.group(2)) / self.vecInterval) * self.vecInterval
                
                # initialize lists if needed
                if not vectorNames[matches.group(1)] in baskets:
                    baskets[vectorNames[matches.group(1)]] = {}
                if not index in baskets[vectorNames[matches.group(1)]]:
                    baskets[vectorNames[matches.group(1)]][index] = []
                    
                # add value to basket
                baskets[vectorNames[matches.group(1)]][index].append(float(matches.group(3)))
            
        file.close()
        
        # summarize data
        for vector in baskets:
            # initialize global baskets if needed
            if not vector in self.globalVecBaskets:
                self.globalVecBaskets[vector] = {}
            if not vector in self.currVecBaskets:
                self.currVecBaskets[vector] = {}
                
            for index in baskets[vector]:
                # initialize global baskets if needed
                if not index in self.globalVecBaskets[vector]:
                    self.globalVecBaskets[vector][index] = []
                if not index in self.currVecBaskets[vector]:
                    self.currVecBaskets[vector][index] = []
                    
                arr = scipy.array(baskets[vector][index])
                mean = arr.mean()
                self.globalVecBaskets[vector][index].append(mean)
                self.currVecBaskets[vector][index].append(mean)
                
        
    def printScaStats(self, fileName, stats):
        """ print scalar data """
        import scipy, scipy.stats # import scipy packages
        
        file = open (fileName + ".sca", "w")
        
        for name in sorted(stats.keys()):
            arr = scipy.array(stats[name])
            value = str(self.non_exp_repr(arr.mean()))
            file.write("scalar " + name + "\t" + value + "\n")
            
        file.close()
        
    def printVecStats(self, fileName, baskets):
        """ print vector data """
        import scipy, scipy.stats # import scipy packages
        
        vecCounter = 1
        maxIndex = 0
        
        file = open (fileName + ".vec.dat", "w")
        
        # header
        line = "#\tColumn\tVector\n"
        line += "#\t1\tsimulation time\n"
        for vector in sorted(baskets.keys()):
            vecCounter += 1
            line += "#\t" + str(vecCounter) + "\t" + vector + "(mean)\n"
            vecCounter += 1
            line += "#\t" + str(vecCounter) + "\t" + vector + "(10-quantile)\n"
            vecCounter += 1
            line += "#\t" + str(vecCounter) + "\t" + vector + "(90-quantile)\n"
            if maxIndex < max(baskets[vector].keys()):
                maxIndex = int(max(baskets[vector].keys()))
                
        line += "#" * 70 + "\n"
        file.write(line)
        
        # data
        for index in range(0, maxIndex+self.vecInterval, self.vecInterval): # loop through buckets
            line = str(index)
            for vector in sorted(baskets.keys()): # loop through vectors
                if index in baskets[vector]:
                    list = baskets[vector][index]
                    arr = scipy.array(list)
                    line += "\t" + str(self.non_exp_repr(arr.mean()))
                    line += "\t" + str(self.non_exp_repr(self.percentile(sorted(list), 0.1)))
                    line += "\t" + str(self.non_exp_repr(self.percentile(sorted(list), 0.9)))
                else:
                    line += "\t0\t0\t0"
                    
            file.write(line + "\n")
            
        file.close()
        
    def percentile(self, N, percent, key=lambda x:x):
        """
        Find the percentile of a list of values.
        Source: http://aspn.activestate.com/ASPN/Cookbook/Python/Recipe/511478

        @parameter N - is a list of values. Note N MUST BE already sorted.
        @parameter percent - a float value from 0.0 to 1.0.
        @parameter key - optional key function to compute value from each element of N.

        @return - the percentile of the values
        """
        if not N:
            return None
        k = (len(N)-1) * percent
        f = math.floor(k)
        c = math.ceil(k)
        if f == c:
            return key(N[int(k)])
        d0 = key(N[int(f)]) * (k-f)
        d1 = key(N[int(c)]) * (c-k)
        return d0+d1
        
    def numeric_compare(self, x, y):
        # from http://wiki.python.org/moin/HowTo/Sorting
        if not self.IsInt(x) or not self.IsInt(y):
          return cmp(x, y) #use default compare function
          
        x = int(x)
        y = int(y)
        
        if x > y:
            return 1
        elif x == y:
            return 0
        else: # x < y
            return -1
    
    def IsInt(self, str):
        # from http://aspn.activestate.com/ASPN/Cookbook/Python/Recipe/52660
        """ Is the given string an integer?	"""
        ok = 1
        try:
            num = int(str)
        except ValueError:
            ok = 0
        return ok
        
    def non_exp_repr(self, x):
        # from http://aspn.activestate.com/ASPN/Cookbook/Python/Recipe/358361
        """Return a floating point representation without exponential notation.

        Result is a string that satisfies:
            float(result)==float(x) and 'e' not in result.
        
        >>> non_exp_repr(1.234e-025)
        '0.00000000000000000000000012339999999999999'
        >>> non_exp_repr(-1.234e+018)
        '-1234000000000000000.0'
        
        >>> for e in xrange(-50,51):
        ...     for m in (1.234, 0.018, -0.89, -75.59, 100/7.0, -909):
        ...         x = m * 10 ** e
        ...         s = non_exp_repr(x)
        ...         assert 'e' not in s
        ...         assert float(x) == float(s)

        """
        s = repr(float(x))
        e_loc = s.lower().find('e')
        if e_loc == -1:
            return s

        mantissa = s[:e_loc].replace('.', '')
        exp = int(s[e_loc+1:])

        assert s[1] == '.' or s[0] == '-' and s[2] == '.', "Unsupported format"     
        sign = ''
        if mantissa[0] == '-':
            sign = '-'
            mantissa = mantissa[1:]

        digitsafter = len(mantissa) - 1     # num digits after the decimal point
        if exp >= digitsafter:
            return sign + mantissa + '0' * (exp - digitsafter) + '.0'
        elif exp <= -1:
            return sign + '0.' + '0' * (-exp - 1) + mantissa
        ip = exp + 1                        # insertion point
        return sign + mantissa[:ip] + '.' + mantissa[ip:]
        
    def makeScaDat(self, exp):
        """ create sca dat files """
        import scipy, scipy.stats # import scipy packages
        
        for index, pName in enumerate(exp.parameterNames): # loop through parameters
            values = exp.parameterValues[index]
            
            dict = {}
            entryCount = 0
            
            for value in values: # loop through values
                param = pName + "=" + value # parameter-value string in fileName
                
                for fileName in exp.fileNames: # search files that contain param
                    if not fileName.find(param) < 0:
                        
                        # process sca files
                        self.currScaStat = {}
                        self.globalScaStat = {}
                        for run in range(1, exp.runs+1): # loop through number of runs
                            baseFileName = fileName + "-run=" + str(run)
                            self.processSca(baseFileName)
                        
                        # check if sca files have different number of scalars (should not happen!)
                        if entryCount == 0:
                            entryCount = len(self.currScaStat)
                        elif not entryCount == len(self.currScaStat) and entryCount >= 0:
                            logging.error("Your sca files don't contain the same number of entries. Maybe you are comparing different Protocols oder using different applications?")
                            entryCount = -1;
                        
                        # save collected data
                        stripped = fileName.replace(param, "")
                        if not stripped in dict:
                            dict[stripped] = {}
                            
                        dict[stripped][value] = self.currScaStat.copy()
                        
            # write sca.dat files
            for stripped in dict:
                file = open (stripped + ".sca.dat", "w")
                first = 1
                
                for value in sorted(dict[stripped].keys(), self.numeric_compare):
                    stats = dict[stripped][value]
                    line = ""
                    
                    if first: # write header
                        first = 0
                        line += "#\tcolumn\tscalar name\n"
                        line += "#\t1\t" + pName + "\n"
                        count = 1
                        for name in sorted(stats.keys()):
                            count += 1
                            line += "#\t" + str(count) + "\t" + name + "(mean)\n"
                            count += 1
                            line += "#\t" + str(count) + "\t" + name + "(10-quantile)\n"
                            count += 1
                            line += "#\t" + str(count) + "\t" + name + "(90-quantile)\n"
                            
                        line += "#" * 70 + "\n"
                        file.write(line)
                        line = ""
                        
                    line += value
                    
                    for name in sorted(stats.keys()):
                        list = stats[name]
                        arr = scipy.array(list)
                        line += "\t" + str(self.non_exp_repr(arr.mean()))
                        line += "\t" + str(self.non_exp_repr(self.percentile(sorted(list), 0.1)))
                        line += "\t" + str(self.non_exp_repr(self.percentile(sorted(list), 0.9)))
                        
                    file.write(line + "\n")
                    
                file.close()
                




################################################################
## Start of program
################################################################
# Parse Command Line
parser = optparse.OptionParser()

parser.add_option("-i", "--ini", dest="iniFile", help="ini file [default: 'generate.ini']", metavar="FILE", default="generate.ini")
parser.add_option("-r", "--runs", dest="runs", help="number of runs if not specified in ini file [default: 3]", metavar="RUNS", type="int", default=3)
parser.add_option("-b", "--bucket", dest="vecInterval", help="default bucketsize for parsing vec files [default: 100]", metavar="VAL", type="int", default=100)
parser.add_option("-q", "--quiet", action="store_false", dest="verbose", default=True, help="don't print status messages")
parser.add_option("-g", "--generate", dest="generate", help="generate ini files (default action if none of the following parameters is specified)", default=True, action="store_true")
parser.add_option("-s", "--stats", dest="stats", help="generate statistics after Oversim run", default=False, action="store_true")
parser.add_option("--cleanup", dest="cleanup", help="delete all generated files (combines --cleanup-ini and cleanup-stats)", default=False, action="store_true")
parser.add_option("--cleanup-ini", dest="cleanup_ini", help="delete all generated ini files", default=False, action="store_true")
parser.add_option("--cleanup-stats", dest="cleanup_stats", help="delete all generated statistic files", default=False, action="store_true")
parser.add_option("--cleanup-oversim", dest="cleanup_oversim", help="delete all files generated by Oversim (.debug, .sca, .vec)", default=False, action="store_true")

(options, args) = parser.parse_args()

# act on command line parameters
iniFile = options.iniFile # check ini file
if not os.access(iniFile, os.R_OK):
    parser.error("Unable to read ini file '" + iniFile + "'")

Experiment.defaultRuns = options.runs # set default runs
CollectStatistics.vecInterval = options.vecInterval # set vec interval

if options.verbose: # set verbose
    logging.basicConfig(level=logging.INFO)
    
# parse ini file
experiments = IniParser(iniFile).parse()

# do something
if options.cleanup: # delete ini and dat files
    options.cleanup_ini = 1
    options.cleanup_stats = 1
    
if options.cleanup_ini: # delete ini files
    GenerateIniFiles().delete(experiments)
    options.generate = 0

if options.cleanup_stats: # delete data files
    CollectStatistics().delete(experiments)
    options.generate = 0
    
if options.cleanup_oversim: # delete files generated by Oversim
    RemoveOversimFiles().delete(experiments)
    options.generate = 0
    
if options.stats: # collect statistics
    CollectStatistics().generate(experiments)
    options.generate = 0
    
if options.generate: # generate ini files
    GenerateIniFiles().generate(experiments)
