//
// Copyright (C) 2007 Institut fuer Telematik, Universitaet Karlsruhe (TH)
//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public License
// as published by the Free Software Foundation; either version 2
// of the License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
//

/**
 * @file P2pns.cc
 * @author Ingmar Baumgart
 */

#include <IPAddressResolver.h>

#include "P2pns.h"

Define_Module(P2pns);

using namespace std;

P2pns::P2pns()
{
    p2pnsCache = NULL;
}

P2pns::~P2pns()
{
}

void P2pns::initializeApp(int stage)
{
    if (stage != MIN_STAGE_APP)
        return;

    twoStageResolution = par("twoStageResolution");

    p2pnsCache = check_and_cast<P2pnsCache*> (parentModule()->
            submodule("p2pnsCache"));
}

void P2pns::handleTimerEvent(cMessage* msg)
{
}

bool P2pns::handleRpc(BaseCallMessage* msg)
{
    // delegate messages
    RPC_SWITCH_START(msg)
    // RPC_DELEGATE( <messageName>[Call|Response], <methodToCall> )
    RPC_DELEGATE( P2pnsRegister, p2pnsRegisterRpc );
    RPC_DELEGATE( P2pnsResolve, p2pnsResolveRpc );
    RPC_SWITCH_END()

    return RPC_HANDLED;
}


void P2pns::handleRpcResponse(BaseResponseMessage* msg,
                              cPolymorphic* context, int rpcId, simtime_t rtt)
{
    RPC_SWITCH_START(msg)
    RPC_ON_RESPONSE( DHTputCAPI ) {
        EV << "DHTputCAPI RPC Response received: id=" << rpcId
           << " msg=" << *_DHTputCAPIResponse << " rtt=" << rtt << endl;
        handleDHTputCAPIResponse(_DHTputCAPIResponse,
                                 check_and_cast<P2pnsRegisterCall*>(context));
        break;
    }
    RPC_ON_RESPONSE( DHTgetCAPI ) {
        EV << "DHTgetCAPI RPC Response received: id=" << rpcId
           << " msg=" << *_DHTgetCAPIResponse << " rtt=" << rtt << endl;
        handleDHTgetCAPIResponse(_DHTgetCAPIResponse,
                                 check_and_cast<P2pnsResolveCall*>(context));
        break;
    }
    RPC_ON_RESPONSE( Lookup ) {
        EV << "Lookup RPC Response received: id=" << rpcId
           << " msg=" << *_LookupResponse << " rtt=" << rtt << endl;
        handleLookupResponse(_LookupResponse,
                             check_and_cast<P2pnsResolveCall*>(context));
        break;
    }
    RPC_SWITCH_END()
}

void P2pns::handleRpcTimeout(BaseCallMessage* msg,
                             const TransportAddress& dest,
                             cPolymorphic* context, int rpcId,
                             const OverlayKey& destKey)
{
    throw new cRuntimeError("P2PNS::handleRpcTimeout(): Not implemented!");
}

void P2pns::p2pnsRegisterRpc(P2pnsRegisterCall* registerCall)
{
    p2pnsCache->addData(registerCall->getP2pName(),
                        registerCall->getAddress());

    DHTputCAPICall* dhtPutMsg = new DHTputCAPICall();

    EV << "RegisterRpc: name: " << registerCall->getP2pName()
       << " addr: " << registerCall->getAddress() << endl;

    dhtPutMsg->setKey(OverlayKey::sha1(registerCall->getP2pName()));
    if (twoStageResolution) {
        dhtPutMsg->setValue(overlay->getThisNode().key.toString());
    } else {
        dhtPutMsg->setValue(registerCall->getAddress());
    }

    dhtPutMsg->setTtl(registerCall->getTtl());
    dhtPutMsg->setIsModifiable(true);

    sendInternalRpcCall(TIER1_COMP, dhtPutMsg, registerCall);
}

void P2pns::p2pnsResolveRpc(P2pnsResolveCall* resolveCall)
{
    DHTgetCAPICall* dhtGetMsg = new DHTgetCAPICall();

    EV << "ResolveRpc: name: " << resolveCall->getP2pName()
           << endl;

    dhtGetMsg->setKey(OverlayKey::sha1(resolveCall->getP2pName()));

    sendInternalRpcCall(TIER1_COMP, dhtGetMsg, resolveCall);
}

void P2pns::handleDHTputCAPIResponse(DHTputCAPIResponse* putResponse,
                                     P2pnsRegisterCall* registerCall)
{
    P2pnsRegisterResponse* registerResponse = new P2pnsRegisterResponse();
    registerResponse->setP2pName(registerCall->getP2pName());
    registerResponse->setAddress(registerCall->getAddress());
    registerResponse->setIsSuccess(putResponse->getIsSuccess());
    sendRpcResponse(registerCall, registerResponse);
}

void P2pns::handleDHTgetCAPIResponse(DHTgetCAPIResponse* getResponse,
                                     P2pnsResolveCall* resolveCall)
{
    std::stringstream valueStream;
    valueStream << getResponse->getValue();

    if ((!getResponse->getIsSuccess())
            || (valueStream.str().size() == 0)) {
        P2pnsResolveResponse* resolveResponse = new P2pnsResolveResponse();
        resolveResponse->setP2pName(resolveCall->getP2pName());
        resolveResponse->setAddress(BinaryValue(""));
        resolveResponse->setIsSuccess(false);
        sendRpcResponse(resolveCall, resolveResponse);
        return;
    }

    p2pnsCache->addData(resolveCall->getP2pName(),
                        getResponse->getValue());

    if (twoStageResolution) {
        OverlayKey key(valueStream.str(), 16);

        LookupCall* lookupCall = new LookupCall();

        lookupCall->setKey(key);
        lookupCall->setNumSiblings(1);

        sendInternalRpcCall(OVERLAY_COMP, lookupCall, resolveCall);

        return;
    }

    EV << "ResolveRpcResponse: name: " << resolveCall->getP2pName()
       << " addr: " << getResponse->getValue() << endl;


    P2pnsResolveResponse* resolveResponse = new P2pnsResolveResponse();
    resolveResponse->setP2pName(resolveCall->getP2pName());
    resolveResponse->setAddress(getResponse->getValue());
    resolveResponse->setIsSuccess(getResponse->getIsSuccess());
    sendRpcResponse(resolveCall, resolveResponse);
}

void P2pns::handleLookupResponse(LookupResponse* lookupResponse,
                                 P2pnsResolveCall* resolveCall)
{
    stringstream sstream;
    sstream << lookupResponse->getSiblings(0);

    P2pnsResolveResponse* resolveResponse = new P2pnsResolveResponse();
    resolveResponse->setP2pName(resolveCall->getP2pName());
    resolveResponse->setAddress(sstream.str());
    resolveResponse->setIsSuccess(lookupResponse->getIsValid());
    sendRpcResponse(resolveCall, resolveResponse);
}


void P2pns::finishApp()
{
}

