//
// Copyright (C) 2006 Institut fuer Telematik, Universitaet Karlsruhe (TH)
//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public License
// as published by the Free Software Foundation; either version 2
// of the License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
//

/**
 * @file SimpleNodeEntry.cc
 * @author Bernhard Heep
 */

#include <sstream>

#include "SimpleNodeEntry.h"
#include "IPAddressResolver.h"


uint NodeRecord::dim;

NodeRecord::NodeRecord(uint dim)
{
    this->dim = dim;
    coords = new double[dim];
}

NodeRecord::~NodeRecord()
{
    delete[] coords;
    coords = NULL;
}

NodeRecord::NodeRecord(const NodeRecord& nodeRecord)
{
    std::cout << "copy ";
    
    coords = new double[dim];
    for (uint i = 0; i < dim; ++i)
        coords[i] = nodeRecord.coords[i]; 
}

NodeRecord& NodeRecord::operator=(const NodeRecord& nodeRecord)
{
    std::cout << "= ";
    
    delete[] coords;
    coords = new double[dim];
    for (uint i = 0; i < dim; ++i)
        coords[i] = nodeRecord.coords[i];
    
    return *this;
}


SimpleNodeEntry::SimpleNodeEntry(cModule* node,
                                 cChannelType* type,
                                 uint fieldSize,
                                 uint sendQueueLength)
{
    ingate = node->submodule("udp")->gate("network_in");
	
    nodeRecord = new NodeRecord(2);
    index = -1;

    //use random values as coordinates
    nodeRecord->coords[0] = uniform(0, fieldSize) - fieldSize / 2;
    nodeRecord->coords[1] = uniform(0, fieldSize) - fieldSize / 2;

    cBasicChannel* temp = dynamic_cast<cBasicChannel*>(type->create("temp"));

    bandwidth = temp->datarate();
    errorRate = temp->error();
    accessDelay = temp->delay();

    txMaxQueueTime = ((float)sendQueueLength * 8) / bandwidth;
    txFinished = simulation.simTime();

    delete temp;
}

SimpleNodeEntry::SimpleNodeEntry(cModule* node,
                                 cChannelType* type,
                                 uint sendQueueLength,
                                 NodeRecord* nodeRecord, int index)
{
    ingate = node->submodule("udp")->gate("network_in");


    //this->nodeRecord = new NodeRecord(nodeRecord);
    this->nodeRecord = nodeRecord;
    this->index = index;

    cBasicChannel* temp = dynamic_cast<cBasicChannel*>(type->create("temp"));

    bandwidth = temp->datarate();
    errorRate = temp->error();
    accessDelay = temp->delay();

    txMaxQueueTime = ((float)sendQueueLength * 8) / bandwidth;
    txFinished = simulation.simTime();

    delete temp;
}

float SimpleNodeEntry::operator-(const SimpleNodeEntry& entry) const
{
    double sum_of_squares = 0;
    for (uint i = 0; i < nodeRecord->dim; i++) {
        sum_of_squares += pow(nodeRecord->coords[i] -
                              entry.nodeRecord->coords[i], 2);
    }
    return sqrt(sum_of_squares);
}

SimpleNodeEntry::SimpleDelay SimpleNodeEntry::calcDelay(const SimpleUDPPacket& msg, const SimpleNodeEntry& dest)
{
    if ((pow(1 - errorRate, msg.byteLength() * 8) <= uniform(0, 1))
        || (pow(1 - dest.errorRate, msg.byteLength() * 8) <= uniform(0, 1))) {
     
        return SimpleDelay(0, false);
    }

    simtime_t now = simulation.simTime();
    simtime_t bandwidthDelay= ((msg.byteLength() * 8) / bandwidth);
    simtime_t newTxFinished = fmax(txFinished, now) + bandwidthDelay;

    // send queue
    if ((newTxFinished > now + txMaxQueueTime) && (txMaxQueueTime != 0)) {
        EV << "[SimpleNodeEntry::calcDelay()]\n"
           << "    Send queue overrun"
           << "\n    newTxFinished = fmax(txFinished, now) + bandwidthDelay"
           << "\n    newTxFinished = " << newTxFinished
           << "\n    txFinished = " << txFinished
           << "\n    now = " << now
           << "\n    bandwidthDelay = " << bandwidthDelay
           << "\n    (newTxFinished > now + txMaxQueueTime) == true"
           << "\n    txMaxQueueTime = " << txMaxQueueTime
           << endl;
        return SimpleDelay(0, false);
    }

    txFinished = newTxFinished;

    simtime_t destBandwidthDelay = (msg.byteLength() * 8) / dest.bandwidth;
    simtime_t coordDelay = 0.001 * (*this - dest);

    return SimpleDelay(txFinished - now
                       + accessDelay
                       + coordDelay
                       + destBandwidthDelay + dest.accessDelay, true);
}

std::string SimpleNodeEntry::info() const
{
    std::ostringstream str;
    str << *this;
    return str.str();
}

std::ostream& operator<<(std::ostream& out, const SimpleNodeEntry& entry)
{
    out << "(x:" << entry.nodeRecord->coords[0]
        << ", y:" << entry.nodeRecord->coords[1]
        << ")\nbandwidth = " << entry.bandwidth
        << ",\ndelay = " << entry.accessDelay
        << ",\nerrorRate = " << entry.errorRate
        << ",\ntxMaxQueueTime = " << entry.txMaxQueueTime
        << ",\ntxFinished = " << entry.txFinished;

    return out;
}
