//
// Copyright (C) 2000 Institut fuer Telematik, Universitaet Karlsruhe
// Copyright (C) 2004 Andras Varga
//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public License
// as published by the Free Software Foundation; either version 2
// of the License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
//

/**
 * @file SimpleNodeEntry.h
 * @author Bernhard Heep
 * @date 2006/11/30
 */

#ifndef __SIMPLENODEENTRY_H
#define __SIMPLENODEENTRY_H


#include <omnetpp.h>
#include <IPvXAddress.h>

#include "SimpleUDPPacket_m.h"


class NodeRecord
{
public:
    NodeRecord(uint dim);
    ~NodeRecord();
    NodeRecord(const NodeRecord& nodeRecord);
    NodeRecord& operator=(const NodeRecord& nodeRecord);
    void debugOutput(int dim);

    double* coords;
    IPvXAddress ip;
    static uint dim;
};

/**
 * representation of a single node in the BootstrapOracle
 *
 * @author Bernhard Heep
 */
class SimpleNodeEntry : public cPolymorphic
{
public:

    ~SimpleNodeEntry()
    {
        if (index == -1) delete nodeRecord;
    }

    /**
     * Constructor for use with 2D random coordinates
     *
     * @param node pointer to new terminal
     * @param type access channel of new terminal
     * @param fieldSize length of one side of the coordinate space
     * @param sendQueueLength initial send queue size
     */
    SimpleNodeEntry(cModule* node, cChannelType* type,
                    uint fieldSize, uint sendQueueLength);

    /**
     * Constructor for more than 2 dimensions
     *
     * @param node pointer to new terminal
     * @param type access channel of new terminal
     * @param sendQueueLength length of the send queue in bytes
     * @param nodeRecord the node's coordinates
     * @param index the position in unusedNodeRecords
     */
    SimpleNodeEntry(cModule* node, cChannelType* type, uint sendQueueLength,
                    NodeRecord* nodeRecord, int index);

    /**
     * Getter for SimpleUDP ingate
     *
     * @return the ingate
     */
    inline cGate* getGate() const
    {
        return ingate;
    };

    /**
     * Setter for maximum coordinate
     *
     * @param size maximum coordinate
     */
//    inline static void setFieldSize(uint size)
//    {
//        fieldSize = size;
//    };

    /**
     * Setter for send queue length
     *
     * @param length send queue length
     */
//    inline static void setSendQueueLength(uint length)
//    {
//        sendQueueLength = length;
//    };

    typedef std::pair<simtime_t, bool> SimpleDelay; //!< type for return value of calcDelay()

    /**
     * Calculates delay between two nodes
     *
     * @param msg reference to message to get its length for delay calculation,
     * @param dest destination terminal
     * @return delay in s and boolean value that is false if message should be deleted
     */
    SimpleDelay calcDelay(const SimpleUDPPacket& msg, const SimpleNodeEntry& dest);

    /**
     * OMNeT++ info method
     *
     * @return infostring
     */
    std::string info() const;

    /**
     * Stream output
     *
     * @param out output stream
     * @param entry the terminal
     * @return reference to stream out
     */
    friend std::ostream& operator<<(std::ostream& out, const SimpleNodeEntry& entry);

    simtime_t getAccessDelay() { return accessDelay; };

    // typo fixed, thanks to huebby
    float getBandwidth() { return bandwidth; };

    float getErrorRate() { return errorRate; };

    inline float getX() { return nodeRecord->coords[0]; };
    inline float getY() { return nodeRecord->coords[1]; };

    int getRecordIndex() { return index; };
    NodeRecord* getNodeRecord() { return nodeRecord; };

protected:

    /**
     * Calculates eulklidean distance between two terminals
     *
     * @param entry destination entry
     * @return the euklidean distance
     */
    float operator-(const SimpleNodeEntry& entry) const;

    cGate* ingate; //!< ingate of the SimpleUDP module of this terminal

    simtime_t txFinished; //!< send queue finished
    simtime_t txMaxQueueTime; //!< maximum time for packets to be queued
    simtime_t accessDelay; //!< first hop delay
    double bandwidth; //!< bandwidth in access net
    double errorRate; //!< packet loss rate

    NodeRecord* nodeRecord;
    int index;
};


#endif // __SIMPLENODEENTRY_H
