//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public License
// as published by the Free Software Foundation; either version 2
// of the License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
//

/**
 * @file SingleHostConfigurator.cc
 * @author Stephan Krause
 * @author Bernhard Heep (migrateRandomNode)
 */

#include <stdio.h>
#include <unistd.h>
#include <fcntl.h>
#include <net/if.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/ioctl.h>
#include <sys/socket.h>
#include <arpa/inet.h>
#include <netinet/in.h>

#include <stun/udp.h>
#include <stun/stun.h>

#include <omnetpp.h>

#include "InterfaceTable.h"
#include "InterfaceEntry.h"
#include "IPv4InterfaceData.h"
#include "IPAddressResolver.h"

#include <PeerInfo.h>
#include <RoutingTable.h>
#include <NodeHandle.h>
#include <BootstrapOracleAccess.h>

#include "SingleHostConfigurator.h"

Define_Module(SingleHostConfigurator);

void SingleHostConfigurator::initializeUnderlay(int stage)
{
    int sock_fd;
    struct ifreq req;
    StunAddress4 publicAddr, stunServerAddr;
    IPvXAddress addr;

    if(stage != MAX_STAGE_UNDERLAY)
        return;

    overlayTerminalCount = 1;

    // Set IP, Routes, etc
    cModule* node = parentModule()->submodule("singleHost", 0);

    std::string interface = std::string(par("nodeInterface").stringValue());
    std::string stunServer = std::string(par("stunServer").stringValue());
    std::string nodeIP = std::string(par("nodeIP").stringValue());

    if (!interface.size() && !stunServer.size() && !nodeIP.size()) {
        throw new cRuntimeError("SingleHostConfigurator::initializeUnderlay(): "
                                "Please specify at least one of "
                                "**.nodeInterface, **.nodeIP, or **.stunServer");
    }

    if (interface.size()) {
        if ((sock_fd = socket(PF_INET, SOCK_DGRAM, IPPROTO_IP)) != -1) {
            snprintf(req.ifr_name, sizeof(req.ifr_name), interface.c_str());

            if (!ioctl(sock_fd, SIOCGIFADDR, &req)) {
                addr = IPAddress(inet_ntoa(((struct sockaddr_in *)
                        (&req.ifr_addr))->sin_addr));
            } else {
                throw new cRuntimeError("SingleHostConfigurator::"
                                        "initializeUnderlay(): "
                                        "Invalid interface!");
            }

            close(sock_fd);
        }
    } else if (nodeIP.size()) {
        addr = IPAddress(nodeIP.c_str());
    }

    if (stunServer.size()) {
        // TODO: use and set overlayPort!
        throw new cRuntimeError("SingleHostConfigurator::initializeUnderlay():"
                                " Not implemented yet!");
        int srcPort = 0;
        publicAddr.addr = 0;
        publicAddr.port = srcPort;
        stunServerAddr.addr = 0;
        char tmpAddr[512];
        strncpy (tmpAddr, stunServer.c_str(), 512);

        if (stunParseServerName(tmpAddr, stunServerAddr)) {
            bool presPort = false;
            bool hairpin = false;

            NatType stype = stunNatType(stunServerAddr, false, &presPort,
                                        &hairpin, srcPort, &publicAddr);
            if (stype != StunTypeOpen) {
                EV << "SingleHostConfigurator::initializeUnderlay(): "
                   << "Node is behind NAT or invalid STUN server configuration!"
                   << std::endl;
            }

            publicAddr.addr = htonl(publicAddr.addr);
            addr = IPAddress(inet_ntoa(*((struct in_addr *)(&(publicAddr.addr)))));
        }
    }

    IPvXAddress gw = addr;
    InterfaceEntry* ifEntry = IPAddressResolver().interfaceTableOf(node)->
	    interfaceByName("outDev");
    RoutingTable* rTable = (RoutingTable*) node->submodule("routingTable", 0);
    ifEntry->ipv4()->setInetAddress(addr.get4());
    ifEntry->ipv4()->setNetmask(IPAddress::ALLONES_ADDRESS);

    RoutingEntry* te = new RoutingEntry();
    te->host = IPAddress::UNSPECIFIED_ADDRESS;
    te->netmask = IPAddress::UNSPECIFIED_ADDRESS;
    te->gateway = gw.get4();
    te->interfaceName = "outDev";
    te->interfacePtr = ifEntry;
    te->type = RoutingEntry::REMOTE;
    te->source = RoutingEntry::MANUAL;
    rTable->addRoutingEntry(te);

    //add node to bootstrap oracle
    PeerInfo* info = new PeerInfo(0, node->id());
    bootstrapOracle->addPeer(addr, info);

    if (strlen(par("bootstrapIP")) > 0) {
        PeerInfo* bootstrapInfo = new PeerInfo(0, -1);
	bootstrapOracle->addPeer(IPAddress(par("bootstrapIP").stringValue()),
				 bootstrapInfo);

	bootstrapOracle->registerPeer(NodeHandle(OverlayKey::ONE,
          IPAddress(par("bootstrapIP").stringValue()), par("bootstrapPort")));
    }

    // update display
    setDisplayString();

}

void SingleHostConfigurator::setDisplayString()
{
    // updates the statistics display string
    char buf[80];
    sprintf(buf, "%i overlay terminals", overlayTerminalCount);
    displayString().setTagArg("t", 0, buf);
}

void SingleHostConfigurator::finishUnderlay()
{
    if (!isInInitPhase()) {
        struct timeval now, diff;
        gettimeofday(&now, NULL);
        timersub(&now, &initFinishedTime, &diff);
        printf("Simulation time: %li.%06li\n", diff.tv_sec, diff.tv_usec);
    }
}
